
/***************************************************************************
 *                         _______ _______ _______                         *
 *                        |  ___  |____   |  ___  |                        *
 *                        | |   |_|    / /| |   |_|                        *
 *                        | |_____    / / | |_____                         *
 *                        |_____  |  / /  |_____  |                        *
 *                         _    | | / /    _    | |                        *
 *                        | |___| |/ /____| |___| |                        *
 *                        |_______|_______|_______|                        *
 *                                                                         *
 *                             Wiimms SZS Tools                            *
 *                           http://szs.wiimm.de/                          *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This file is part of the SZS project.                                 *
 *   Visit http://szs.wiimm.de/ for project details and sources.           *
 *                                                                         *
 *   Copyright (c) 2011-2016 by Dirk Clemens <wiimm@wiimm.de>              *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   See file gpl-2.0.txt or http://www.gnu.org/licenses/gpl-2.0.txt       *
 *                                                                         *
 ***************************************************************************/

// all about BMG files:
//	http://wiki.tockdom.de/index.php?title=BMG

///////////////////////////////////////////////////////////////////////////////

#ifndef SZS_LIB_KMP_H
#define SZS_LIB_KMP_H 1

#include "lib-std.h"
#include "db-object.h"

//
///////////////////////////////////////////////////////////////////////////////
///////////////			definitions			///////////////
///////////////////////////////////////////////////////////////////////////////

#define KMP_MAGIC		"RKMD"
#define KMP_MAGIC_NUM		0x524B4D44

#define KMP_TEXT_MAGIC		"#KMP"
#define KMP_TEXT_MAGIC_NUM	0x234b4d50

#define KMP_MAX_PH_LINK		  6
#define KMP_MAX_GROUP		255
#define KMP_MAX_POINT		255
#define KMP_MAX_NAME_LEN	 20

#define KMP_DIST_HINT		10.0
#define KMP_DIST_WARN		 1.0

#define GOBJ_ELINE_CONTROL	0x014
#define GOBJ_ICE		0x017
#define GOBJ_ITEMBOX		0x065
#define GOBJ_SUN_DS		0x072
#define	GOBJ_HEYHO_SHIP		0x0ce
#define	GOBJ_HEYHO_BALL		0x0ea
#define GOBJ_OBAKE_BLOCK	0x13c
#define GOBJ_PYLON01		0x144

//
///////////////////////////////////////////////////////////////////////////////
///////////////			enum kmp_entry_t		///////////////
///////////////////////////////////////////////////////////////////////////////

typedef enum kmp_entry_t
{
    KMP_KTPT,	//	Starting positions of racers
    KMP_ENPT,	//	Enemy route points
    KMP_ENPH,	//	Enemy route points header: Divide enemy points into groups
    KMP_ITPT,	//	Item route points
    KMP_ITPH,	//	Item points header: Divide item points into groups
    KMP_CKPT,	//	Check points
    KMP_CKPH,	//	Checkpoint header: Divide checkpoints into groups
    KMP_GOBJ,	// 'O',	Global objects
    KMP_POTI,	// 'R',	Routes
    KMP_AREA,	// 'A',	Camera and other areas
    KMP_CAME,	// 'C',	Cameras
    KMP_JGPT,	// 'J',	Respawn positions
    KMP_CNPT,	// 'N',	Cannon target positions
    KMP_MSPT,	// 'M',	End battle positions
    KMP_STGI,	// 'S',	Stage information

    KMP_N_SECT	// number of KMP sections

} kmp_entry_t;

extern const uint kmp_entry_size[KMP_N_SECT+1];
extern const KeywordTab_t kmp_section_name[];

//
///////////////////////////////////////////////////////////////////////////////
///////////////			kmp_file_t			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct kmp_file_t
{
  /*0x00*/  char magic[4];	// always KMP_MAGIC = "RKMD"
  /*0x04*/  u32  file_size;	// size of file
  /*0x08*/  u16  n_sect;	// number of sections
  /*0x0a*/  u16  sect_base_off;	// base offset for all sections, usually 0x4c
  /*0x0c*/  u32  unknown_0x0c;

  /*0x10*/  u32  sect_off[KMP_N_SECT];	// offsets of all sections,
					// relative to 'sect_base_off'
  /*0x4c*/
}
__attribute__ ((packed)) kmp_file_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////			kmp list & entries		///////////////
///////////////////////////////////////////////////////////////////////////////
// [[kmp_list_head_t]]

typedef struct kmp_list_head_t
{
  /*0x00*/  char	magic[4];	// related string
  /*0x04*/  u16		n_entry;	// number of entries
  /*0x06*/  u16		value;		// addititional value
					//   POTI: total number of points

  /*0x08*/  u8		entry[0];	// entries
}
__attribute__ ((packed)) kmp_list_head_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_ktpt_entry_t]]

typedef struct kmp_ktpt_entry_t
{
  // Kart Points: Starting positions of racers.

  /*0x00*/  float32	position[3];	// start position
  /*0x0c*/  float32	rotation[3];	// start rotation
  /*0x18*/  s16		player_index;	// index or -1 for all players
  /*0x1a*/  u16		unknown;	// padding?
  /*0x1c*/
}
__attribute__ ((packed)) kmp_ktpt_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_enpt_entry_t]]  [[kmp_itpt_entry_t]]

typedef struct kmp_enpt_entry_t
{
  // Enemy Points: Routes of the cpu racers

  /*0x00*/  float32	position[3];	// start position
  /*0x0c*/  float32	scale;
  /*0x10*/  u16		prop[2];	// point properties
  /*0x14*/
}
__attribute__ ((packed)) kmp_enpt_entry_t;

// Item Points: Routes of red shells and bullet bill
typedef kmp_enpt_entry_t kmp_itpt_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_enph_entry_t]]

typedef struct kmp_enph_entry_t
{
  // Enemy Points grouping: Route links of cpu racers

  /*0x00*/  u8		pt_start;
  /*0x01*/  u8		pt_len;
  /*0x02*/  u8		prev[KMP_MAX_PH_LINK];
  /*0x08*/  u8		next[KMP_MAX_PH_LINK];
  /*0x0e*/  u16		unknown;	// padding?
  /*0x10*/
}
__attribute__ ((packed)) kmp_enph_entry_t;

// Item Points grouping: Route links of cpu racers
typedef kmp_enph_entry_t kmp_itph_entry_t;

// Check point links
typedef kmp_enph_entry_t kmp_ckph_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_ckpt_entry_t]]

typedef struct kmp_ckpt_entry_t
{
  // Check Points

  /*0x00*/  float32	left[2];	// left hand position of checkpoint line
  /*0x08*/  float32	right[2];	// right hand position of checkpoint line
  /*0x10*/  u8		respawn;	// index into JGPT section
  /*0x11*/  u8		mode;		// 0:lap counter, -1:other, *:must cross
  /*0x12*/  u8		prev;		// previous checkpoint in sequence
  /*0x13*/  u8		next;		// next checkpoint in sequence
  /*0x14*/
}
__attribute__ ((packed)) kmp_ckpt_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_gobj_entry_t]]

typedef struct kmp_gobj_entry_t
{
  // Global objects

  /*0x00*/  u16		obj_id;		// Id of global object
  /*0x02*/  u16		unknown;	// padding?
  /*0x04*/  float32	position[3];
  /*0x10*/  float32	rotation[3];
  /*0x1c*/  float32	scale[3];
  /*0x28*/  u16		route_id;
  /*0x2a*/  u16		setting[8];	// object setting
  /*0x3a*/  u16		presence_flags;
  /*0x3c*/
}
__attribute__ ((packed)) kmp_gobj_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_poti_group_t]]

typedef struct kmp_poti_group_t
{
  // Routes

  /*0x00*/  u16		n_point;	// number of related points
  /*0x02*/  u8		smooth;		// 1: enable smooth moving
  /*0x03*/  u8		back;		// 1: enable forward+backward
  /*0x04*/
}
__attribute__ ((packed)) kmp_poti_group_t;

//-----------------------------------------------------------------------------
// [[kmp_poti_point_t]]

typedef struct kmp_poti_point_t
{
  // Routes

  /*0x00*/  float32	position[3];
  /*0x0c*/  u16		speed;		// speed in 60pt/s
  /*0x0e*/  u16		unknown;	// Nintendo: 0 (~94%), 1,2,3,4,5,6,9,12
  /*0x10*/
}
__attribute__ ((packed)) kmp_poti_point_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_area_entry_t]]

typedef struct kmp_area_entry_t
{
  // Camera and other areas

  /*0x00*/  u8		mode;		// mode of area
  /*0x01*/  u8		type;		// type of area
  /*0x02*/  u8		dest_id;	// index into CAME
  /*0x03*/  u8		unknown;
  /*0x04*/  float32	position[3];
  /*0x10*/  float32	rotation[3];
  /*0x1c*/  float32	scale[3];
  /*0x28*/  u16		setting[4];
  /*0x30*/
}
__attribute__ ((packed)) kmp_area_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_came_entry_t]]

typedef struct kmp_came_entry_t
{
  // Cameras

  /*0x00*/  u8		type;		// type of camera
  /*0x01*/  u8		next;		// next camera
  /*0x02*/  u8		unknown_0x02;
  /*0x03*/  u8		route;		// route index -> POTI
  /*0x04*/  u16		came_speed;	// unsure!
  /*0x06*/  u16		zoom_speed;	// zoom speed in units per 100/60 sec
  /*0x08*/  u16		viewpt_speed;	// view point speed in units per 100/60 sec
  /*0x0a*/  u16		unknown_0x0a;
  /*0x0c*/  float32	position[3];	// camera position
  /*0x18*/  float32	rotation[3];	// rotation? (always 0)
  /*0x24*/  float32	zoom_begin;	// zoom value begin
  /*0x28*/  float32	zoom_end;	// zoom value end
  /*0x2c*/  float32	viewpt_begin[3];// view point begin
  /*0x38*/  float32	viewpt_end[3];	// view point end
  /*0x44*/  float32	time;		// active time of the camera
  /*0x48*/
}
__attribute__ ((packed)) kmp_came_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_jgpt_entry_t]]

typedef struct kmp_jgpt_entry_t
{
  // Jugem points: Respawn positions

  /*0x00*/  float32	position[3];
  /*0x0c*/  float32	rotation[3];
  /*0x18*/  s16		id;
  /*0x1a*/  u16		effect;
  /*0x1c*/
}
__attribute__ ((packed)) kmp_jgpt_entry_t;

// Cannon Points: Cannon target positions
typedef kmp_jgpt_entry_t kmp_cnpt_entry_t;

// After battles and competitions have ended the players are placed on this point(s)
typedef kmp_jgpt_entry_t kmp_mspt_entry_t;

///////////////////////////////////////////////////////////////////////////////
// [[kmp_stgi_entry_t]]

typedef struct kmp_stgi_entry_t
{
  //  Stage Information

  /*0x00*/  u8		lap_count;		// XML suggest but ignored
  /*0x01*/  u8		pole_pos;		// 0:left, 1:right
  /*0x02*/  u8		narrow_start;		// 0:normal, 1:closer together
  /*0x03*/  u8		unknown_03;		// 0 or 1
  /*0x04*/  u32		unknown_04;		// 0xe6e6e6 or 0xffffff (color?)
  /*0x08*/  u8		unknown_08;		// 0x32 or 0x4b
  /*0x09*/  u8		unknown_09;		// always 0
  /*0x0a*/  u16		unknown_0a;		// nintendo: always 0
						// custom: used as speed modifier
  /*0x0c*/
}
__attribute__ ((packed)) kmp_stgi_entry_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  kmp_group_t			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef char kmp_group_name_t[KMP_MAX_NAME_LEN+1];

//-----------------------------------------------------------------------------
// [[kmp_group_t]]

typedef struct kmp_group_t
{
    int			first_index;		// index of first point
    int			last_index;		// index of last point
    bool		no_rotate;		// true: do not rotate links

    kmp_group_name_t	name;			// name of group
    kmp_group_name_t	link[KMP_MAX_PH_LINK];	// link names

} kmp_group_t;

//-----------------------------------------------------------------------------
// [[kmp_group_list_t]]

typedef struct kmp_group_list_t
{
    uint		used;			// used groups
    kmp_group_t		group[KMP_MAX_GROUP];	// list with groups
    char		selector;		// '!' selector
    Line_t		line;			// '$LINE' settings

} kmp_group_list_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////		kmp_rtobj_t, kmp_rtobj_list_t		///////////////
///////////////////////////////////////////////////////////////////////////////
// [[kmp_rtobj_t]]

typedef struct kmp_rtobj_t
{
    u16			obj;			// object to use
    float3		shift;			// add to position
    float3		scale;			// store as 'scale' factor
    float3		rotate;			// add to rotation

} kmp_rtobj_t;

//-----------------------------------------------------------------------------
// [[kmp_rtobj_list_t]]

typedef struct kmp_rtobj_list_t
{
    kmp_rtobj_t		rtobj;			// object to use
    u16			index;			// index to POTI group

} kmp_rtobj_list_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////			 kmp_mode_t			///////////////
///////////////////////////////////////////////////////////////////////////////
// [[kmp_mode_t]]

typedef enum kmp_mode_t
{
    //--- general options

    KMPMD_FORCE		= 0x00000001,	// like --force: accept little repairable issues
     KMPMD_M_GENERAL	 = KMPMD_FORCE,


    //--- KMP patching options

    KMPMD_NEW		= 0x00000010,	// create a new KMP file even if nothing changed
    KMPMD_RM_SPCITEM	= 0x00000020,	// remove special items
    KMPMD_REVERSE	= 0x00000040,	// reverse the track order

    KMPMD_1_LAP		= 0x00000100,	// set lap count to #1
    KMPMD_2_LAPS	= 0x00000200,	// set lap count to #2
    KMPMD_3_LAPS	= 0x00000300,	// set lap count to #3
    KMPMD_4_LAPS	= 0x00000400,	// set lap count to #4
    KMPMD_5_LAPS	= 0x00000500,	// set lap count to #5
    KMPMD_6_LAPS	= 0x00000600,	// set lap count to #6
    KMPMD_7_LAPS	= 0x00000700,	// set lap count to #7
    KMPMD_8_LAPS	= 0x00000800,	// set lap count to #8
    KMPMD_9_LAPS	= 0x00000900,	// set lap count to #9
     KMPMD_M_LAPS	= 0x00000f00,
     KMPMD_SHIFT_LAPS	= 8,

    KMPMD_POLE_LEFT	= 0x00001000,	// set pole position to left side
    KMPMD_POLE_RIGHT	= 0x00002000,	// set pole position to right side
    KMPMD_SP_WIDE	= 0x00004000,	// set start mode 'wide'
    KMPMD_SP_NARROW	= 0x00008000,	// set start mode 'narrow'
     KMPMD_M_POLE	 = KMPMD_POLE_LEFT
			 | KMPMD_POLE_RIGHT,
     KMPMD_M_STARTPOS	 = KMPMD_SP_WIDE
			 | KMPMD_SP_NARROW,
     KMPMD_M_STGI	 = KMPMD_M_POLE
			 | KMPMD_M_STARTPOS
			 | KMPMD_M_LAPS,

    KMPMD_FIX_CKPH	= 0x00010000,	// automatic calculations of section CKPH
    KMPMD_FIX_ENPH	= 0x00020000,	// automatic calculations of section ENPH
    KMPMD_FIX_ITPH	= 0x00040000,	// automatic calculations of section ITPH
    KMPMD_FIX_CKNEXT	= 0x00080000,	// CKPT: auto calc of 'prev' and 'next'
    KMPMD_FIX_CKJGPT	= 0x00100000,	// CKPT: fix invalid JGPT indices
     KMPMD_M_FIX_PH	 = KMPMD_FIX_CKPH
			 | KMPMD_FIX_ENPH
			 | KMPMD_FIX_ITPH,
     KMPMD_M_FIX_CK	 = KMPMD_FIX_CKPH
			 | KMPMD_FIX_CKNEXT
			 | KMPMD_FIX_CKJGPT,
     KMPMD_M_FIX_ALL	 = KMPMD_M_FIX_PH
			 | KMPMD_M_FIX_CK,

     KMPMD_M_PATCH	 = KMPMD_NEW
			 | KMPMD_RM_SPCITEM
			 | KMPMD_REVERSE
			 | KMPMD_M_STGI
			 | KMPMD_M_FIX_ALL,



    //--- hidden options

    KMPMD_LOG		= 0x01000000,	// enable kmp trace log to stdout
    KMPMD_SILENT	= 0x02000000,	// disable logging of patching operations
    KMPMD_INPLACE	= 0x04000000,	// force inplace patching
     KMPMD_M_HIDDEN	 = KMPMD_LOG
			 | KMPMD_SILENT
			 | KMPMD_INPLACE,


    //--- test modes

    KMPMD_TEST0		= 0x00000000,
    KMPMD_TEST1		= 0x10000000,
    KMPMD_TEST2		= 0x20000000,
    KMPMD_TEST3		= 0x30000000,
     KMPMD_M_TEST	 = KMPMD_TEST3, // all bits!


    //--- summaries

    KMPMD_F_HIDE	= 0x40000000,		// help flag to hide
						// modes in PrintKmpMode()

    KMPMD_M_DEFAULT	= 0,			// default settings

    KMPMD_M_ALL		= KMPMD_M_GENERAL	// all relevant bits
			| KMPMD_M_PATCH
			| KMPMD_M_HIDDEN
			| KMPMD_M_TEST,

    KMPMD_M_PRINT	= KMPMD_M_ALL		// all relevant bits for obj output
			& ~KMPMD_M_HIDDEN,

} kmp_mode_t;

//-----------------------------------------------------------------------------
// [[kmp_tform_t]]

typedef enum kmp_tform_t
{
    //--- single modes

    KMPTF_AREA_POS	= 0x000001,
    KMPTF_AREA_ROTATE	= 0x000002,
    KMPTF_AREA_SCALE	= 0x000004,

    KMPTF_CAME_POS	= 0x000008,

    KMPTF_CKPT_POS	= 0x000010,

    KMPTF_CNPT_POS	= 0x000020,
    KMPTF_CNPT_ROTATE	= 0x000040,

    KMPTF_ENPT_POS	= 0x000080,
    KMPTF_ENPT_SCALE	= 0x000100,

    KMPTF_GOBJ_POS	= 0x000200,
    KMPTF_GOBJ_ROTATE	= 0x000400,
    KMPTF_GOBJ_SCALE	= 0x000800,

    KMPTF_ITPT_POS	= 0x001000,
    KMPTF_ITPT_SCALE	= 0x002000,

    KMPTF_JGPT_POS	= 0x004000,
    KMPTF_JGPT_ROTATE	= 0x008000,

    KMPTF_KTPT_POS	= 0x010000,
    KMPTF_KTPT_ROTATE	= 0x020000,

    KMPTF_MSPT_POS	= 0x040000,
    KMPTF_MSPT_ROTATE	= 0x080000,

    KMPTF_POTI_POS	= 0x100000,


    //--- section summaries

    KMPTF_AREA		= KMPTF_AREA_POS | KMPTF_AREA_ROTATE | KMPTF_AREA_SCALE,
    KMPTF_CAME		= KMPTF_CAME_POS,
    KMPTF_CKPT		= KMPTF_CKPT_POS,
    KMPTF_CNPT		= KMPTF_CNPT_POS | KMPTF_CNPT_ROTATE,
    KMPTF_ENPT		= KMPTF_ENPT_POS | KMPTF_ENPT_SCALE,
    KMPTF_GOBJ		= KMPTF_GOBJ_POS | KMPTF_GOBJ_ROTATE | KMPTF_GOBJ_SCALE,
    KMPTF_ITPT		= KMPTF_ITPT_POS | KMPTF_ITPT_SCALE,
    KMPTF_JGPT		= KMPTF_JGPT_POS | KMPTF_JGPT_ROTATE,
    KMPTF_KTPT		= KMPTF_KTPT_POS | KMPTF_KTPT_ROTATE,
    KMPTF_MSPT		= KMPTF_MSPT_POS | KMPTF_MSPT_ROTATE,
    KMPTF_POTI		= KMPTF_POTI_POS,


    //--- masks

    KMPTF_M_ALL		= 0x1fffff,
    KMPTF_M_DEFAULT	= KMPTF_M_ALL,

    //--- flags

    KMPTF_F_HIDE	= 0x1000000

} kmp_tform_t;

//-----------------------------------------------------------------------------
// [[kmp_pt_flag_t]]

typedef enum kmp_pt_flag_t
{
	KPFL_FALL	= 0x01,  // F: fall-down enabled
	KPFL_SNAP	= 0x02,  // S: used for SNAP
	KPFL_JGPT	= 0x04,  // J: used for automatic JGPT creation

	KPFL_M_ENPT	= KPFL_FALL
			| KPFL_SNAP
			| KPFL_JGPT,

	KPFL_M_ITPT	= KPFL_FALL
			| KPFL_JGPT,

	KPFL_M_JGPT	= KPFL_FALL
			| KPFL_SNAP,

	KPFL_M_DEFAULT	= KPFL_M_ENPT
			| KPFL_M_ITPT
			| KPFL_M_JGPT,
}
__attribute__ ((packed)) kmp_pt_flag_t;

//-----------------------------------------------------------------------------

extern kmp_mode_t KMP_MODE;
extern kmp_tform_t KMP_TFORM;
extern bool  speed_mod_active;
extern float speed_mod_factor;
extern u16   speed_mod_val;

int  ScanOptKmp ( ccp arg );
uint PrintKmpMode( char *buf, uint bufsize, kmp_mode_t mode );
ccp  GetKmpMode();

int  ScanOptTformKmp ( ccp arg );
uint PrintKmpTform( char *buf, uint bufsize, kmp_tform_t mode );
ccp  GetKmpTform();

int   ScanOptSpeedMod ( ccp arg );
u16   SpeedMod2u16 ( float speed );
float SpeedMod2float ( u16 hex );

bool KMP_ACTION_LOG ( bool is_patch, const char * format, ... )
	__attribute__ ((__format__(__printf__,2,3)));

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    kmp_t			///////////////
///////////////////////////////////////////////////////////////////////////////
// [[kmp_t]]

struct kcl_t;

typedef struct kmp_t
{
    //--- base info

    ccp			fname;			// alloced filename of loaded file
    FileAttrib_t	fatt;			// file attribute
    file_format_t	fform;			// FF_KMP | FF_KMP_TXT

    //--- data

    List_t		dlist[KMP_N_SECT];	// data lists
    List_t		poti_point;		// POTI points list

    int			revision;		// tool revision
    u32			unknown_0x0c;		// unknown value if file header
    u16			value[KMP_N_SECT+1];	// 'value' of section header
    IndexList_t		index[KMP_N_SECT];	// named indices

    // [[2do]] [[obsolete]] -> replaced by 'index'
    kmp_group_name_t  ckph_name[KMP_MAX_GROUP];	// group names of check points
    kmp_group_name_t  enph_name[KMP_MAX_GROUP];	// group names of enemy points
    kmp_group_name_t  itph_name[KMP_MAX_GROUP];	// group names of item points

    char		* info;			// string behind last section
    const u8		* max_scanned;		// raw scanning: max scanned area

    // snap vaiables: if not VAR_UNSET => move each JGPT to next ENPT and and snap value
    // priority: 'snap2' before 'snap' before 'hsnap'
    Var_t		jgpt_snap2;		// not VAR_UNSET: snap2 JGPT
    Var_t		cnpt_snap2;		// not VAR_UNSET: snap2 CNPT
    Var_t		mspt_snap2;		// not VAR_UNSET: snap2 MSPT
    Var_t		jgpt_snap;		// not VAR_UNSET: snap JGPT
    Var_t		cnpt_snap;		// not VAR_UNSET: snap CNPT
    Var_t		mspt_snap;		// not VAR_UNSET: snap MSPT
    Var_t		jgpt_hsnap;		// not VAR_UNSET: hsnap JGPT
    Var_t		cnpt_hsnap;		// not VAR_UNSET: hsnap CNPT
    Var_t		mspt_hsnap;		// not VAR_UNSET: hsnap MSPT

    // route objects
    kmp_rtobj_t		rtobj;			// predefined route object
    kmp_rtobj_t		rtobj_enpt;		// >0: for each ENPT 1 object
    kmp_rtobj_t		rtobj_itpt;		// >0: for each ITPT 1 object
    kmp_rtobj_t		rtobj_jgpt;		// >0: for each JGPT 1 object
    kmp_rtobj_t		rtobj_cnpt;		// >0: for each CNPT 1 object
    kmp_rtobj_t		rtobj_mspt;		// >0: for each MSPT 1 object
    kmp_rtobj_list_t	rtobj_poti[50];		// objects for POTI routes
    uint		n_rtobj_poti;		// number of valid elements
						// in 'rtobj_poti[]'

    // objects on check points
    int			ckpt_object_mode;	// view mode
    int			ckpt_object1;		// object 1
    int			ckpt_object2;		// object 2
    Var_t		ckpt_object1_scale;	// scale of object 1
    Var_t		ckpt_object2_scale;	// scale of object 2
    float		ckpt_object1_base;	// y base of object 1
    float		ckpt_object2_base;	// y base of object 2

    // special settings

    u8			ckpt_auto_respawn[KMP_MAX_POINT]; // >0: assign nearest JGPT
    u8			enpt_flags[KMP_MAX_POINT]; // kmp_pt_flag_t
    u8			itpt_flags[KMP_MAX_POINT]; // kmp_pt_flag_t
    u8			jgpt_flags[KMP_MAX_POINT]; // kmp_pt_flag_t
    
    u16			enpt_mask[2];
    u16			itpt_mask[2];

    int			jgpt_auto_define;	// >0 and valid section: auto define JGPT
    float		jgpt_auto_distance;	// >0.0: define only if no other pt in distance
    Var_t		jgpt_auto_adjust;	// add this vector to aauto defined point
    int			jgpt_auto_remove;	// >0: remove unused: 1=auto defined, 2=all

    // misc
    bool		multi_ckpt_mode_0_ok;	// true: disable warnings about mode 0
    struct kcl_t	*fall_kcl;		// NULL or pointer to own fall-down kcl
    bool		enable_fall_kcl;	// true: enable 'fall_kcl' support

    //--- raw data, created by CreateRawKMP()

    u8			* raw_data;		// NULL or data
    uint		raw_data_size;		// size of 'raw_data'

} kmp_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP interface			///////////////
///////////////////////////////////////////////////////////////////////////////

void InitializeKMP ( kmp_t * kmp );
void ResetKMP ( kmp_t * kmp );
void ClearSectionsKMP ( kmp_t * kmp );
const VarMap_t * SetupVarsKMP();

///////////////////////////////////////////////////////////////////////////////

enumError ScanRawKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    const void		* data,		// data to scan
    uint		data_size	// size of 'data'
);

//-----------------------------------------------------------------------------

enumError ScanTextKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    const void		* data,		// data to scan
    uint		data_size	// size of 'data'
);

//-----------------------------------------------------------------------------

enumError ScanKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    CheckMode_t		mode		// not NULL: call CheckKMP(mode)
);

//-----------------------------------------------------------------------------

struct raw_data_t;

enumError ScanRawDataKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    struct raw_data_t	* raw,		// valid raw data
    CheckMode_t		mode		// not NULL: call CheckKMP(mode)
);

///////////////////////////////////////////////////////////////////////////////

enumError LoadKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    ccp			fname,		// valid pointer to filenname
    bool		ignore_no_file,	// ignore if file does not exists
					// and return warning ERR_NOT_EXISTS
    CheckMode_t		mode		// not NULL: call CheckKMP(mode)
);

//-----------------------------------------------------------------------------

enumError CreateRawKMP
(
    kmp_t		* kmp		// pointer to valid KMP
);

//-----------------------------------------------------------------------------

enumError SaveRawKMP
(
    kmp_t		* kmp,		// pointer to valid KMP
    ccp			fname,		// filename of destination
    bool		set_time	// true: set time stamps
);

//-----------------------------------------------------------------------------

enumError SaveTextKMP
(
    const kmp_t		* kmp,		// pointer to valid KMP
    ccp			fname,		// filename of destination
    bool		set_time	// true: set time stamps
);

///////////////////////////////////////////////////////////////////////////////

bool SortGOBJ
(
    kmp_t		* kmp,		// pointer to valid KMP
    int			sort_mode	// <=0: dont sort
					//   1: group only
					//   2: secondary sort by angle
);

//-----------------------------------------------------------------------------

bool TransformKMP
(
    kmp_t		* kmp		// pointer to valid KMP
);

//-----------------------------------------------------------------------------

bool PatchKMP
(
    kmp_t		* kmp		// pointer to valid KMP
);

//-----------------------------------------------------------------------------

bool PatchRawDataKMP
(
    void		* data,		// data to scan
    uint		data_size	// size of 'data'
);

///////////////////////////////////////////////////////////////////////////////

int GetNextPointKMP
(
    const kmp_t		* kmp,		// KMP data structure
    uint		sect,		// KMP PT section index
    int			pt_index,	// index of point, <0: add N
    int			sub_index,	// index into 'next'
					//   -1: use 'pt_index+1' in a group
					//       find first valid at group end
					//  !-1 && KMP_CKPT: try to use 'next' link
    bool		unique,		// return 'not_found_val' on duplicates
    int			not_found_val,	// return this value, if next point not found
    int			*ret_pt_index,	// not NULL: store fixed 'pt_index' here
    int			*ret_sub_index,	// not NULL: store used sub index here
    bool		*ret_new_group	// not NULL: store true, if new group
);

///////////////////////////////////////////////////////////////////////////////

int CheckConvexCKPT
(
    // returns
    //  -1: illegal index
    //   0: convex
    //   1: concave

    const kmp_t		*kmp,		// KMP data structure
    uint		ckpt1,		// index of first check point (robust)
    uint		ckpt2,		// index of second check point (robust)
    double		*min_distance	// not NULL: store minimum distance here
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP itembox support		///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct Itembox_t
{
    u16		obj_id;		// >0: ID of analyzed itembox
    u32		obj_flags;	// >0: ObjectFlags_t of analyzed itembox
    u16		item_player;	// mode of player item 
    u16		item_enemy;	// mode of enemy item 
    bool	unsure_player;	// true: not sure, if 'item_player' is correct
    bool	unsure_enemy;	// true: not sure, if 'item_enemy' is correct
}
Itembox_t;

// initialze 'ibox' and returns 'ibox->obj_id'
u16 AnalyseItembox ( Itembox_t *ibox, const kmp_gobj_entry_t *op );

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP draw support		///////////////
///////////////////////////////////////////////////////////////////////////////
// [[draw_mode_t]]

typedef enum draw_mode_t
{
    DMD_NONE		= 0,

    DMD_KTPT		= 1 << KMP_KTPT,
    DMD_ENPT		= 1 << KMP_ENPT,
    DMD_ITPT		= 1 << KMP_ITPT,
    DMD_CKPT		= 1 << KMP_CKPT,
    //DMD_GOBJ		= 1 << KMP_GOBJ,
    DMD_POTI		= 1 << KMP_POTI,
    //DMD_AREA		= 1 << KMP_AREA,
    //DMD_CAME		= 1 << KMP_CAME,
    DMD_JGPT		= 1 << KMP_JGPT,
    DMD_CNPT		= 1 << KMP_CNPT,
    //DMD_MSPT		= 1 << KMP_MSPT,

     DMD_SHIFT_PFLAGS	= KMP_N_SECT,		// shift right to get presence flags
    DMD_1OFFLINE	= 1 << KMP_N_SECT,
    DMD_2OFFLINE	= DMD_1OFFLINE << 1,
    DMD_3OFFLINE	= DMD_2OFFLINE << 1,
    DMD_1ONLINE		= DMD_3OFFLINE << 1,
    DMD_2ONLINE		= DMD_1ONLINE << 1,
    DMD_3ONLINE		= DMD_2ONLINE << 1,

    DMD_CJGPT		= DMD_3ONLINE << 1,
    DMD_ITEMBOX		= DMD_CJGPT << 1,
    DMD_ROADOBJ		= DMD_ITEMBOX << 1,
    DMD_SOLIDOBJ	= DMD_ROADOBJ << 1,
    DMD_KCL		= DMD_SOLIDOBJ << 1,
    DMD_DETAILED	= DMD_KCL << 1,
    DMD_WARNINGS	= DMD_DETAILED << 1,

    DMD_M_OFFLINE	= DMD_1OFFLINE	// offline presence flags
			| DMD_2OFFLINE
			| DMD_3OFFLINE,

    DMD_M_ONLINE	= DMD_1ONLINE	// online presence flags
			| DMD_2ONLINE
			| DMD_3ONLINE,

    DMD_M_PFLAGS	= DMD_M_OFFLINE	// all presence flags
			| DMD_M_ONLINE,

    DMD_M_FLAGS		= DMD_DETAILED	// flags
			| DMD_WARNINGS,

    DMD_M_RESPAWN	= DMD_CKPT	// all about respawning
			| DMD_JGPT
			| DMD_CJGPT
			| DMD_KTPT
			| DMD_KCL,

    DMD_M_PTS		= DMD_JGPT	// single points
			| DMD_KTPT,

    DMD_M_GOBJ		= DMD_ITEMBOX	// global objects
			| DMD_ROADOBJ
			| DMD_SOLIDOBJ,

    DMD_M_ADD		= DMD_M_PTS	// modes that add faces
			| DMD_M_GOBJ
			| DMD_ENPT
			| DMD_ITPT
			| DMD_CKPT
			| DMD_POTI
			| DMD_CNPT
			| DMD_CJGPT,

    DMD_M_ALL		= DMD_M_ADD	// all possible modes
			| DMD_KCL
			| DMD_M_PFLAGS
			| DMD_M_FLAGS,

    DMD_M_RESET		= DMD_M_ALL	// reset these bits
			& ~DMD_M_FLAGS,

    DMD_M_NEED_REF	= DMD_CNPT,	// needs a reference KCL

    DMD_M_DEFAULT	= DMD_M_ALL	// default setting
			& ~DMD_M_PFLAGS

} draw_mode_t;

///////////////////////////////////////////////////////////////////////////////

extern uint KMP_DRAW;

int ScanOptDraw ( ccp arg );
ccp GetDrawMode();

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// [[DrawRouteMode_t]]

typedef enum DrawRouteMode_t
{
    // order by priority, lowest priority first

    DRTM_OFF,			// draw none
    DRTM_POTI,			// draw standard route
  //DRTM_CAME,			// draw came route
    DRTM_GOBJ,			// draw route for GOBJ objects
    DRTM_ITEMBOX,		// draw route for standard itembox
    DRTM_ITEMBOX_ENEMY,		// draw route for itembox with special item for enemies
    DRTM_ITEMBOX_PLAYER,	// draw route for itembox with special item for players
}
__attribute__ ((packed)) DrawRouteMode_t;

///////////////////////////////////////////////////////////////////////////////
// [[DrawKCL_t]]

#define MAX_ROUTE 0x100

typedef struct DrawKCL_t
{
    struct kcl_t	*draw;		// valid pointer to draw-KCL, maybe same as 'kcl'

    const struct kcl_t	*kcl;		// NULL or pointer to related source KCL
    const struct kmp_t	*kmp;		// NULL or pointer to related source KMP

    DrawRouteMode_t	route[MAX_ROUTE]; // List of routes to draw
    bool		warnings;	// true: draw warnings in red
    bool		detailed;	// true: print more detailed
    bool		as_road;	// true: print as road (KCL flag 0x0000)
}
DrawKCL_t;

//-----------------------------------------------------------------------------

static inline void InitializeDrawKCL ( DrawKCL_t *dk )
	{ DASSERT(dk); memset(dk,0,sizeof(*dk)); }

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP draw functions		///////////////
///////////////////////////////////////////////////////////////////////////////

void AddCheckPoints2KCL
(
    DrawKCL_t		* dk,		// pointer to initalized object
    bool		draw_jgpt	// true: draw connections to JGPT
);

//-----------------------------------------------------------------------------

void AddArrow2KCL
(
    DrawKCL_t		* dk,		// pointer to initalized object
    uint		sect,		// KMP section
    uint		arrow_color,	// color of arrow
    uint		area_color,	// color of area behind arrow
    double		area_len,	// >0.0: width of area behind arrow
    double		area_wd		// >0.0: length of area behind arrow
);

//-----------------------------------------------------------------------------

void AddRoutes2KCL
(
    DrawKCL_t		* dk,		// pointer to initalized object
    uint		sect,		// KMP section
    uint		color		// color index
);

//-----------------------------------------------------------------------------

void AddPotiRoutes2KCL
(
    DrawKCL_t		* dk		// pointer to initalized object
);

//-----------------------------------------------------------------------------

void AddCannons2KCL
(
    DrawKCL_t		* dk		// pointer to initalized object
);

//-----------------------------------------------------------------------------

void AddObject2KCL
(
    DrawKCL_t		* dk,		// pointer to initalized object
    const
     kmp_gobj_entry_t	* op		// object to draw
);

//-----------------------------------------------------------------------------

void AddObjectByFlag2KCL
(
    DrawKCL_t		* dk,		// pointer to initalized object
    u32			flag_mask,	// mask the object flag with this ...
    u32			flag_result	// ... and draw it, if the result is this.
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			pos file support		///////////////
///////////////////////////////////////////////////////////////////////////////
// [[pos_param_t]] [[pos_file_t]]

#define MAX_POS_FILE_INDEX	50
#define MAX_POS_PARAM		15
#define MAX_POS_FILTER		 3

typedef struct pos_param_t
{
    //--- columns, all column indeces are 1 based

    uint		idx_x;		// >0: column index of x-pos 
    uint		idx_y;		// >0: column index of y-pos 
    uint		idx_z;		// >0: column index of z-pos 
    uint		idx_d;		// >0: column index of directiion (degree)

    //--- color

    uint		autocolor;	// >0: column for auto color, ignore color
    uint		color;		// kcl_flag_*
    int			transp;		// 0,1,2 for transparency selection, -1:disabled

    //--- filter

    uint  idx_f[MAX_POS_FILTER];	// column index of filter A
    char  filter[MAX_POS_FILTER][20];	// text of filter A
}
pos_param_t;

typedef struct pos_file_t
{
    struct pos_file_t	*next;		// link to next entry
    ccp			fname;		// filename (alloced). NULL if invalid

    uint		n_param;	// number of used 'param' elements
    pos_param_t		param[MAX_POS_PARAM];
}
pos_file_t;

//-----------------------------------------------------------------------------

extern pos_param_t default_pos_param;
extern pos_file_t  current_pos_param;
extern pos_file_t *first_pos_file;

//-----------------------------------------------------------------------------

int ScanOptPosFile ( ccp arg );
int ScanOptPosMode ( ccp arg );

void DrawPosFiles( DrawKCL_t *draw );
enumError DrawPosFile ( DrawKCL_t *draw, pos_file_t *pf );

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  KMP misc			///////////////
///////////////////////////////////////////////////////////////////////////////

extern uint kmp_start_pos_mirror;   // last used setting
extern uint kmp_start_pos_narrow;   // last used setting
extern bool kmp_start_pos_auto;	    // last call used 'automode';

const double3 * GetStartPosKMP
(
    // Return NULL on failure.
    // Otherwise return a pointer to 'tab' or to an internal static buffer
    // with N elements. This buffer is overwritten on next call.

    double3	*tab,		// table with 'n' elements to store the result.
				//   if NULL: use internal static buffer
    uint	n,		// number of players: 1..12
    bool	mirror,		// true: mirror the data (pole is right)
    bool	narrow,		// true: use 'narrow' mode

    kmp_t	*kmp,		// not NULL: Use KTPT/#0 and STGI/#0
    uint	automode	// bit fieled: Use KMP to override parameters
				//  1: return absolute positions, if KMP available
				//  2: use pole position of KMP if available
				//  4: use narrow mode of KMP if available
);

///////////////////////////////////////////////////////////////////////////////

int CheckKMP
(
    // returns number of errors

    const kmp_t		* kmp,		// pointer to valid KMP
    CheckMode_t		mode		// print mode
);

///////////////////////////////////////////////////////////////////////////////

extern int have_kmp_patch_count;

int ScanOptRmGobj ( ccp arg );
enumError cmd_stgi();

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    END				///////////////
///////////////////////////////////////////////////////////////////////////////

#endif // SZS_LIB_KMP_H
