
/***************************************************************************
 *                         _______ _______ _______                         *
 *                        |  ___  |____   |  ___  |                        *
 *                        | |   |_|    / /| |   |_|                        *
 *                        | |_____    / / | |_____                         *
 *                        |_____  |  / /  |_____  |                        *
 *                         _    | | / /    _    | |                        *
 *                        | |___| |/ /____| |___| |                        *
 *                        |_______|_______|_______|                        *
 *                                                                         *
 *                             Wiimms SZS Tools                            *
 *                           http://szs.wiimm.de/                          *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This file is part of the SZS project.                                 *
 *   Visit http://szs.wiimm.de/ for project details and sources.           *
 *                                                                         *
 *   Copyright (c) 2011-2016 by Dirk Clemens <wiimm@wiimm.de>              *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   See file gpl-2.0.txt or http://www.gnu.org/licenses/gpl-2.0.txt       *
 *                                                                         *
 ***************************************************************************/

#include "lib-kmp.h"
#include "lib-kcl.h"
#include "lib-szs.h"
#include "ui.h"

#include "kmp.inc"

#include <math.h>

//
///////////////////////////////////////////////////////////////////////////////
///////////////			kmp constants			///////////////
///////////////////////////////////////////////////////////////////////////////

enum
{
    //----- auto mode

    KMP_AM_OFF = 0,	// default
    KMP_AM_NORM,	// normalize modes
    KMP_AM_RENUMBER,	// renumber  modes
    KMP_AM_1_LAP,	// define modes as 1 lap race
    KMP_AM_SHORT,	// define modes to get a very short race
    KMP_AM_UNLIMIT,	// remove lap counter for unlimited race

    //----- sort mode

    KMP_SORT_OFF = 0,	// sort is off
    KMP_SORT_GROUPING,	// grouping only
    KMP_SORT_ANGLE,	// grouping and sorting by angle
};

//----- default values for ROUTE-OBJECT

#define RTOBJ_1		0x12f	// o$castletree1
#define RTOBJ_2		0x163	// o$TownTreeDS
#define RTOBJ_BASE	-2000
#define RTOBJ_SCALE_X	 0.05
#define RTOBJ_SCALE_Y	12.00
#define RTOBJ_SCALE_Z	 0.20
#define RTOBJ_SCALE_MIN	 0.01

enum // bit numbers for 'show route'
{
    SHRT_ENPT = 0,
    SHRT_ITPT,
    SHRT_JGPT,
    SHRT_CNPT,
    SHRT_MSPT,
    SHRT_POTI,
};

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    KMP action log		///////////////
///////////////////////////////////////////////////////////////////////////////

bool KMP_ACTION_LOG ( bool is_patch, const char * format, ... )
{
    #if HAVE_PRINT
    {
	char buf[200];
	snprintf(buf,sizeof(buf),">>>[KMP]<<< %s",format);
	va_list arg;
	va_start(arg,format);
	PRINT_ARG_FUNC(buf,arg);
	va_end(arg);
    }
    #endif

    if ( verbose > 2 || KMP_MODE & KMPMD_LOG || is_patch && ~KMP_MODE & KMPMD_SILENT )
    {
	fflush(stdout);
	fprintf(stdlog,"    %s>[KMP]%s ",
		colset->heading,
		is_patch ? colset->highlight : colset->info );
	va_list arg;
	va_start(arg,format);
	vfprintf(stdlog,format,arg);
	va_end(arg);
	fputs(colset->reset,stdlog);
	fflush(stdlog);
	return true;
    }

    return false;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			 kmp_mode_t			///////////////
///////////////////////////////////////////////////////////////////////////////

kmp_mode_t KMP_MODE = KMPMD_M_DEFAULT;

///////////////////////////////////////////////////////////////////////////////

static const KeywordTab_t opt_kmp_tab[] =
{
  { 0,			"CLEAR",	"RESET",	KMPMD_M_ALL | KMPMD_F_HIDE },
  { KMPMD_M_DEFAULT,	"DEFAULT",	0,		KMPMD_M_ALL | KMPMD_F_HIDE },

  { KMPMD_FORCE,	"FORCE",	0,		0 },
  { KMPMD_NEW,		"NEW",		0,		0 },
  { KMPMD_RM_SPCITEM,	"RM-SPCITEM",	"RMSPCITEM",	0 },
  { KMPMD_REVERSE,	"REVERSE",	0,		0 },

  { KMPMD_POLE_LEFT,	"LEFT",		"LT",		KMPMD_M_POLE },
  { KMPMD_POLE_RIGHT,	"RIGHT",	"RT",		KMPMD_M_POLE },
  { KMPMD_SP_WIDE,	"WIDE",		0,		KMPMD_M_STARTPOS },
  { KMPMD_SP_NARROW,	"NARROW",	0,		KMPMD_M_STARTPOS },

  { KMPMD_1_LAP,	"1LAPS",	"L1",		KMPMD_M_LAPS },
  { KMPMD_2_LAPS,	"2LAPS",	"L2",		KMPMD_M_LAPS },
  { KMPMD_3_LAPS,	"3LAPS",	"L3",		KMPMD_M_LAPS },
  { KMPMD_4_LAPS,	"4LAPS",	"L4",		KMPMD_M_LAPS },
  { KMPMD_5_LAPS,	"5LAPS",	"L5",		KMPMD_M_LAPS },
  { KMPMD_6_LAPS,	"6LAPS",	"L6",		KMPMD_M_LAPS },
  { KMPMD_7_LAPS,	"7LAPS",	"L7",		KMPMD_M_LAPS },
  { KMPMD_8_LAPS,	"8LAPS",	"L8",		KMPMD_M_LAPS },
  { KMPMD_9_LAPS,	"9LAPS",	"L9",		KMPMD_M_LAPS },

  { KMPMD_FIX_CKPH,	"FIX-CKPH",	"FIXCKPH",	0 },
  { KMPMD_FIX_ENPH,	"FIX-ENPH",	"FIXENPH",	0 },
  { KMPMD_FIX_ITPH,	"FIX-ITPH",	"FIXITPH",	0 },
  { KMPMD_FIX_CKNEXT,	"FIX-CKNEXT",	"FIXCKNEXT",	0 },
  { KMPMD_FIX_CKJGPT,	"FIX-CKJGPT",	"FIXCKJGPT",	0 },

  { KMPMD_M_FIX_PH,	"FIX-PH",	"FIXPH",	KMPMD_M_FIX_PH  | KMPMD_F_HIDE },
  { KMPMD_M_FIX_CK,	"FIX-CK",	"FIXCK",	KMPMD_M_FIX_CK  | KMPMD_F_HIDE },
  { KMPMD_M_FIX_ALL,	"FIX-ALL",	"FIXALL",	KMPMD_M_FIX_ALL | KMPMD_F_HIDE },

  { KMPMD_LOG,		"LOG",		0,		0 },
  { KMPMD_SILENT,	"SILENT",	0,		0 },
  { KMPMD_INPLACE,	"INPLACE",	0,		0 },

  { KMPMD_TEST0,	"TEST0",	"T0",		KMPMD_M_TEST | KMPMD_F_HIDE },
  { KMPMD_TEST1,	"TEST1",	"T1",		KMPMD_M_TEST },
  { KMPMD_TEST2,	"TEST2",	"T2",		KMPMD_M_TEST },
  { KMPMD_TEST3,	"TEST3",	"T3",		KMPMD_M_TEST },

  { 0,0,0,0 }
};

///////////////////////////////////////////////////////////////////////////////

static void SetKmpMode ( uint new_mode )
{
    noPRINT("SET-KMP/IN:  PC=%d, KPC=%d\n",
		have_patch_count, have_kmp_patch_count );

    if ( KMP_MODE & KMPMD_M_PATCH )
	have_patch_count--, have_kmp_patch_count--;

    KMP_MODE = new_mode & KMPMD_M_ALL;
    if ( PATCH_FILE_MODE & PFILE_F_LOG_ALL )
	KMP_MODE |= KMPMD_LOG;
    force_kmp = ( KMP_TFORM & KMPMD_FORCE ) != 0;

    if ( KMP_MODE & KMPMD_M_PATCH )
	have_patch_count++, have_kmp_patch_count++;

    noPRINT("SET-KMP/OUT: PC=%d, KPC=%d, force=%d\n",
		have_patch_count, have_kmp_patch_count, force_kmp );
}

//-----------------------------------------------------------------------------

int ScanOptKmp ( ccp arg )
{
    if (!arg)
	return 0;

    s64 stat = ScanKeywordList(arg,opt_kmp_tab,0,true,0,KMP_MODE,
					"Option --kmp",ERR_SYNTAX);
    if ( stat != -1 )
    {
	SetKmpMode(stat);
	return 0;
    }
    return 1;
}

///////////////////////////////////////////////////////////////////////////////

ccp GetKmpMode()
{
    static char buf[200] = {0};
    if (!*buf)
	PrintKeywordList( buf, sizeof(buf), 0, opt_kmp_tab,
				KMP_MODE, KMPMD_M_DEFAULT, KMPMD_F_HIDE );
    return buf;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			 kmp_tform_t			///////////////
///////////////////////////////////////////////////////////////////////////////

kmp_tform_t KMP_TFORM = KMPTF_M_DEFAULT;

///////////////////////////////////////////////////////////////////////////////

static const KeywordTab_t opt_tformkmp_tab[] =
{
  { 0,			"NONE",		"CLEAR",	KMPTF_M_ALL | KMPTF_F_HIDE },
  { KMPTF_M_DEFAULT,	"DEFAULT",	0,		KMPTF_M_ALL | KMPTF_F_HIDE },
  { KMPTF_M_ALL,	"ALL",		0,		KMPTF_M_ALL },

  { KMPTF_AREA,		"AREA",		0,		KMPTF_AREA },
  { 0,			"-AREA",	0,		KMPTF_AREA | KMPTF_F_HIDE },
  { KMPTF_AREA_POS,	"AREA-POS",	"AREAPOS",	0 },
  { KMPTF_AREA_ROTATE,	"AREA-ROTATE",	"AREAROTATE",	0 },
  { KMPTF_AREA_SCALE,	"AREA-SCALE",	"AREASCALE",	0 },

  { KMPTF_CAME,		"CAME",		0,		0 },
  { KMPTF_CAME_POS,	"CAME-POS",	"CAMEPOS",	0 },

  { KMPTF_CKPT,		"CKPT",		0,		0 },
  { KMPTF_CKPT_POS,	"CKPT-POS",	"CKPTPOS",	0 },

  { KMPTF_CNPT,		"CNPT",		0,		KMPTF_CNPT },
  { 0,			"-CNPT",	0,		KMPTF_CNPT | KMPTF_F_HIDE },
  { KMPTF_CNPT_POS,	"CNPT-POS",	"CNPTPOS",	0 },
  { KMPTF_CNPT_ROTATE,	"CNPT-ROTATE",	"CNPTROTATE",	0 },

  { KMPTF_ENPT,		"ENPT",		0,		KMPTF_ENPT },
  { 0,			"-ENPT",	0,		KMPTF_ENPT | KMPTF_F_HIDE },
  { KMPTF_ENPT_POS,	"ENPT-POS",	"ENPTPOS",	0 },
  { KMPTF_ENPT_SCALE,	"ENPT-SCALE",	"ENPTSCALE",	0 },

  { KMPTF_GOBJ,		"GOBJ",		0,		KMPTF_GOBJ },
  { 0,			"-GOBJ",	0,		KMPTF_GOBJ | KMPTF_F_HIDE },
  { KMPTF_GOBJ_POS,	"GOBJ-POS",	"GOBJPOS",	0 },
  { KMPTF_GOBJ_ROTATE,	"GOBJ-ROTATE",	"GOBJROTATE",	0 },
  { KMPTF_GOBJ_SCALE,	"GOBJ-SCALE",	"GOBJSCALE",	0 },

  { KMPTF_ITPT,		"ITPT",		0,		KMPTF_ITPT },
  { 0,			"-ITPT",	0,		KMPTF_ITPT | KMPTF_F_HIDE },
  { KMPTF_ITPT_POS,	"ITPT-POS",	"ITPTPOS",	0 },
  { KMPTF_ITPT_SCALE,	"ITPT-SCALE",	"ITPTSCALE",	0 },

  { KMPTF_JGPT,		"JGPT",		0,		KMPTF_JGPT },
  { 0,			"-JGPT",	0,		KMPTF_JGPT | KMPTF_F_HIDE },
  { KMPTF_JGPT_POS,	"JGPT-POS",	"JGPTPOS",	0 },
  { KMPTF_JGPT_ROTATE,	"JGPT-ROTATE",	"JGPTROTATE",	0 },

  { KMPTF_KTPT,		"KTPT",		0,		KMPTF_KTPT },
  { 0,			"-KTPT",	0,		KMPTF_KTPT | KMPTF_F_HIDE },
  { KMPTF_KTPT_POS,	"KTPT-POS",	"KTPTPOS",	0 },
  { KMPTF_KTPT_ROTATE,	"KTPT-ROTATE",	"KTPTROTATE",	0 },

  { KMPTF_MSPT,		"MSPT",		0,		KMPTF_MSPT },
  { 0,			"-MSPT",	0,		KMPTF_MSPT | KMPTF_F_HIDE },
  { KMPTF_MSPT_POS,	"MSPT-POS",	"MSPTPOS",	0 },
  { KMPTF_MSPT_ROTATE,	"MSPT-ROTATE",	"MSPTROTATE",	0 },

  { KMPTF_POTI,		"POTI",		0,		0 },
  { KMPTF_POTI_POS,	"POTI-POS",	"POTIPOS",	0 },

  { 0,0,0,0 }
};

///////////////////////////////////////////////////////////////////////////////

int ScanOptTformKmp ( ccp arg )
{
    if (!arg)
	return 0;

    s64 start = 0;
    if ( *arg == '+' )
	start = KMP_TFORM;
    else if ( *arg == '-' )
	start = KMPTF_M_ALL;

    s64 stat = ScanKeywordList(arg,opt_tformkmp_tab,0,true,0,start,
					"Option --tform-kmp",ERR_SYNTAX);
    if ( stat != -1 )
    {
	noPRINT("TFORM: %x -> %llx\n",KMP_TFORM,stat & KMPTF_M_ALL);
	KMP_TFORM = stat & KMPTF_M_ALL;
	return 0;
    }
    return 1;
}

///////////////////////////////////////////////////////////////////////////////

uint PrintKmpTform ( char *buf, uint bufsize, kmp_tform_t mode )
{
    DASSERT(buf);
    DASSERT(bufsize>10);
    char *dest = buf;
    char *end = buf + bufsize - 1;

    mode = mode & KMPTF_M_ALL | KMPTF_F_HIDE;
    kmp_mode_t mode1 = mode;

    const KeywordTab_t *ct;
    for ( ct = opt_tformkmp_tab; ct->name1 && dest < end; ct++ )
    {
	if ( ct->opt & KMPMD_F_HIDE )
	    continue;

	if ( ct->opt ? (mode & ct->opt) == ct->id : mode & ct->id )
	{
	    if ( dest > buf )
		*dest++ = ',';
	    dest = StringCopyE(dest,end,ct->name1);
	    mode &= ~(ct->id|ct->opt);
	}
    }

    if ( mode1 == (KMPTF_M_DEFAULT|KMPTF_F_HIDE) )
	dest = StringCopyE(dest,end," (default)");
    else if (!mode1)
	dest = StringCopyE(dest,end,"(none)");

    *dest = 0;
    return dest-buf;
}

///////////////////////////////////////////////////////////////////////////////

ccp GetKmpTform()
{
    static char buf[200] = {0};
    if (!*buf)
	PrintKmpTform(buf,sizeof(buf),KMP_TFORM);
    return buf;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			speed modifier			///////////////
///////////////////////////////////////////////////////////////////////////////

bool  speed_mod_active = false;
float speed_mod_factor = 0.0;
u16   speed_mod_val = 0;

///////////////////////////////////////////////////////////////////////////////

u16 SpeedMod2u16 ( float speed )
{
    u8 val[4];
    write_bef4(val,speed);
    if ( val[2] >= 0x80 && !++val[1] )
	val[0]++;
    return be16(val);
}

///////////////////////////////////////////////////////////////////////////////

float SpeedMod2float ( u16 hex )
{
    u16 val[2] = { htons(hex), 0 };
    return bef4(val);
}

///////////////////////////////////////////////////////////////////////////////

int ScanOptSpeedMod ( ccp arg )
{
    if (!arg)
	return 0;

    ScanInfo_t si;
    InitializeSI(&si,arg,strlen(arg),"Option --speed-mod",0);

    Var_t temp;
    enumError err = ScanExprSI(&si,&temp);
    float factor = GetDoubleV(&temp);
    CheckEolSI(&si);
    ResetSI(&si);

    if (speed_mod_active)
	have_patch_count--, have_kmp_patch_count--;

    if ( factor < 0.0 )
	speed_mod_active = false;
    else
    {
	speed_mod_active = true;
	speed_mod_val    = SpeedMod2u16(factor);
	speed_mod_factor = SpeedMod2float(speed_mod_val);
	have_patch_count++;
	have_kmp_patch_count++;
    }

    return err != ERR_OK;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			helper functions		///////////////
///////////////////////////////////////////////////////////////////////////////

int GetNextPointKMP
(
    const kmp_t		* kmp,		// KMP data structure
    uint		sect,		// KMP PT section index
    int			pt_index,	// index of point, <0: add N
    int			sub_index,	// index into 'next'
					//   -1: use 'pt_index+1' in a group
					//       find first valid at group end
					//  !-1 && KMP_CKPT: try to use 'next' link
    bool		unique,		// return 'not_found_val' on duplicates
    int			not_found_val,	// return this value, if next point not found
    int			*ret_pt_index,	// not NULL: store fixed 'pt_index' here
    int			*ret_sub_index,	// not NULL: store used sub index here
    bool		*ret_new_group	// not NULL: store true, if new group
)
{
    DASSERT(kmp);
    DASSERT( sect == KMP_CKPT || sect == KMP_ENPT || sect == KMP_ITPT );

    if (ret_sub_index)
	*ret_sub_index = -1;
    if (ret_new_group)
	*ret_new_group = false;

    const List_t *ptlist, *phlist;
    if ( sect == KMP_CKPT )
    {
	ptlist = kmp->dlist + KMP_CKPT;
	phlist = kmp->dlist + KMP_CKPH;
    }
    else if ( sect == KMP_ENPT )
    {
	ptlist = kmp->dlist + KMP_ENPT;
	phlist = kmp->dlist + KMP_ENPH;
    }
    else
    {
	DASSERT( sect == KMP_ITPT );
	ptlist = kmp->dlist + KMP_ITPT;
	phlist = kmp->dlist + KMP_ITPH;
    }

    if ( pt_index < 0 )
	pt_index += ptlist->used;
    if ( pt_index < 0 || pt_index >= ptlist->used )
    {
	if (ret_pt_index)
	    *ret_pt_index = -1;
	return not_found_val;
    }

    if (ret_pt_index)
	*ret_pt_index = pt_index;

    const kmp_enph_entry_t *ph = (kmp_enph_entry_t*)phlist->list;
    uint n = phlist->used;
    while ( pt_index >= ph->pt_start + ph->pt_len && n-- > 0 )
	ph++;

    noPRINT(">> %u %zu[%u+%u]\n", idx,
		ph-(kmp_enph_entry_t*)phlist->list,ph->pt_start,ph->pt_len);

    if ( pt_index == ph->pt_start + ph->pt_len - 1 )
    {
	// end of group reached
	if (ret_new_group)
	    *ret_new_group = true;

	pt_index = -1; // invalidate
	uint ph_idx = ~0;

	if ( sub_index >= 0 && sub_index < KMP_MAX_PH_LINK )
	{
	    ph_idx = ph->next[sub_index];
	    if (unique)
	    {
		uint i;
		for ( i = 0; i < sub_index; i++ )
		    if ( ph_idx == ph->next[i] )
		    {
			ph_idx = ~0;
			break;
		    }
	    }
	}
	else
	{
	    uint idx2;
	    for ( idx2 = 0; idx2 < KMP_MAX_PH_LINK; idx2++ )
	    {
		if (!IS_M1(ph->next[idx2]))
		{
		    ph_idx = ph->next[idx2];
		    if (ret_sub_index)
			*ret_sub_index = idx2;
		    break;
		}
	    }
	}

	if ( ph_idx < phlist->used )
	{
	    ph = (kmp_enph_entry_t*)phlist->list + ph_idx;
	    pt_index = ph->pt_start;
	}
    }
    else if ( unique && sub_index >= 1 && sub_index < KMP_MAX_PH_LINK )
	pt_index = -1;
    else if ( sub_index != -1 && sect == KMP_CKPT )
    {
	const kmp_ckpt_entry_t *pt = (kmp_ckpt_entry_t*)ptlist->list + pt_index;
	uint next = pt->next;
	pt_index = next < ptlist->used ? next : pt_index + 1;
    }
    else
	pt_index++;

    return pt_index >= 0 && pt_index < ptlist->used ? pt_index : not_found_val;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

u16 AnalyseItembox ( Itembox_t *ibox, const kmp_gobj_entry_t *op )
{
    // initialze 'ibox' and returns 'ibox->obj_id'

    DASSERT(ibox);
    DASSERT(op);

    memset(ibox,0,sizeof(*ibox));
    if (op)
      switch (op->obj_id)
      {
	case 0x76: // s_itembox
	case 0xd4: // w_itembox
	case 0xd5: // w_itemboxline
	case 0xee: // sin_itembox
	    ibox->unsure_player	= true;
	    ibox->unsure_enemy	= true;
	    // fall through

	case 0x65: // itembox
	case 0xc9: // f_itembox
	    ibox->obj_id	= op->obj_id;
	    ibox->obj_flags	= ObjectInfo[op->obj_id].flags;
	    ibox->item_player	= op->setting[1];
	    ibox->item_enemy	= op->setting[2];
	    break;
      }

    return ibox->obj_id;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  kmp_group_t			///////////////
///////////////////////////////////////////////////////////////////////////////

static void InitializeGL ( kmp_group_list_t * gl, int selector )
{
    DASSERT(gl);
    memset(gl,0,sizeof(*gl));
    gl->selector = toupper(selector);
    InitializeLine(&gl->line);
}

///////////////////////////////////////////////////////////////////////////////

static enumError AddGroupGL
(
    kmp_group_list_t	* gl,		// group list
    ScanInfo_t		* si,		// valid scan info
    char		var_prefix,	// Prefix for var names
    uint		index,		// index of first point
    uint		n_dim		// Line_t support: number of dimensions
)
{
    DASSERT(gl);
    DASSERT(si);
    ScanFile_t *sf = si->cur_file;
    DASSERT(sf);

    char name[KMP_MAX_NAME_LEN+3];
    ScanNameSI(si,name,sizeof(name),true,true,0);

    const bool split_group = !strcmp(name,"$SPLIT");
    if (split_group)
    {
	if ( !gl->used || index == gl->group[gl->used-1].first_index )
	{
	    si->total_err++;
	    sf->line_err++;
	    CheckEolSI(si);
	    if (!si->no_warn)
		ERROR0(ERR_WARNING,
		    "Command '$SPLIT' ignored [%s @%u].\n", sf->name, sf->line );
	    return ERR_WARNING;
	}
    }
    else if ( strcmp(name,"$GROUP") )
    {
	if (!strcmp(name,"$LINE"))
	    return ScanLine(&gl->line,si,n_dim,LM_LINE);
	if (!strcmp(name,"$BORDER"))
	    return ScanLine(&gl->line,si,n_dim,LM_BORDER);
	if (!strcmp(name,"$BEZIER"))
	    return ScanLine(&gl->line,si,n_dim,LM_BEZIER);

	si->total_err++;
	sf->line_err++;
	CheckEolSI(si);
	return ERROR0(ERR_WARNING,
		"Line ignored [%s @%u]: %s ...\n",
		sf->name, sf->line, name );
    }

    if ( gl->used >= KMP_MAX_GROUP )
    {
	si->total_err++;
	sf->line_err++;
	CheckEolSI(si);
	return ERROR0(ERR_WARNING,
		"To many groups, new group ignored: %s @%u\n",
		sf->name, sf->line );
    }

    kmp_group_t *grp = gl->group + gl->used++;
    grp->first_index = grp->last_index = index;

    if ( NextCharSI(si,false) == '-' )
    {
	sf->ptr++;
	grp->no_rotate = true;
    }

    ScanNameSI(si,name+2,sizeof(name)-2,true,true,0);
    memcpy(grp->name,name+2,sizeof(grp->name));
    name[0] = var_prefix;
    name[1] = '.';
    DefineIntVar(&si->gvar,name,gl->used-1);

    if (split_group)
    {	
	kmp_group_t *prev = gl->group + gl->used - 2;
	DASSERT( prev >= gl->group );
	memcpy(grp->link,prev->link,sizeof(grp->link));
	memset(prev->link,0,sizeof(prev->link));
	memcpy(prev->link[0],name,sizeof(prev->link[0]));

	CheckEolSI(si);
	return ERR_OK;
    }

    for(;;)
    {
	char ch = NextCharSI(si,false);
	noPRINT("--> %02x : %.6s\n",ch,sf->ptr);
	if (!ch)
	    break;
	sf->ptr++;
	if ( ch == ':' )
	    break;
    }

    uint idx = 0;
    while ( idx < KMP_MAX_PH_LINK )
    {
	ScanNameSI(si,name,sizeof(name),true,true,0);
	if ( gl->selector && *sf->ptr == '!' )
	{
	    int ch = *++sf->ptr;
	    if (ch)
	    {
		sf->ptr++;
		if ( toupper(ch) != gl->selector )
		    continue;
	    }
	}

	uint n = 1;
	if ( *sf->ptr == '*' )
	{
	    sf->ptr++;
	    const uint num = *sf->ptr - '0';
	    if ( num < 10 )
	    {
		sf->ptr++;
		if ( gl->selector == 'E' )
		    n = num;
	    }
	}

	noPRINT_IF(*name,"LINK#%zu.%u: %s *%u\n",grp-gl->group,idx,name,n);
	while ( n-- > 0 && idx < KMP_MAX_PH_LINK )
	{
	    memcpy(grp->link[idx],name,sizeof(grp->link[idx]));
	    idx++;
	}
    }

    CheckEolSI(si);
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static uint AddPointGL
(
    kmp_group_list_t	* gl,		// group list
    uint		index		// index if first point
)
{
    if (!gl->used)
	gl->used++;

    kmp_group_t *grp = gl->group + gl->used - 1;
    grp->last_index = index;
    return index - grp->first_index;
}

///////////////////////////////////////////////////////////////////////////////

static void InsertLink
(
    u8			* vector,	// insert into vector
    u8			link,		// link to insert
    ScanInfo_t		* si,		// not NULL: print error message
    ccp			name,		// NULL or name of line for error message
    bool		allow_dup	// true: allow multiple links into same section
)
{
    DASSERT(vector);

    uint i;
    for ( i = 0; i < KMP_MAX_PH_LINK; i++ )
    {
	if ( vector[i] == 0xff )
	{
	    vector[i] = link;
	    return;
	}
	if ( !allow_dup && vector[i] == link )
	    return;
    }

    ScanFile_t *sf = si->cur_file;
    DASSERT(sf);
    ERROR0(ERR_WARNING,
		"To many links [%s @%u]: %s\n",
		sf->name, sf->line, name ? name : "?" );
}

///////////////////////////////////////////////////////////////////////////////

static void TermGL
(
    kmp_group_list_t	* gl,		// group list
    ScanInfo_t		* si,		// not NULL: print error message
    kmp_t		* kmp,		// KMP data structure
    uint		sect,		// PH section
    kmp_group_name_t	* gname,	// NULL or pointer to group name array
    int			rotate_links	// >0: rotate links um x positions
)
{
    DASSERT(gl);
    DASSERT(kmp);
    DASSERT( sect < KMP_N_SECT );

    CheckLevelSI(si);

    if (gname)
	memset(gname,0,sizeof(*gname)*KMP_MAX_GROUP);

    List_t *list = kmp->dlist + sect;
    list->used = 0;

    uint i;
    kmp_group_t *grp = gl->group;
    for ( i = 0; i < gl->used; i++, grp++ )
    {
	kmp_enph_entry_t * entry = AppendList(list);
	entry->pt_start = grp->first_index;
	entry->pt_len = grp->last_index - grp->first_index + 1;

	memset(entry->next,-1,sizeof(entry->next));
	memset(entry->prev,-1,sizeof(entry->prev));
	if ( gname && i < KMP_MAX_GROUP )
	    memcpy(gname+i,grp->name,sizeof(*gname));
    }

    rotate_links = rotate_links <= 0 ? 0 : rotate_links % KMP_MAX_PH_LINK;
    PRINT_IF(rotate_links,"ROTATE-LINKS %d\n",rotate_links);

    grp = gl->group;
    for ( i = 0; i < gl->used; i++, grp++ )
    {
	uint l;
	for ( l = 0; l < KMP_MAX_PH_LINK; l++ )
	{
	    uint j = grp->no_rotate
			? l
			: ( l + rotate_links ) % KMP_MAX_PH_LINK;
	    ccp lname = grp->link[j];
	    if (!*lname)
		continue;

	    noPRINT(">> '%s' links to '%s'\n",grp->name,lname);
	    kmp_group_t *grp2 = gl->group;
	    for ( j = 0; j < gl->used; j++, grp2++ )
	    {
		if (!strcmp(lname,grp2->name))
		{
		    kmp_enph_entry_t * entry;
		    entry = (kmp_enph_entry_t*)(list->list) + i;
		    InsertLink(entry->next,j,si,lname,true);
		    entry = (kmp_enph_entry_t*)(list->list) + j;
		    InsertLink(entry->prev,i,si,lname,false);
		    break;
		}
	    }
	    // [[warn]] not found warning
	}
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			RenameGroup()			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct group_t
{
    u8  done;
    u8  index;
    u8  subindex;
    u8	order;
}
group_t;

///////////////////////////////////////////////////////////////////////////////

static u64 IterateGroup
(
    const kmp_enph_entry_t	* elist,
    group_t			* grp,
    uint			idx,
    uint			n,
    uint			last_index,
    u64				steps
)
{
    DASSERT(elist);
    DASSERT(grp);
    DASSERT(idx<n);
    noPRINT("IterateGroup(%p,%p,%u,%u,%u,%u)\n",elist,grp,idx,n,last_index,steps);

    if ( steps < 1000000 ) // limit steps avoid nearly endless loops
    {
	noPRINT_IF( !(steps % 10000000), "  steps=%9llu M\n", steps/1000000);
	steps++;
	last_index++;
	const kmp_enph_entry_t * entry = elist + idx;

	uint ni;
	for ( ni=0; ni < KMP_MAX_PH_LINK; ni++ )
	{
	    const uint next = entry->next[ni];
	    if ( next < n )
	    {
		group_t * g = grp + next;
		if ( !g->done && g->index < last_index )
		{
		    g->index = last_index;
		    g->done++;
		    noPRINT(" -> ni=%u, next=%u/%u, last=%u\n",ni,next,n,last_index);
		    steps = IterateGroup(elist,grp,next,n,last_index,steps);
		    g->done--;
		}
	    }
	}
    }
    return steps;
}

///////////////////////////////////////////////////////////////////////////////

static void RenameGroup
(
    kmp_t		* kmp,		// KMP data structure
    uint		sect_ph,	// ph section index
    kmp_group_name_t	* gname		// pointer to group name array
)
{
    DASSERT(kmp);
    DASSERT( sect_ph == KMP_CKPH || sect_ph == KMP_ENPH || sect_ph == KMP_ITPH );
    DASSERT(gname);

    const List_t *phlist = kmp->dlist + sect_ph;
    const uint n = phlist->used < KMP_MAX_GROUP ? phlist->used : KMP_MAX_GROUP;
    memset(gname,0,sizeof(*gname)*KMP_MAX_GROUP);

    uint last_index = 0;
    group_t grp[KMP_MAX_GROUP];
    memset(grp,0,sizeof(grp));

    const kmp_enph_entry_t * elist = (kmp_enph_entry_t*)(phlist->list);


    //--- find index number

    for(;;)
    {
	//--- find first non named group

	uint g1 = 0;
	while ( g1 < n && grp[g1].done )
	    g1++;
	if ( g1 == n )
	    break;

	grp[g1].done = 1;
	grp[g1].index = ++last_index;
 #if HAVE_PRINT0
	const uint steps = IterateGroup(elist,grp,g1,n,last_index,0);
	PRINT(" SECT=%s, steps=%8u\n",kmp_section_name[sect_ph].name1,steps);
 #else
	IterateGroup(elist,grp,g1,n,last_index,0);
 #endif

	uint gi;
	group_t * gp;
	for ( gi = 0, gp = grp; gi < n; gi++, gp++ )
	    if (gp->index)
	    {
		gp->done = 1;
		if ( last_index < gp->index )
		     last_index = gp->index;
	    }
    }


    //--- calc subindex

    u8 count[KMP_MAX_GROUP];
    memset(count,0,sizeof(count));

    uint gi;
    group_t * gp;
    for ( gi = 0, gp = grp; gi < n; gi++, gp++ )
	gp->subindex = count[gp->index]++;


    //--- print names

    const int fw = snprintf(*gname,sizeof(*gname),"%u",last_index);
    for ( gi = 0, gp = grp; gi < n; gi++, gp++, gname++ )
    {
	uint nc = count[gp->index];
	if ( nc <= 1 )
	    snprintf(*gname,sizeof(*gname), "G%0*u", fw,gp->index );
	else if ( nc <= 26 )
	    snprintf(*gname,sizeof(*gname),
			"G%0*u%c", fw,gp->index, 'A' + gp->subindex );
	else
	    snprintf(*gname,sizeof(*gname),
			"G%0*u%c%c", fw,gp->index,
			'A' + gp->subindex/26, 'A' + gp->subindex%26 );
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP pathes			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct kmp_link_t
{
    u8	cur;		// item index of current point
    u8	next;		// item index of next point
    u8	grp_cur;	// group index of current point
    u8	grp_next;	// group index of next point
}
kmp_link_t;

//-----------------------------------------------------------------------------

typedef struct kmp_path_t
{
    uint	sect_ph;			// analysed PH section
    uint	sect_pt;			// analysed PT section

    uint	n_ph;				// number of groups
    uint	n_pt;				// number of items

    group_t	grp[KMP_MAX_GROUP];		// group_t data
    u8		grp_order[KMP_MAX_GROUP];	// order => index into 'grp'
    u8		itm_grp[KMP_MAX_POINT];		// group of each item

    kmp_link_t	*link;				// allocated list of links
    uint	n_link;				// number of element in 'link'
}
kmp_path_t;

///////////////////////////////////////////////////////////////////////////////

static inline void InitializeKmpPath ( kmp_path_t *path )
{
    DASSERT(path);
    memset(path,0,sizeof(*path));
}

///////////////////////////////////////////////////////////////////////////////

static void ResetKmpPath ( kmp_path_t *path )
{
    DASSERT(path);
    if (path->link)
	FREE(path->link);
    InitializeKmpPath(path);
}

///////////////////////////////////////////////////////////////////////////////

static void AnalyzeKmpPath
(
    const kmp_t		*kmp,		// KMP data structure
    uint		sect_ph,	// ph section index
    kmp_path_t		*path,		// store group list here
    bool		init_path,	// true: initialize 'path'
    uint		first_grp	// index of first group (robust)
)
{
 #if HAVE_PRINT
    uint start_time = GetTimerUSec();
 #endif

    DASSERT(kmp);
    DASSERT( sect_ph == KMP_CKPH || sect_ph == KMP_ENPH || sect_ph == KMP_ITPH );
    DASSERT(path);

    if (init_path)
	InitializeKmpPath(path);
    else
	ResetKmpPath(path);

    DASSERT( KMP_CKPT+1 == KMP_CKPH );
    DASSERT( KMP_ENPT+1 == KMP_ENPH );
    DASSERT( KMP_ITPT+1 == KMP_ITPH );
    path->sect_ph = sect_ph;
    path->sect_pt = sect_ph - 1;

    const List_t *phlist = kmp->dlist + sect_ph;
    const uint n_ph = phlist->used < KMP_MAX_GROUP ? phlist->used : KMP_MAX_GROUP;
    const kmp_enph_entry_t * elist = (kmp_enph_entry_t*)(phlist->list);

    const List_t *ptlist = kmp->dlist + path->sect_pt;
    const uint n_pt = ptlist->used < KMP_MAX_POINT ? ptlist->used : KMP_MAX_POINT;

    uint last_index = 0;
    path->n_ph = n_ph;
    path->n_pt = n_pt;

    for (;;)
    {
	if ( first_grp >= n_ph )
	    first_grp = 0;
	PRINT("GRP: ---- %u ---\n",first_grp);

	for(;;)
	{
	    uint g1 = first_grp;
	    while ( g1 < n_ph && path->grp[g1].done )
		g1++;
	    if ( g1 == n_ph )
		break;

	    path->grp[g1].done = 1;
	    path->grp[g1].index = ++last_index;
	 #if HAVE_PRINT0
	    const uint steps = IterateGroup(elist,path->grp,g1,n_ph,last_index,0);
	    PRINT(" SECT=%s, steps=%8u\n",kmp_section_name[sect_ph].name1,steps);
	 #else
	    IterateGroup(elist,path->grp,g1,n_ph,last_index,0);
	 #endif

	    uint gi;
	    group_t * gp;
	    for ( gi = 0, gp = path->grp; gi < n_ph; gi++, gp++ )
	    {
		if (gp->index)
		{
		    gp->done = 1;
		    if ( last_index < gp->index )
			 last_index = gp->index;
		    noPRINT("GRP %3u: %u.%u\n",gi,gp->index,gp->subindex);
		}
	    }
	}

	if (!first_grp)
	    break;
	first_grp = 0;
    }


    //--- calc subindex

    u8 count[KMP_MAX_GROUP];
    memset(count,0,sizeof(count));

    uint gi;
    group_t *gp;
    for ( gi = 0, gp = path->grp; gi < n_ph; gi++, gp++ )
    {
	gp->subindex = count[gp->index]++;
	noPRINT("GRP %3u: %u.%u\n",gi,gp->index,gp->subindex);
    }


    //--- sort groups

    u8 index[KMP_MAX_GROUP];
    memset(index,0,sizeof(index));
    for ( gi = 1; gi < n_ph; gi++ )
	index[gi] = index[gi-1] + count[gi-1];

    for ( gi = 0, gp = path->grp; gi < n_ph; gi++, gp++ )
    {
	gp->done = 0; // clear done for further processing
	gp->order = index[gp->index] + gp->subindex;
	noPRINT("GRP %3u: %3u => %u.%u\n",gi,gp->grp_order,gp->index,gp->subindex);
	path->grp_order[gp->order] = gi;
    }

    //--- setup item lists

    kmp_link_t item[KMP_MAX_PH_LINK*(KMP_MAX_GROUP+2)];
    kmp_link_t *ip = item, *iend = (kmp_link_t*)((u8*)item + sizeof(item)) - KMP_MAX_PH_LINK;
    memset(item,0,sizeof(item));
    memset(path->itm_grp,0xff,sizeof(path->itm_grp));

    uint i;
    for ( i = 0; i < n_ph; i++ )
    {
	uint gi = path->grp_order[i];

	const kmp_enph_entry_t *ep = elist + gi;
	uint idx = ep->pt_start;
	uint end = idx + ep->pt_len;
	if ( end > n_pt )
	     end = n_pt;
	end--;
	for ( ; idx <= end && ip < iend; idx++ )
	{
	    if (path->itm_grp[idx] == 0xff )
	    {
		path->itm_grp[idx] = gi;
		if ( idx == end )
		{
		    // end of group -> multi links
		    uint link;
		    for ( link = 0; link < KMP_MAX_PH_LINK; link++ )
		    {
			int next = GetNextPointKMP(kmp,path->sect_pt,idx,link,true,-1,0,0,0);
			if ( next >= 0 )
			{
			    ip->cur      = idx;
			    ip->next     = next;
			    ip->grp_cur	 = gi;
			    ip++;
			}
		    }
		}
		else
		{
		    ip->cur      = idx;
		    ip->next     = idx + 1;
		    ip->grp_cur	 = gi;
		    ip++;
		}
	    }
	}
    }


    //--- calculate 'next group'

    kmp_link_t *jp;
    for ( jp = item; jp < ip; jp++ )
	jp->grp_next = path->itm_grp[jp->next];


    //--- add points not handled by a group

    for ( i = 0; i < n_pt && ip < iend; i++ )
	if ( path->itm_grp[i] == 0xff )
	{
	    ip->cur      = i;
	    ip->next     = i + 1;
	    ip->grp_cur	 = 0xff;
	    ip->grp_next = 0xff;
	    ip++;
	}

    path->n_link = ip - item;
    path->link = MEMDUP(item,path->n_link*sizeof(*item));


    //--- debugging

    PRINT("AnalyzeKmpPath(%s) done in %llu µsec\n",
		kmp_section_name[sect_ph].name1, GetTimerUSec()-start_time);

    #if HAVE_PRINT0
    {
	PRINT("----- GROUP ORDER '%s' -----\n",kmp_section_name[sect_ph].name1);
	uint i;
	for ( i = 0; i < n_ph; i++ )
	{
	    uint gi = path->grp_order[i];
	    const group_t *gp = path->grp + gi;
	    const kmp_enph_entry_t *ep = elist + gi;

	    PRINT("%3u. GRP %3u: %3u.%u : %3u .. %3u [+%3u]\n",
			gp->order, gi, gp->index, gp->subindex,
			ep->pt_start, ep->pt_start + ep->pt_len-1, ep->pt_len );
	}
    }
    #endif


    #if HAVE_PRINT0
    {
	PRINT("----- ITEM ORDER '%s' -----\n",kmp_section_name[sect_ph].name1);
	uint i;
	for ( i = 0; i < path->n_link; i++ )
	{
	    const kmp_link_t *link = path->link + i;
	    PRINT("%3u. ITM %3u -> %3u : %2u -> %3u\n",
			i, link->cur, link->next, link->grp_cur, link->grp_next );
	}
    }
    #endif
}

///////////////////////////////////////////////////////////////////////////////

const kmp_link_t * FindCKPT 
(
    const kmp_t		*kmp,		// KMP data structure
    kmp_path_t		*ck_path,	// ckeck point path data, already initialized
					//	data is calculated in first call
    float		x,		// x-coordiante of point
    float		z,		// y-coordiante of point
    const kmp_link_t	*start_link	// NULL or start link
)
{
    DASSERT(kmp);
    DASSERT(ck_path);

    const List_t *ptlist = kmp->dlist + KMP_CKPT;
    const uint n_pt = ptlist->used < KMP_MAX_POINT ? ptlist->used : KMP_MAX_POINT;
    if ( n_pt < 2 )
	return 0;

    if ( !ck_path->n_link || ck_path->n_pt != n_pt || ck_path->sect_ph != KMP_CKPH )
    {
	AnalyzeKmpPath(kmp,KMP_CKPH,ck_path,false,0);
	start_link = 0;
    }
    if (!start_link)
	start_link = ck_path->link;
    const kmp_link_t *link, *endlink = ck_path->link + ck_path->n_link;

    float pts[2];
    pts[0] = x;
    pts[1] = z;

    link = start_link;
    const kmp_ckpt_entry_t *pt_base = (kmp_ckpt_entry_t*)ptlist->list;

    for(;;)
    {
	DASSERT( link->cur < n_pt );
	DASSERT( link->next < n_pt );
	float quad[8];

	const kmp_ckpt_entry_t *pt1 = pt_base + link->cur;
	quad[0] = pt1->left[0];
	quad[1] = pt1->left[1];
	quad[2] = pt1->right[0];
	quad[3] = pt1->right[1];

	const kmp_ckpt_entry_t *pt2 = pt_base + link->next;
	quad[4] = pt2->right[0];
	quad[5] = pt2->right[1];
	quad[6] = pt2->left[0];
	quad[7] = pt2->left[1];

	int dir;
	if ( PointsInConvexPolygonF(pts,1,quad,4,true,&dir) && dir >= 0 )
	    return link;

	if ( ++link == endlink )
	    link = ck_path->link;
	if ( link == start_link )
	    return 0;
    }
}

///////////////////////////////////////////////////////////////////////////////

int CheckConvexCKPT
(
    // returns
    //  -1: illegal index
    //   0: convex
    //   1: concave

    const kmp_t		*kmp,		// KMP data structure
    uint		ckpt1,		// index of first check point (robust)
    uint		ckpt2,		// index of second check point (robust)
    double		*min_distance	// not NULL: store minimum distance here
)
{
    DASSERT(kmp);

    int stat = -1;
    double min_dist = INFINITY;

    const List_t *ptlist = kmp->dlist + KMP_CKPT;
    const uint n_pt = ptlist->used;

    if ( ckpt1 < n_pt && ckpt2 < n_pt )
    {
	float p[5][2];

	const kmp_ckpt_entry_t *pt1 = (kmp_ckpt_entry_t*)ptlist->list + ckpt1;
	p[0][0] = p[4][0] = pt1->left[0];
	p[0][1] = p[4][1] = pt1->left[1];
	p[1][0] = pt1->right[0];
	p[1][1] = pt1->right[1];

	const kmp_ckpt_entry_t *pt2 = (kmp_ckpt_entry_t*)ptlist->list + ckpt2;
	p[2][0] = pt2->right[0];
	p[2][1] = pt2->right[1];
	p[3][0] = pt2->left[0];
	p[3][1] = pt2->left[1];

	double dir[5];
	uint j;
	for ( j = 0; j < 4; j++ )
	{
	    noPRINT(" - p: [%8.1f,%8.1f] [%8.1f,%8.1f]\n",
			p[j][0], p[j][1], p[j+1][0], p[j+1][1] );
	    const double dx = p[j+1][0] - p[j][0];
	    const double dz = p[j+1][1] - p[j][1];
	    dir[j] = atan2(dx,dz) * (180.0/M_PI);
	    const double len = dx*dx + dz*dz;
	    if ( min_dist > len )
		 min_dist = len;
	    noPRINT(" - DIR: %6.1f %9.1f [%8.1f,%8.1f]\n",
			dir[j], len, dx, dz );
	}
	dir[4] = dir[0];

	stat = 0;
	for ( j = 0; j < 4; j++ )
	{
	    double delta = dir[j+1] - dir[j];
	    if ( delta < 0.0 )
		delta += 360.0;
	    else if ( delta > 360.0 )
		delta -= 360.0;

	    if ( delta > 180.0 )
	    {
		stat = 1;
		break;
	    }
	}
    }

    if (min_distance)
	*min_distance = sqrt(min_dist);
    return stat;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    KMP data			///////////////
///////////////////////////////////////////////////////////////////////////////

const uint kmp_entry_size[KMP_N_SECT+1] =
{
	sizeof(kmp_ktpt_entry_t),
	sizeof(kmp_enpt_entry_t),
	sizeof(kmp_enph_entry_t),
	sizeof(kmp_itpt_entry_t),
	sizeof(kmp_itph_entry_t),
	sizeof(kmp_ckpt_entry_t),
	sizeof(kmp_ckph_entry_t),
	sizeof(kmp_gobj_entry_t),
	sizeof(kmp_poti_group_t),
	sizeof(kmp_area_entry_t),
	sizeof(kmp_came_entry_t),
	sizeof(kmp_jgpt_entry_t),
	sizeof(kmp_cnpt_entry_t),
	sizeof(kmp_mspt_entry_t),
	sizeof(kmp_stgi_entry_t),
	0
};

const KeywordTab_t kmp_section_name[] =
{
	{ KMP_KTPT, "KTPT",	"Start position", 0 },
	{ KMP_ENPT, "ENPT",	"Enemy route point", 0 },
	{ KMP_ENPH, "ENPH",	"Enemy route group", 0 },
	{ KMP_ITPT, "ITPT",	"Item route point", 0 },
	{ KMP_ITPH, "ITPH",	"Item route group", 0 },
	{ KMP_CKPT, "CKPT",	"Check point", 0 },
	{ KMP_CKPH, "CKPH",	"Check point group", 0 },
	{ KMP_GOBJ, "GOBJ",	"Global object", 0 },
	{ KMP_POTI, "POTI",	"Route", 0 },
	{ KMP_AREA, "AREA",	"Area", 0 },
	{ KMP_CAME, "CAME",	"Camera", 0 },
	{ KMP_JGPT, "JGPT",	"Respawn point", 0 },
	{ KMP_CNPT, "CNPT",	"Canon point", 0 },
	{ KMP_MSPT, "MSPT",	"Battle position", 0 },
	{ KMP_STGI, "STGI",	"Stage info", 0 },
	{ KMP_N_SECT, "SETUP",	0, 1  },
	{ 0,0,0,0 }
};

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    KMP setup			///////////////
///////////////////////////////////////////////////////////////////////////////

void InitializeKMP ( kmp_t * kmp )
{
    static bool done = false;
    if (!done)
    {
	// do this only once
	done = true;
	SetKmpMode(KMP_MODE);
	KMP_ACTION_LOG(false,"Global KMP Modes: %s\n",GetKmpMode());
	noPRINT("Global KMP Modes: %s\n",GetKmpMode());
    }

    DASSERT(kmp);
    memset(kmp,0,sizeof(*kmp));

    kmp->fname = EmptyString;

    uint sect;
    for ( sect = 0; sect < KMP_N_SECT; sect++ )
    {
	InitializeList( kmp->dlist+sect, kmp_entry_size[sect] );
	InitializeIL( kmp->index+sect, 0 );
    }
    InitializeList( &kmp->poti_point, sizeof(kmp_poti_point_t) );

    kmp->index[KMP_GOBJ].prefix = "o";
    kmp->index[KMP_POTI].prefix = "r";
    kmp->index[KMP_AREA].prefix = "a";
    kmp->index[KMP_CAME].prefix = "c";
    kmp->index[KMP_JGPT].prefix = "j";
    kmp->index[KMP_CNPT].prefix = "n";
    kmp->index[KMP_MSPT].prefix = "m";
    kmp->index[KMP_STGI].prefix = "s";

    kmp->revision		= REVISION_NUM;
    kmp->unknown_0x0c		= 0x9d8;

    kmp->rtobj.obj		= GOBJ_OBAKE_BLOCK;
    kmp->rtobj.scale.x		= 0.2;
    kmp->rtobj.scale.y		= 0.4;
    kmp->rtobj.scale.z		= 1.0;

    kmp->ckpt_object1		= RTOBJ_1;
    kmp->ckpt_object2		= RTOBJ_2;
    kmp->ckpt_object1_base	= RTOBJ_BASE;
    kmp->ckpt_object2_base	= RTOBJ_BASE;

    kmp->jgpt_auto_distance	= 1000.0;
    kmp->jgpt_auto_adjust.x	=    0.0;
    kmp->jgpt_auto_adjust.y	=  200.0;
    kmp->jgpt_auto_adjust.z	=    0.0;

    kmp->enable_fall_kcl	= ( KCL_MODE & KCLMD_ADD_ROAD ) != 0;

    kmp->enpt_mask[0] = kmp->enpt_mask[1] = 0xffff;
    kmp->itpt_mask[0] = kmp->itpt_mask[1] = 0xffff;

    memset( kmp->enpt_flags, KPFL_M_DEFAULT & KPFL_M_ENPT, sizeof(kmp->enpt_flags) );
    memset( kmp->itpt_flags, KPFL_M_DEFAULT & KPFL_M_ITPT, sizeof(kmp->itpt_flags) );
    memset( kmp->jgpt_flags, KPFL_M_DEFAULT & KPFL_M_JGPT, sizeof(kmp->jgpt_flags) );
}

///////////////////////////////////////////////////////////////////////////////

void ClearSectionsKMP ( kmp_t * kmp )
{
    DASSERT(kmp);

    uint sect;
    for ( sect = 0; sect < KMP_N_SECT; sect++ )
    {
	ResetList(kmp->dlist+sect);
	ResetIL(kmp->index+sect);
    }
    ResetList(&kmp->poti_point);
}

///////////////////////////////////////////////////////////////////////////////

void ResetKMP ( kmp_t * kmp )
{
    DASSERT(kmp);
    UnloadAutoReferenceKCL();
    FreeString(kmp->fname);
    FreeString(kmp->info);
    ClearSectionsKMP(kmp);
    FREE(kmp->raw_data);

    if (kmp->fall_kcl)
    {
	if ( kcl_ref_prio == kmp->fall_kcl )
	    kcl_ref_prio = 0;
	ResetKCL(kmp->fall_kcl);
	FREE(kmp->fall_kcl);
    }

    InitializeKMP(kmp);
}

///////////////////////////////////////////////////////////////////////////////

static ccp GetObjectName ( uint obj_id )
{
    const uint bufsize = OBJ_FW_NAME + 3;
    char *buf = GetCircBuf(bufsize);

    if ( obj_id < N_KMP_GOBJ && ObjectInfo[obj_id].name )
	snprintf(buf,bufsize,"o$%s",ObjectInfo[obj_id].name);
    else
	snprintf(buf,bufsize,"0x%x",obj_id);
    return buf;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////		    kmp parser function get()		///////////////
///////////////////////////////////////////////////////////////////////////////

enum // get mode
{
    GTMD_N,		// number of elements

    GTMD_HD_U16,	// head 'value' as u16
    GTMD_HD_U8HI,	// hight byte of head 'value'
    GTMD_HD_U8LO,	// low byte of head 'value'

    GTMD__DATA,		// separator of header info and data info

    GTMD_U8,		// unsigned 8 bit
    GTMD_U8M1,		// unsigned 8 bit, but 0xff is -1
    GTMD_U16,		// unsigned 16 bit
    GTMD_U16M1,		// unsigned 16 bit, but 0xffff is -1

    GTMD_DBL,		// double value
    GTMD_DBL2,		// 2d vector
    GTMD_DBL3,		// 3d double vector
};

///////////////////////////////////////////////////////////////////////////////

#define POTI_PT_OFF 0x10

typedef struct kmp_func_t
{
    char	name[18];	// variable name
    u16		sect;		// section
    u16		off;		// data offset, for POTI: off>POTI_PT_OFF: route point
    u16		get_mode;	// see enum above
    u8		n_var;		// number of variables
    u8		n_param;	// number of function parameters
    ccp		get_info;	// NULL or function info for get function
    ccp		set_info;	// NULL or function info for set function

} kmp_func_t;

///////////////////////////////////////////////////////////////////////////////

static const kmp_func_t kmp_func_list[] =
{
    { "KTPT$N",			KMP_KTPT,    0,  GTMD_N,	0,0,
			"Returns the number of start points.",
			0 },
    { "KTPT$POS",		KMP_KTPT, 0x00,  GTMD_DBL3,	1,1,
			"Returns a start point position.",
			0 },
    { "KTPT$ROT",		KMP_KTPT, 0x0c,  GTMD_DBL3,	1,1,
			"Returns a start point rotation.",
			0 },
    { "KTPT$ID",		KMP_KTPT, 0x18,  GTMD_U16M1,	1,1,
			"Returns a start point ID.",
			0 },
    { "KTPT$UNKNOWN",		KMP_KTPT, 0x1a,  GTMD_U16,	1,1,
			"Returns the unknown value of a start point.",
			0 },

    //---------------

    { "ENPT$N",			KMP_ENPT,    0,  GTMD_N,	0,0,
			"Returns the number of enemy points.",
			0 },
    { "ENPT$POS",		KMP_ENPT, 0x00,  GTMD_DBL3,	1,1,
			"Returns the position of an enemy point.",
			0 },
    { "ENPT$SCALE",		KMP_ENPT, 0x0c,  GTMD_DBL,	1,1,
			"Returns the scale of an enemy point.",
			0 },
    { "ENPT$PROP",		KMP_ENPT, 0x10,  GTMD_U16,	2,2,
			"Returns property 0 or 1 of an enemy point.",
			0 },

    //---------------

    { "ENPH$N",			KMP_ENPH,    0,  GTMD_N,	0,0,
			"Returns the number of enemy point groups.",
			0 },
    { "ENPH$START",		KMP_ENPH, 0x00,  GTMD_U8,	1,1,
			"Returns the enemy point start index of an enemy point group.",
			0 },
    { "ENPH$LEN",		KMP_ENPH, 0x01,  GTMD_U8,	1,1,
			"Returns the number of enemy points of an enemy point group.",
			0 },
    { "ENPH$PREV",		KMP_ENPH, 0x02,  GTMD_U8M1,	KMP_MAX_PH_LINK,2,
			"Returns link 0..5 to the previous enemy point group.",
			0 },
    { "ENPH$NEXT",		KMP_ENPH, 0x08,  GTMD_U8M1,	KMP_MAX_PH_LINK,2,
			"Returns link 0..5 to the next enemy point group.",
			0 },
    { "ENPH$UNKNOWN",		KMP_ENPH, 0x0e,  GTMD_U16,	1,1,
			"Returns the unknown value of an enemy point group.",
			0 },

    //---------------

    { "ITPT$N",			KMP_ITPT,    0,  GTMD_N,	0,0,
			"Returns the number of item points.",
			0 },
    { "ITPT$POS",		KMP_ITPT, 0x00,  GTMD_DBL3,	1,1,
			"Returns the position of an item point.",
			"Set the position of an item point"
			" and returns 1 on success ans 0 , if failed." },
    { "ITPT$SCALE",		KMP_ITPT, 0x0c,  GTMD_DBL,	1,1,
			"Returns the scale of an item point.",
			0 },
    { "ITPT$PROP",		KMP_ITPT, 0x10,  GTMD_U16,	2,2,
			"Returns property 0 or 1 of an item point.",
			0 },

    //---------------

    { "ITPH$N",			KMP_ITPH,    0,  GTMD_N,	0,0,
			"Returns the number of item point groups.",
			0 },
    { "ITPH$START",		KMP_ITPH, 0x00,  GTMD_U8,	1,1,
			"Returns the item point start index of an item point group.",
			0 },
    { "ITPH$LEN",		KMP_ITPH, 0x01,  GTMD_U8,	1,1,
			"Returns the number of item points of an item point group.",
			0 },
    { "ITPH$PREV",		KMP_ITPH, 0x02,  GTMD_U8M1,	KMP_MAX_PH_LINK,2,
			"Returns link 0..5 to the previous item point group.",
			0 },
    { "ITPH$NEXT",		KMP_ITPH, 0x08,  GTMD_U8M1,	KMP_MAX_PH_LINK,2,
			"Returns link 0..5 to the next item point group.",
			0 },
    { "ITPH$UNKNOWN",		KMP_ITPH, 0x0e,  GTMD_U16,	1,1,
			"Returns the unknown value of an item point group.",
			0 },

    //---------------

    { "CKPT$N",			KMP_CKPT,    0,  GTMD_N,	0,0,
			"Returns the number of check points.",
			0 },
    { "CKPT$LEFT",		KMP_CKPT, 0x00,  GTMD_DBL2,	1,1,
			"Returns the left position of a check point.",
			0 },
    { "CKPT$RIGHT",		KMP_CKPT, 0x08,  GTMD_DBL2,	1,1,
			"Returns the right position of a check point.",
			0 },
    { "CKPT$RESPAWN",		KMP_CKPT, 0x10,  GTMD_U8,	1,1,
			"Returns the respawn index of a check point.",
			0 },
    { "CKPT$MODE",		KMP_CKPT, 0x11,  GTMD_U8M1,	1,1,
			"Returns the mode of check point.",
			0 },
    { "CKPT$PREV",		KMP_CKPT, 0x12,  GTMD_U8M1,	1,1,
			"Returns the index of the previous check point.",
			0 },
    { "CKPT$NEXT",		KMP_CKPT, 0x13,  GTMD_U8M1,	1,1,
			"Returns the index of the next check point.",
			0 },

    //---------------

    { "CKPH$N",			KMP_CKPH,    0,  GTMD_N,	0,0,
			"Returns the number of check point groups.",
			0 },
    { "CKPH$START",		KMP_CKPH, 0x00,  GTMD_U8,	1,1,
			"Returns the check point start index of a check point group.",
			0 },
    { "CKPH$LEN",		KMP_CKPH, 0x01,  GTMD_U8,	1,1,
			"Returns the number of check points of a check point group.",
			0 },
    { "CKPH$PREV",		KMP_CKPH, 0x02,  GTMD_U8M1,	KMP_MAX_PH_LINK,2,
			"Returns link 0..5 to the previous check point group.",
			0 },
    { "CKPH$NEXT",		KMP_CKPH, 0x08,  GTMD_U8M1,	KMP_MAX_PH_LINK,2,
			"Returns link 0..5 to the next check point group.",
			0 },
    { "CKPH$UNKNOWN",		KMP_CKPH, 0x0e,  GTMD_U16,	1,1,
			"Returns the unknown value of a check point group.",
			0 },

    //---------------

    { "GOBJ$N",			KMP_GOBJ,    0,  GTMD_N,	0,0,
			"Returns the number of global objects.",
			0 },
    { "GOBJ$ID",		KMP_GOBJ, 0x00,  GTMD_U16,	1,1,
			"Returns the object ID of a global object.",
			0 },
    { "GOBJ$UNKNOWN",		KMP_GOBJ, 0x02,  GTMD_U16,	1,1,
			"Returns the unknown value of a global object.",
			0 },
    { "GOBJ$POS",		KMP_GOBJ, 0x04,  GTMD_DBL3,	1,1,
			"Returns the position of a global object.",
			0 },
    { "GOBJ$ROT",		KMP_GOBJ, 0x10,  GTMD_DBL3,	1,1,
			"Returns the rotation of a global object.",
			0 },
    { "GOBJ$SCALE",		KMP_GOBJ, 0x1c,  GTMD_DBL3,	1,1,
			"Returns the scale of a global object.",
			0 },
    { "GOBJ$ROUTE",		KMP_GOBJ, 0x28,  GTMD_U16M1,	1,1,
			"Returns -1 or the index of a related route"
			" of a global object.",
			0 },
    { "GOBJ$SET",		KMP_GOBJ, 0x2a,  GTMD_U16,	8,2,
			"Returns setting 0..7 of a global object.",
			0 },
    { "GOBJ$PFLAGS",		KMP_GOBJ, 0x3a,  GTMD_U16,	1,1,
			"Returns the presence flags of a global object.",
			0 },

    //---------------

    { "POTI$N",			KMP_POTI,    0,  GTMD_N,	0,0,
			"Returns the number of routes.",
			0 },
     { "POTI$NN",		KMP_POTI,    0,  GTMD_HD_U16,	0,0,
			"Returns the total number of points of all routes.",
			0 },
    { "POTI$NP",		KMP_POTI, 0x00,  GTMD_U8,	1,1,
			"Returns the number of points of a route.",
			0 },
    { "POTI$SMOOTH",		KMP_POTI, 0x02,  GTMD_U8,	1,1,
			"Returns the smooth value of a route"
			" (route header byte at offset 0x02).",
			0 },
    { "POTI$BACK",		KMP_POTI, 0x03,  GTMD_U8,	1,1,
			"Returns the forward+backward value of a route"
			" (route header byte at offset 0x03).",
			0 },
     { "POTI$POS",		KMP_POTI, 0x10,  GTMD_DBL3,	1,2, // off+POTI_PT_OFF
			"Returns the position of a point of a route."
			" The point index is relative to the route.",
			0 },
     { "POTI$SET",		KMP_POTI, 0x1c,  GTMD_U16,	2,3, // off+POTI_PT_OFF
			"Returns setting 0 or 1 of a point of a route."
			" The point index is relative to the route.",
			0 },

    //---------------

    { "AREA$N",			KMP_AREA,    0,  GTMD_N,	0,0,
			"Returns the number of areas.",
			0 },
    { "AREA$MODE",		KMP_AREA, 0x00,  GTMD_U8M1,	1,1,
			"Returns the mode of an area.",
			0 },
    { "AREA$TYPE",		KMP_AREA, 0x01,  GTMD_U8M1,	1,1,
			"Returns the type of an area. Cameras are type 0.",
			0 },
    { "AREA$DEST",		KMP_AREA, 0x02,  GTMD_U8M1,	1,1,
			"Returns the destination index (e.g. camera index) of an area.",
			0 },
    { "AREA$UNKNOWN",		KMP_AREA, 0x03,  GTMD_U8,	1,1,
			"Returns the unknown value of an area.",
			0 },
    { "AREA$POS",		KMP_AREA, 0x04,  GTMD_DBL3,	1,1,
			"Returns the position of an area.",
			0 },
    { "AREA$ROT",		KMP_AREA, 0x10,  GTMD_DBL3,	1,1,
			"Returns the rotation of an area.",
			0 },
    { "AREA$SCALE",		KMP_AREA, 0x1c,  GTMD_DBL3,	1,1,
			"Returns the scale of an area.",
			0 },
    { "AREA$SET",		KMP_AREA, 0x28,  GTMD_U16,	4,2,
			"Returns setting 0..3 of an area.",
			0 },

    //---------------

    { "CAME$N",			KMP_CAME,    0,  GTMD_N,	0,0,
			"Returns the number of cameras.",
			0 },
     { "CAME$OCAM",		KMP_CAME,    0,  GTMD_HD_U8HI,	0,0,
			"Returns the index of the opening camera.",
			0 },
     { "CAME$SCAM",		KMP_CAME,    0,  GTMD_HD_U8LO,	0,0,
			"Returns the index of the selection camera.",
			0 },
    { "CAME$TYPE",		KMP_CAME, 0x00,  GTMD_U8,	1,1,
			"Returns the type of a camera.",
			0 },
    { "CAME$NEXT",		KMP_CAME, 0x01,  GTMD_U8M1,	1,1,
			"Returns -1 or the link to the next camera.",
			0 },
    { "CAME$UNKNOWN_02",	KMP_CAME, 0x02,  GTMD_U8,	1,1,
			"Returns the unknown value at offser 0x02 of a camera.",
			0 },
    { "CAME$ROUTE",		KMP_CAME, 0x03,  GTMD_U8M1,	1,1,
			"Returns -1 or the link to the related route.",
			0 },
    { "CAME$UNKNOWN_04",	KMP_CAME, 0x04,  GTMD_U16,	1,1,
			"Returns the unknown value at offser 0x04 of a camera.",
			0 },
    { "CAME$ZOOM_SPEED",	KMP_CAME, 0x06,  GTMD_U16,	1,1,
			"Returns the zoom speed in units per 100/60 sec.",
			0 },
    { "CAME$VIEWPT_SPEED",	KMP_CAME, 0x08,  GTMD_U16,	1,1,
			"Returns the view point speed in units per 100/60 sec.",
			0 },
    { "CAME$UNKNOWN_0A",	KMP_CAME, 0x0a,  GTMD_U16,	1,1,
			"Returns the unknown value at offser 0x0a of a camera.",
			0 },
    { "CAME$POS",		KMP_CAME, 0x0c,  GTMD_DBL3,	1,1,
			"Returns the position of a camera.",
			0 },
    { "CAME$UNKNOWN_18",	KMP_CAME, 0x18,  GTMD_DBL3,	1,1,
			"Returns the unknown value at offser 0x18 of a camera.",
			0 },
    { "CAME$ZOOM_BEGIN",	KMP_CAME, 0x24,  GTMD_DBL,	1,1,
			"Returns the zoom beginning angle of a camera.",
			0 },
    { "CAME$ZOOM_END",		KMP_CAME, 0x28,  GTMD_DBL,	1,1,
			"Returns the zoom ending angle of a camera.",
			0 },
    { "CAME$VIEWPT_BEGIN",	KMP_CAME, 0x2c,  GTMD_DBL3,	1,1,
			"Returns the start position of the view point.",
			0 },
    { "CAME$VIEWPT_END",	KMP_CAME, 0x38,  GTMD_DBL3,	1,1,
			"Returns the end position of the view point.",
			0 },
    { "CAME$TIME",		KMP_CAME, 0x44,  GTMD_DBL,	1,1,
			"Returns the run time of a camera in 1/60 sec.",
			0 },

    //---------------

    { "JGPT$N",			KMP_JGPT,    0,  GTMD_N,	0,0,
			"Returns the number of respan points.",
			0 },
    { "JGPT$POS",		KMP_JGPT, 0x00,  GTMD_DBL3,	1,1,
			"Returns the position of a respawn point.",
			0 },
    { "JGPT$ROT",		KMP_JGPT, 0x0c,  GTMD_DBL3,	1,1,
			"Returns the rotation of a respawn point.",
			0 },
    { "JGPT$ID",		KMP_JGPT, 0x18,  GTMD_U16,	1,1,
			"Returns the ID of a respawn point.",
			0 },
    { "JGPT$EFFECT",		KMP_JGPT, 0x1a,  GTMD_U16,	1,1,
			"Returns the effect value of a respawn point.",
			0 },

    //---------------

    { "CNPT$N",			KMP_CNPT,    0,  GTMD_N,	0,0,
			"Returns the number of canon point points.",
			0 },
    { "CNPT$POS",		KMP_CNPT, 0x00,  GTMD_DBL3,	1,1,
			"Returns the position of a canon point.",
			0 },
    { "CNPT$ROT",		KMP_CNPT, 0x0c,  GTMD_DBL3,	1,1,
			"Returns the rotation of a canon point.",
			0 },
    { "CNPT$ID",		KMP_CNPT, 0x18,  GTMD_U16,	1,1,
			"Returns the ID of a canon point.",
			0 },
    { "CNPT$EFFECT",		KMP_CNPT, 0x1a,  GTMD_U16,	1,1,
			"Returns the efffect value of a canon point.",
			0 },

    //---------------

    { "MSPT$N",			KMP_MSPT,    0,  GTMD_N,	0,0,
			"Returns the number of battle points.",
			0 },
    { "MSPT$POS",		KMP_MSPT, 0x00,  GTMD_DBL3,	1,1,
			"Returns the position of a battle point.",
			0 },
    { "MSPT$ROT",		KMP_MSPT, 0x0c,  GTMD_DBL3,	1,1,
			"Returns the rotation of a battle point.",
			0 },
    { "MSPT$ID",		KMP_MSPT, 0x18,  GTMD_U16,	1,1,
			"Returns the ID of a battle point.",
			0 },
    { "MSPT$EFFECT",		KMP_MSPT, 0x1a,  GTMD_U16,	1,1,
			"Returns the efffect value of a battle point.",
			0 },

    //---------------

    { "STGI$N",			KMP_STGI,    0,  GTMD_N,	0,0,
			"Returns the number of stage infos.",
			0 },
    { "STGI$BYTE",		KMP_STGI, 0x00,  GTMD_U8,	12,2,
			"Returns setting byte 0..11 of a stage info.",
			0 },
};

#define KMP_FUNC_LIST_N (sizeof(kmp_func_list)/sizeof(*kmp_func_list))

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static enumError F_get
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    const FuncParam_t	* fpar		// pointer to user parameter
)
{
    DASSERT(res);
    DASSERT(param);
    DASSERT(fpar);
    DASSERT( fpar->user_id >= 0 && fpar->user_id < KMP_FUNC_LIST_N );

    if ( si && si->kmp )
    {
	const kmp_func_t *fp = kmp_func_list + fpar->user_id;
	DASSERT( n_param >= fp->n_param );
	DASSERT( fp->sect < KMP_N_SECT );
	DASSERT( fp->get_info );

	const List_t *dlist = si->kmp->dlist + fp->sect;
	uint max_index1 = dlist->used;
	int val = n_param > 0 ? GetIntV(param) : 0;
	uint index1 = val < 0 ? val + max_index1 : val;
	const u8 *data = dlist->list + dlist->elem_size * index1 + fp->off;

	if ( fp->sect == KMP_POTI && fp->off >= POTI_PT_OFF && index1 < max_index1 )
	{
	    param++;
	    n_param--;

	    const kmp_poti_group_t *pgrp = (kmp_poti_group_t*)dlist->list;
	    uint pt_start = 0;
	    while ( index1-- > 0 )
		pt_start += (pgrp++)->n_point;

	    dlist = &si->kmp->poti_point;
	    max_index1 = pgrp->n_point;
	    val = n_param > 0 ? GetIntV(param) : 0;
	    index1 = val < 0 ? val + max_index1 : val;
	    PRINT("POTI/PT: index= %u/%u  +%u\n",index1,max_index1,pt_start);
	    data = dlist->list
		 + dlist->elem_size * ( pt_start + index1 )
		 + fp->off - POTI_PT_OFF;
	}

	val = n_param > 1 ? GetIntV(param+1) : 0;
	const uint index2 = val < 0 ? val + fp->n_var : val;

	const bool valid_index12 = index1 < max_index1 && index2 < fp->n_var;
	noPRINT("index1=%d/%d, index2=%d/%d, valid=%d\n",
		index1, max_index1, index2, fp->n_var, valid_index12 );

	switch(fp->get_mode)
	{
	    case GTMD_N:
		res->i = dlist->used;
		res->mode = VAR_INT;
		return ERR_OK;

	    case GTMD_HD_U16:
		res->i = si->kmp->value[fp->sect];
		res->mode = VAR_INT;
		return ERR_OK;

	    case GTMD_HD_U8HI:
		res->i = si->kmp->value[fp->sect] >> 8;
		res->mode = VAR_INT;
		return ERR_OK;

	    case GTMD_HD_U8LO:
		res->i = si->kmp->value[fp->sect] & 0xff;
		res->mode = VAR_INT;
		return ERR_OK;

	    case GTMD_U8:
	    case GTMD_U8M1:
		if (valid_index12)
		{
		    const u8 val = data[index2];
		    res->i = val == 0xff && fp->get_mode == GTMD_U8M1 ? -1 : (int)val;
		    res->mode = VAR_INT;
		    return ERR_OK;
		}
		break;

	    case GTMD_U16:
	    case GTMD_U16M1:
		if (valid_index12)
		{
		    const u16 val = ((u16*)data)[index2];
		    res->i = val == 0xffff && fp->get_mode == GTMD_U16M1
				? -1 : (int)val;
		    res->mode = VAR_INT;
		    return ERR_OK;
		}
		break;

	    case GTMD_DBL:
		if (valid_index12)
		    return ResultFloat( res, (float*)data + index2, 1 );
		break;

	    case GTMD_DBL2:
		if (valid_index12)
		    return ResultFloat( res, (float*)data + 2 * index2, 2 );
		break;

	    case GTMD_DBL3:
		if (valid_index12)
		    return ResultFloat( res, (float*)data + 3 * index2, 3 );
		break;
	}
    }

    res->i = 0;
    res->mode = VAR_UNSET;
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////
#ifdef TEST

static enumError F_set
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    const FuncParam_t	* fpar		// pointer to user parameter
)
{
    DASSERT(res);
    DASSERT(param);
    DASSERT(fpar);
    DASSERT( fpar->user_id >= 0 && fpar->user_id < KMP_FUNC_LIST_N );

    if ( si && si->kmp )
    {
	const kmp_func_t *fp = kmp_func_list + fpar->user_id;
	DASSERT( n_param >= fp->n_param );
	DASSERT( fp->sect < KMP_N_SECT );
	DASSERT( fp->set_info );

	PRINT(">>> F_SET <<<\n");
	// ...
    }

    res->i = 0;
    res->mode = VAR_UNSET;
    return ERR_OK;
}

#endif
//
///////////////////////////////////////////////////////////////////////////////
///////////////			more kmp functions		///////////////
///////////////////////////////////////////////////////////////////////////////

static int get_next
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    int			user_id,	// user defined id
    int			* cur_idx	// not NULL: store current index here
)
{
    DASSERT(res);
    DASSERT(param);
    DASSERT(n_param>=1);
    DASSERT( user_id == KMP_ENPT || user_id == KMP_ITPT );

    // [[2do]] use GetNextPointKMP()

    if (cur_idx)
	*cur_idx = -1;

    int idx = GetIntV(param);
    res->i = 0;
    res->mode = VAR_UNSET;

    if ( si && si->kmp )
    {
	const kmp_t *kmp = si->kmp;

	const List_t *ptlist, *phlist;
	if ( user_id == KMP_ENPT )
	{
	    ptlist = kmp->dlist + KMP_ENPT;
	    phlist = kmp->dlist + KMP_ENPH;
	}
	else
	{
	    ptlist = kmp->dlist + KMP_ITPT;
	    phlist = kmp->dlist + KMP_ITPH;
	}

	if ( idx < 0 )
	    idx += ptlist->used;

	if ( idx >= 0 && idx < ptlist->used )
	{
	    if (cur_idx)
		*cur_idx = idx;

	    const kmp_enph_entry_t *ph = (kmp_enph_entry_t*)phlist->list;
	    uint n = phlist->used;
	    while ( idx >= ph->pt_start + ph->pt_len && n-- > 0 )
		ph++;

	    noPRINT(">> %u %zu[%u+%u]\n", idx,
			ph-(kmp_enph_entry_t*)phlist->list,ph->pt_start,ph->pt_len);
	    if ( idx == ph->pt_start + ph->pt_len - 1 )
	    {
		idx = -1; // invalidate
		uint ph_idx = ~0;

		if ( n_param > 1 )
		{
		    uint idx2 = GetIntV(param+1);
		    if ( idx2 < KMP_MAX_PH_LINK )
			ph_idx = ph->next[idx2];
		}
		else
		{
		    uint idx2;
		    for ( idx2 = 0; idx2 < KMP_MAX_PH_LINK; idx2++ )
		    {
			if (!IS_M1(ph->next[idx2]))
			{
			    ph_idx = ph->next[idx2];
			    break;
			}
		    }
		}

		if ( ph_idx < phlist->used )
		{
		    ph = (kmp_enph_entry_t*)phlist->list + ph_idx;
		    idx = ph->pt_start;
		}
	    }
	    else
		idx++;

	    if ( idx >= 0 && idx < ptlist->used )
		return idx;
	}
    }
    return -1;
}

///////////////////////////////////////////////////////////////////////////////

static enumError F_get_next
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    const FuncParam_t	* fpar		// pointer to user parameter
)
{
    DASSERT(res);
    DASSERT(param);
    DASSERT(n_param>=1);
    DASSERT(fpar);
    DASSERT( fpar->user_id == KMP_ENPT || fpar->user_id == KMP_ITPT );

    int next = get_next(res,param,n_param,si,fpar->user_id,0);
    if ( next >= 0 )
    {
	res->i = next;
	res->mode = VAR_INT;
    }
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError F_get_hdir
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    const FuncParam_t	* fpar		// pointer to user parameter
)
{
    DASSERT(res);
    DASSERT(param);
    DASSERT(n_param>=1);
    DASSERT(fpar);
    DASSERT( fpar->user_id == KMP_ENPT || fpar->user_id == KMP_ITPT );

    int current;
    int next = get_next(res,param,n_param,si,fpar->user_id,&current);
    if ( next >= 0 )
    {
	DASSERT(si);
	const kmp_t *kmp = si->kmp;
	DASSERT(kmp);

	const kmp_enpt_entry_t *pt = (kmp_enpt_entry_t*)kmp->dlist[fpar->user_id].list;
	res->d = atan2( pt[next].position[0] - pt[current].position[0],
			pt[next].position[2] - pt[current].position[2] ) * (180/M_PI);
	res->mode = VAR_DOUBLE;
    }
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError F_isKMP
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    const FuncParam_t	* fpar		// pointer to user parameter
)
{
    DASSERT(res);
    DASSERT(param);

    uint result = 0;
    if ( si && si->kmp )
    {
	result = 1;

	if (si->kmp->fname)
	{
	    ccp fname = strrchr(si->kmp->fname,'/');
	    if (fname)
		fname++;
	    else
		fname = si->kmp->fname;

	    if ( !strcasecmp(fname,"course.kmp")
		|| !strcasecmp(fname,"course.txt")
		|| !strcasecmp(fname,"course.txt.kmp") )
	    {
		result = 2;
	    }
	}
    }

    res->i    = result;
    res->mode = VAR_INT;
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static const struct FuncTable_t kmp_func_tab[] =
{
    { 1, 2, "ENPT$NEXT", F_get_next, KMP_ENPT,
	"int", "enpt$next(enpt_index[,next_index])",
	"Returns the index of the next enemy point."
	" The 'next_index' is only used, if multiple links are available." },

    { 1, 2, "ITPT$NEXT", F_get_next, KMP_ITPT,
	"int", "itpt$next(itpt_index[,next_index])",
	"Returns the index of the next item point."
	" The 'next_index' is only used, if multiple links are available." },

    { 1, 2, "ENPT$HDIR", F_get_hdir, KMP_ENPT,
	"float", "enpt$hDir(enpt_index[,next_index])",
	"Calculate the horizontal direction in degree of an enemy point."
	" The 'next_index' is only used, if multiple links are available." },

    { 1, 2, "ITPT$HDIR", F_get_hdir, KMP_ITPT,
	"float", "itpt$hDir(itpt_index[,next_index])",
	"Calculate the horizontal direction in degree of an item point."
	" The 'next_index' is only used, if multiple links are available." },

    { 0, 0, "ISKMP", F_isKMP, 0,	// replace only function call
	"int", "isKMP()", 0 },		// but not info

    {0,0,0,0,0,0,0,0}
};

//
///////////////////////////////////////////////////////////////////////////////
///////////////			setup kmp parser		///////////////
///////////////////////////////////////////////////////////////////////////////

const VarMap_t * SetupVarsKMP()
{
    static VarMap_t vm = {0};
    if (!vm.used)
    {
	//--- setup parser function get()

	uint i;
	const kmp_func_t *fp = kmp_func_list;
	for ( i = 0; i < KMP_FUNC_LIST_N; i++, fp++ )
	{
	  char syntax[80];
	  if ( fp->get_info)
	  {
	    DefineParserFunc(fp->name,fp->n_param,fp->n_param,FF_KMP,F_get,i);

	    ccp retval = fp->get_mode == GTMD_DBL
				? "float"
				: fp->get_mode > GTMD_DBL
					? "vector"
					: "int";
	    char *dest = syntax;
	    ccp src = fp->name;
	    while (*src)
		*dest++ = tolower((int)*src++);
	    *dest++ = '(';
	    if ( fp->n_param > 0 )
	    {
		src = fp->sect == KMP_POTI ? "route" : kmp_section_name[fp->sect].name1;
		while (*src)
		    *dest++ = tolower((int)*src++);

		switch(fp->n_param)
		{
		    case 1:
			src = "_idx";
			break;

		    case 2:
			src = fp->sect == KMP_POTI && fp->off >= POTI_PT_OFF
				    ? "_idx,pt_idx"
				    : "_idx,val_idx";
			break;

		    case 3:
			src = "_idx,pt_idx,val_idx";
			break;

		};
		while (*src)
		    *dest++ = *src++;
	    }
	    *dest++ = ')';
	    *dest = 0;
	    DASSERT(dest < syntax + sizeof(syntax) );

	    DefineParserFuncInfo(retval,syntax,fp->get_info);
	  }

    #ifdef TEST // [[2do]]
	  if (fp->set_info)
	  {
	    char namebuf[40];
	    StringCat2S(namebuf,sizeof(namebuf),"SET_",fp->name);
	    DefineParserFunc(namebuf,fp->n_param+1,fp->n_param+1,FF_KMP,F_set,i);

	    char *dest = syntax;
	    ccp src = namebuf;
	    while (*src)
		*dest++ = tolower((int)*src++);
	    *dest++ = '(';
	    if ( fp->n_param > 0 )
	    {
		src = fp->sect == KMP_POTI ? "route" : kmp_section_name[fp->sect].name1;
		while (*src)
		    *dest++ = tolower((int)*src++);

		switch(fp->n_param)
		{
		    case 1:
			src = "_idx";
			break;

		    case 2:
			src = fp->sect == KMP_POTI && fp->off >= POTI_PT_OFF
				    ? "_idx,pt_idx"
				    : "_idx,val_idx";
			break;

		    case 3:
			src = "_idx,pt_idx,val_idx";
			break;

		};
		while (*src)
		    *dest++ = *src++;
	    }
	    DASSERT(dest < syntax + sizeof(syntax) );
	    StringCopyE(dest,syntax+sizeof(syntax),",value)");
	    DefineParserFuncInfo("int",syntax,fp->set_info);
	  }
    #endif
	}


	//--- setup additional functions

	DefineParserFuncTab(kmp_func_tab,FF_KMP);
	SetupReferenceKCL(&vm);


	//--- setup object names

	char varname[50], *varname_end = varname + sizeof(varname)-1;
	const ObjectInfo_t *oi, *oi_end = ObjectInfo + N_KMP_GOBJ;
	int index = 0;
	for ( oi = ObjectInfo; oi < oi_end; oi++, index++ )
	    if (oi->name)
	    {
		char * vn = varname;
		*vn++ = 'O';
		*vn++ = '$';
		ccp src = oi->name;
		while ( *src && vn < varname_end )
		    *vn++ = toupper((int)*src++);
		*vn = 0;

		DefineIntVar(&vm,varname,index);
	    }

	const KeywordTab_t *key;
	for ( key = kmp_section_name; key->name2; key++ )
	{
	    snprintf(varname,sizeof(varname),"SECT$%s",key->name1);
	    DefineIntVar(&vm,varname,key->id+1);
	}

	//--- setup integer variables

	struct inttab_t { ccp name; int val; };
	static const struct inttab_t inttab[] =
	{
	    { "MODE$SLOT",		0 },
	    { "MODE$D",			0 },

	    { "MODE$TEST",		0 },
	    { "MODE$OCAM",		0 },
	    { "MODE$LAPS",		3 },
	    { "MODE$SHOWRT",		0 },
	    { "MODE$SHOWCK",		0 },
	    { "MODE$AUTOMODE",		0 },

	    { "AM$OFF",			KMP_AM_OFF },
	    { "AM$NORM",		KMP_AM_NORM },
	    { "AM$RENUMBER",		KMP_AM_RENUMBER },
	    { "AM$1LAP",		KMP_AM_1_LAP },
	    { "AM$SHORT",		KMP_AM_SHORT },
	    { "AM$UNLIMIT",		KMP_AM_UNLIMIT },

	    { "SORT$OFF",		KMP_SORT_OFF },
	    { "SORT$GROUPING",		KMP_SORT_GROUPING },
	    { "SORT$ANGLE",		KMP_SORT_ANGLE },

	    { "F$FALL",			KPFL_FALL },
	    { "F$SNAP",			KPFL_SNAP },
	    { "F$JGPT",			KPFL_JGPT },

	    { "ITEM$RANDOM",		0x00 },
	    { "ITEM$BANANA",		0x01 },
	    { "ITEM$MUSHROOM",		0x02 },
	    { "ITEM$3MUSHROOMS",	0x03 },
	    { "ITEM$STAR",		0x04 },
	    { "ITEM$3GREEN",		0x05 },
	    { "ITEM$BANANA_MUSHROOM",	0x06 },
	    { "ITEM$GREEN",		0x07 },
	    { "ITEM$RED",		0x08 },
	    { "ITEM$MEGA",		0x0a },
	    { "ITEM$CLOUD",		0x0b },
	    { "ITEM$STAR_MUSHROOM",	0x0c },
	    { "ITEM$MUSHROOM_GREEN",	0x0d },
	    { "ITEM$3BANANAS_GREEN",	0x24 },

	    { "PF$OFFLINE1",		0x01 },
	    { "PF$OFFLINE2",		0x02 },
	    { "PF$OFFLINE3",		0x04 },
	    { "PF$OFFLINE",		0x07 },
	    { "PF$ONLINE1",		0x08 },
	    { "PF$ONLINE2",		0x10 },
	    { "PF$ONLINE3",		0x20 },
	    { "PF$ONLINE",		0x38 },
	    { "PF$ALL",			0x3f },

	    {0,0}
	};

	const struct inttab_t * ip;
	for ( ip = inttab; ip->name; ip++ )
	    DefineIntVar(&vm,ip->name,ip->val);


	//--- setup basic parser variables

	DefineParserVars(&vm);
    }

    return &vm;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP: scan helpers		///////////////
///////////////////////////////////////////////////////////////////////////////

static uint test_unknown_sections
(
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    ccp			fname		// file name for error messages
)
{
    const kmp_file_t * kfile = data;
    const uint base_off = be16(&kfile->sect_base_off);
    const uint max_off  = data_size - base_off - 4;
    const uint n_sect = be16(&kfile->n_sect);

    uint err_count = 0;
    int s;
    for ( s = -1; s < (int)n_sect; s++ )
    {
	const uint off = be32(kfile->sect_off + s);
	noPRINT("%3d: %x/%x\n",s,off,max_off);
	if ( off > max_off || off&3 )
	    continue;

	ccp expected_sect_name = s < 0 ? "----" : kmp_section_name[s].name1;
	const kmp_list_head_t * temp
	    = (kmp_list_head_t*) ( data + base_off + off );
	if (memcmp(temp->magic,expected_sect_name,sizeof(temp->magic)))
	{
	    uint i;
	    for ( i = 0; i < KMP_N_SECT; i++ )
	    {
		const uint off = be32(kfile->sect_off + i);
		if ( off > max_off || off&3 )
		    continue;
		const kmp_list_head_t * temp2
			= (kmp_list_head_t*) ( data + base_off + off );
		if (!memcmp(temp2->magic,expected_sect_name,sizeof(temp2->magic)))
		    break;
	    }

	    if ( i < KMP_N_SECT )
		ERROR0(ERR_WARNING,
		    "KMP: Section [%s] found at index %u, but not at %d: %s\n",
		    expected_sect_name, i, s, fname );
	    else if ( s >= 0 )
		ERROR0(ERR_WARNING,
		    "KMP: Unknown section [%s]: %s\n",
		    PrintID(temp->magic,sizeof(temp->magic),0), fname );
	    err_count++;
	}
    }
    return err_count;
}

///////////////////////////////////////////////////////////////////////////////

static const void * get_kmp_pointer
(
    uint		* res_n,	// not NULL: result: number of entries
    uint		* res_val,	// not NULL: result: unknown value
    kmp_t		* kmp,		// KMP data structure
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    uint		sect		// KMP section index
)
{
    DASSERT(kmp);
    DASSERT(data);
    DASSERT(sect<KMP_N_SECT);

    ccp sect_name = kmp_section_name[sect].name1;
    const kmp_file_t * kfile = data;
    const uint base_off = be16(&kfile->sect_base_off);
    const uint max_off  = data_size - base_off - 4;

    const kmp_list_head_t * list = 0;
    const uint n_sect = be16(&kfile->n_sect);

    if ( sect < n_sect )
    {
	// optimization: try standard slot
	const uint off = be32( kfile->sect_off + sect );
	if ( off <= max_off && !(off&3) )
	{
	    list = (kmp_list_head_t*) ( data + base_off + off );
	    if (memcmp(list->magic,sect_name,sizeof(list->magic)))
		list = 0;
	}
    }

    if (!list)
    {
	// not found yet: try iteration

	noPRINT("'%s' FAILED, TRY ITERATION, ns=%d, maxoff=%x\n",
		sect_name, n_sect, max_off );

	int s;
	for ( s = -1; s < (int)n_sect; s++ )
	{
	    const uint off = be32(kfile->sect_off + s);
	    noPRINT("%3d: %x/%x\n",s,off,max_off);
	    if ( off > max_off || off&3 )
		continue;

	    const kmp_list_head_t * temp
		= (kmp_list_head_t*) ( data + base_off + off );
	    if (!memcmp(temp->magic,sect_name,sizeof(temp->magic)))
	    {
		list = temp;
		break;
	    }
	}

	if (!list)
	{
	    ERROR0(ERR_WARNING,
		"KMP: Section [%s] not found: %s\n",
		kmp_section_name[sect].name1, kmp->fname );
	    if (res_n)
		*res_n = 0;
	    if (res_val)
		*res_val = 0;
	    return 0;
	}

	noPRINT("'%s' FOUND AT SLOT %u\n",sect_name,s);
    }

    kmp->value[sect] = be16(&list->value);
    const u8 * elem = list->entry;
    uint n = be16(&list->n_entry);
    uint max = ( (u8*)data + data_size - elem ) / kmp_entry_size[sect];
    if ( n > max )
    {
	ERROR0(ERR_WARNING,
		"Number of [%s] entries exceed file size (%u>%u): %s\n",
		kmp_section_name[sect].name1, n, max, kmp->fname );
	n = max;
    }
    if (res_n)
	*res_n = n;
    if (res_val)
	*res_val = be16(&list->value);

    const u8 *end = elem + n * kmp_entry_size[sect];
    if ( kmp->max_scanned < end )
	 kmp->max_scanned = end;

    return elem;
}

///////////////////////////////////////////////////////////////////////////////

static void scan_kmp_pt
(
    kmp_t		* kmp,		// KMP data structure
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    uint		sect		// KMP section index
)
{
    uint n;
    const kmp_enpt_entry_t * pt = get_kmp_pointer(&n,0,kmp,data,data_size,sect);
    for ( ; n > 0; n--, pt++ )
    {
	kmp_enpt_entry_t * item = AppendList(kmp->dlist+sect);
	DASSERT(item);
	bef4n(item->position,pt->position,4);
	be16n(item->prop,pt->prop,2);
    }
}

///////////////////////////////////////////////////////////////////////////////

static int sort_ph ( const kmp_enph_entry_t * a, const kmp_enph_entry_t * b )
{
    int stat = a->pt_start - b->pt_start;
    if (!stat)
	stat = a->pt_len - b->pt_len;
    return stat;
}

//-----------------------------------------------------------------------------

static void scan_kmp_ph
(
    kmp_t		* kmp,		// KMP data structure
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    uint		sect_ph,	// PH section index
    kmp_group_name_t	* gname		// NULL or list with names
)
{
    DASSERT(kmp);
    DASSERT(data);

    uint n;
    const kmp_enph_entry_t *ph = get_kmp_pointer(&n,0,kmp,data,data_size,sect_ph);
    for ( ; n > 0; n--, ph++ )
    {
	kmp_enph_entry_t * item = AppendList(kmp->dlist+sect_ph);
	DASSERT(item);
	memcpy(item,ph,sizeof(*item));
    }

    //--- sort groups

    kmp_enph_entry_t *list = (kmp_enph_entry_t*)(kmp->dlist[sect_ph].list);
    qsort( list, kmp->dlist[sect_ph].used, sizeof(*list), (qsort_func)sort_ph );

    //--- assign group names

    if (gname)
	RenameGroup(kmp,sect_ph,gname);
}

///////////////////////////////////////////////////////////////////////////////

static void scan_kmp_pt2
(
    kmp_t		* kmp,		// KMP data structure
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    uint		sect		// KMP section index
)
{
    DASSERT(kmp);
    DASSERT( sect < KMP_N_SECT );

    uint n;
    const kmp_jgpt_entry_t * pt = get_kmp_pointer(&n,0,kmp,data,data_size,sect);
    FillIL(kmp->index+sect,n);
    for ( ; n > 0; n--, pt++ )
    {
	kmp_jgpt_entry_t * item = AppendList(kmp->dlist+sect);
	DASSERT(item);
	bef4n(item->position,pt->position,6);
	be16n((u16*)&item->id,(u16*)&pt->id,2);
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			kcl fall helpers		///////////////
///////////////////////////////////////////////////////////////////////////////
// [[kcl_fall_t]]

typedef struct kcl_fall_t
{
    //--- input

    kcl_t	*fall_kcl;		// KCL to use, never NULL if 'kf->valid'

    int		width;			// input for "KCL-FALL-WIDTH"
    Var_t	pre_add;		// input for "KCL-FALL-PRE-ADD"
    Var_t	post_add;		// input for "KCL-FALL-POST-ADD"
    u32		types;			// input for "KCL-FALL-TYPES"

    //---- status

    bool	valid;			// true: calculation can be done

} kcl_fall_t;

///////////////////////////////////////////////////////////////////////////////

#define KCL_FALL_VARS(k,v) \
	{ "KCL-FALL-ADD-ROAD",	SPM_BOOL, &k->enable_fall_kcl }, \
	{ "KCL-FALL-WIDTH",	SPM_INT,  &v.width }, \
	{ "KCL-FALL-PRE-ADD",	SPM_VAR,  &v.pre_add }, \
	{ "KCL-FALL-POST-ADD",	SPM_VAR,  &v.post_add }, \
	{ "KCL-FALL-TYPES",	SPM_U32,  &v.types },

///////////////////////////////////////////////////////////////////////////////

void ResetKclFall ( kcl_fall_t *kf )
{
    DASSERT(kf);
    memset(kf,0,sizeof(*kf));
    kf->types = M1(kf->types);
}

///////////////////////////////////////////////////////////////////////////////

bool CheckKclFall ( kcl_fall_t *kf )
{
    DASSERT(kf);

    kf->fall_kcl = 0;
    kf->valid = kf->width > 0 && kf->types;
    if (kf->valid)
    {
	ToVectorYV(&kf->pre_add);
	ToVectorYV(&kf->post_add);

	kf->fall_kcl = GetReferenceKCL();
	if (!kf->fall_kcl)
	    kf->valid = false;
    }
    return kf->valid;
}

///////////////////////////////////////////////////////////////////////////////

void CalcKclFall ( kcl_fall_t *kf, float *vector )
{
    DASSERT(kf);
    if (kf->valid)
    {
	DASSERT(kcl_ref);
	DASSERT(kf->fall_kcl);

	double3 pt;
	pt.x = vector[0] + kf->pre_add.x;
	pt.y = vector[1] + kf->pre_add.y;
	pt.z = vector[2] + kf->pre_add.z;
	double temp = FallKCL(kf->fall_kcl,0,pt,kf->width,0,kf->types,0);
	if ( temp >= 0.0 &&  temp < pt.y )
	{
	    vector[0] = kf->post_add.x + pt.x;
	    vector[1] = kf->post_add.y + temp;
	    vector[2] = kf->post_add.z + pt.z;
	}
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			route object helpers		///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct rtobj_info_t
{
    //--- input

    int			route_object;		// input for "ROUTE-OBJECT"
    Var_t		route_shift;		// input for "ROUTE-OBJECT-SHIFT"
    Var_t		route_scale;		// input for "ROUTE-OBJECT-SCALE"
    Var_t		route_rotate;		// input for "ROUTE-OBJECT-ROTATE"
    int			show_route;		// input for "SHOW-ROUTE"
    int			route_itembox;		// input for "ROUTE-ITEMBOX"

    //--- data

    kmp_rtobj_t		*def_obj;		// default object
    kmp_rtobj_t		rtobj;			// resulting data
    bool		valid;			// true: rtobj is set

} rtobj_info_t;

///////////////////////////////////////////////////////////////////////////////

#define ROUTE_OBJECT_VARS(v) \
	{ "ROUTE-OBJECT",	SPM_INT, &v.route_object }, \
	{ "ROUTE-OBJECT-SHIFT",	SPM_VAR, &v.route_shift }, \
	{ "ROUTE-OBJECT-SCALE",	SPM_VAR, &v.route_scale }, \
	{ "ROUTE-OBJECT-ROTATE",SPM_VAR, &v.route_rotate }, \
	{ "SHOW-ROUTE",		SPM_INT, &v.show_route }, \
	{ "ROUTE-ITEMBOX",	SPM_INT, &v.route_itembox },

///////////////////////////////////////////////////////////////////////////////

void ResetRouteInfo ( rtobj_info_t *ri, kmp_rtobj_t *def_obj )
{
    DASSERT(ri);
    DASSERT(def_obj);

    memset(ri,0,sizeof(*ri));
    ri->def_obj		= def_obj;
    memcpy(&ri->rtobj,ri->def_obj,sizeof(ri->rtobj));
    ri->route_object	= N_KMP_GOBJ;
}

///////////////////////////////////////////////////////////////////////////////

bool CheckRouteInfo ( rtobj_info_t *ri )
{
    DASSERT(ri);
    DASSERT(ri->def_obj);
    noPRINT_IF( ri->route_object != N_KMP_GOBJ,
		">> CheckRouteInfo() obj=%d, shrt=%d\n",
		ri->route_object, ri->show_route );

    if (!ri->route_object)
	ri->rtobj.obj = 0;
    else if ( ri->route_object < 0 )
    {
	memcpy(&ri->rtobj,ri->def_obj,sizeof(ri->rtobj));
	ri->show_route = 1;
    }
    else if ( ri->route_object < N_KMP_GOBJ )
    {
	ri->rtobj.obj = ri->route_object;
	ri->show_route = 1;
    }
    ri->route_object = N_KMP_GOBJ;

    if ( ri->route_shift.mode != VAR_UNSET )
    {
	ri->rtobj.shift = GetVectorFV(&ri->route_shift);
	ri->route_shift.mode = VAR_UNSET;
    }

    if ( ri->route_scale.mode != VAR_UNSET )
    {
	ri->rtobj.scale = GetVectorFV(&ri->route_scale);
	ri->route_scale.mode = VAR_UNSET;
    }

    if ( ri->route_rotate.mode != VAR_UNSET )
    {
	ri->rtobj.rotate = GetVectorFV(&ri->route_rotate);
	ri->route_rotate.mode = VAR_UNSET;
    }

    if ( ri->show_route > 0 && !ri->rtobj.obj )
	ri->rtobj.obj = ri->def_obj->obj;

    if ( ri->route_itembox > 0 && !ri->valid )
    {
	ri->show_route = 1;
	ri->rtobj.obj = GOBJ_ITEMBOX;
    }

    ri->valid = ri->show_route > 0 && ri->rtobj.obj;
    return ri->valid;
}

///////////////////////////////////////////////////////////////////////////////

bool AssignRouteInfo ( rtobj_info_t *ri, kmp_rtobj_t *dest )
{
    DASSERT(ri);
    DASSERT(dest);
    if (ri->valid)
	memcpy(dest,&ri->rtobj,sizeof(*dest));
    else
	memset(dest,0,sizeof(*dest));

    return ri->valid;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////		   ScanTextKMP() sub functions		///////////////
///////////////////////////////////////////////////////////////////////////////

static kmp_pt_flag_t ScanPtFlags ( ScanInfo_t *si )
{
    DASSERT(si);
    
    if ( NextCharSI(si,false) == '(' )
    {
	si->cur_file->ptr++;
	Var_t val;
	CheckWarnSI(si,')',ScanExprSI(si,&val));
	return GetIntV(&val);
    }

    char fbuf[20];
    ScanNameSI(si,fbuf,sizeof(fbuf),true,true,0);

    kmp_pt_flag_t flags = 0;
    if ( strchr(fbuf,'F') ) flags |= KPFL_FALL;
    if ( strchr(fbuf,'S') ) flags |= KPFL_SNAP;
    if ( strchr(fbuf,'J') ) flags |= KPFL_JGPT;

 #if 0
    if ( si->cur_file->revision < xxxx && !( flags & KPFL_UPDATE_Rxxxx ) )
	flags |= KPFL_UPDATE_Rxxxx;
 #endif
    return flags;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanText_PT_PH
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si,		// valid data
    uint		sect_pt,	// KMP section index
    uint		sect_ph,	// related ph section index
    kmp_group_name_t	* gname,	// NULL or pointer to group name array
    u8			* pt_flags	// NULL or flags vector
)
{
    DASSERT(kmp);
    DASSERT(si);
    DASSERT( sect_pt < KMP_N_SECT );
    DASSERT( sect_ph < KMP_N_SECT );

    ccp pt_name = kmp_section_name[sect_pt].name1;
    ccp ph_name = kmp_section_name[sect_ph].name1;
    TRACE("ScanText_PT_PH(%s,%s)\n",pt_name,ph_name);

    char pt_hd_name[20], ph_hd_name[20];
    snprintf( pt_hd_name,   sizeof(pt_hd_name),   "%s-HEAD-VALUE",      pt_name );
    snprintf( ph_hd_name,   sizeof(ph_hd_name),	  "%s-HEAD-VALUE",      ph_name );

    rtobj_info_t rtinfo;
    ResetRouteInfo(&rtinfo,&kmp->rtobj);
    kcl_fall_t kfall;
    ResetKclFall(&kfall);
    int auto_group_name = 0, rotate_group_links = 0;
    double auto_fill = 0.0;
    u16 mask1, mask2;
    mask1 = mask2 = 0xffff;

    ScanParam_t ptab[] =
    {
	{ pt_hd_name,		SPM_U16,    kmp->value+sect_pt },
	{ ph_hd_name,		SPM_U16,    kmp->value+sect_ph },
	{ "AUTO-GROUP-NAME",	SPM_INT,    &auto_group_name },
	{ "ROTATE-GROUP-LINKS",	SPM_INT,    &rotate_group_links },
	{ "AUTO-FILL",		SPM_DOUBLE, &auto_fill },
	{ "MASK-1",		SPM_U16,    &mask1 },
	{ "MASK-2",		SPM_U16,    &mask2 },
	ROUTE_OBJECT_VARS(rtinfo)
	KCL_FALL_VARS(kmp,kfall)
	{0}
    };

    kmp_group_list_t gl;
    InitializeGL(&gl,*pt_name);
    List_t *ptlist = kmp->dlist + sect_pt;
    ptlist->used = 0;

    enumError err = ERR_OK;
    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    CheckRouteInfo(&rtinfo);
	    CheckKclFall(&kfall);
	    continue;
	}

	if ( ch == '$' )
	{
	    AddGroupGL(&gl,si,kmp_section_name[sect_ph].name1[0],ptlist->used,3);
	    continue;
	}

	u16 idx;
	if (ScanU16SI(si,&idx,1,0))
	{
	    err = ERR_WARNING;
	    break;
	}

	kmp_enpt_entry_t temp;
	ScanFloatV3SI(si,temp.position,1);
	ScanFloatSI(si,&temp.scale,1);
	if ( si->cur_file->szs_modifier > 0 )
	    ScanBE32SI_swap(si,(u32*)temp.prop,1,1);
	else
	    ScanU16SI(si,temp.prop,2,0);
	const u8 flags = pt_flags && NextCharSI(si,false)
			? ScanPtFlags(si) : KPFL_M_DEFAULT;
	CheckEolSI(si);

	const kmp_enpt_entry_t * prev = GetListElem(ptlist,-1,&temp);
	if (!ptlist->used) // no previous point available
	    gl.line.n = 1;
	else if (  auto_fill >= 1.0
		&& gl.line.n == 1
		&& gl.used
		&& ptlist->used != gl.group[gl.used-1].first_index )
	{
	    double3 dist3;
	    dist3.x = temp.position[0] - prev->position[0];
	    dist3.y = temp.position[1] - prev->position[1];
	    dist3.z = temp.position[2] - prev->position[2];
	    const double dist = Length3(dist3);
	    if ( dist > auto_fill )
	    {
		gl.line.n = (int)trunc(dist/auto_fill) + 1;
		PRINT("DIST=%8.2f, FILL=%8.2f => N=%2u\n",dist,auto_fill,gl.line.n);
	    }
	}

	uint p;
	for ( p = 1; p <= gl.line.n; p++ )
	{
	    AddPointGL(&gl,ptlist->used);
	    if ( pt_flags && ptlist->used < KMP_MAX_POINT )
		pt_flags[ptlist->used] = flags;

	    kmp_enpt_entry_t * entry = AppendList(ptlist);
	    memcpy(entry,&temp,sizeof(*entry));
	    CalcLine(&gl.line,entry->position,prev->position,temp.position,4,p);
	    if ( flags & KPFL_FALL )
		CalcKclFall(&kfall,entry->position);
	}
	InitializeLine(&gl.line);
    }

    TermGL(&gl,si,kmp,sect_ph,gname,rotate_group_links);
    if (gname)
    {
	if (auto_group_name)
	    RenameGroup(kmp,sect_ph,gname);

     #if HAVE_PRINT0
	{
	    uint i;
	    for ( i = 0; i < gl.used; i++ )
		PRINT("GROUP %3u -> %s\n",i,gname[i]);
	}
     #endif
    }

    switch (sect_pt)
    {
	case KMP_ENPT:
	    kmp->enpt_mask[0] = mask1;
	    kmp->enpt_mask[1] = mask2;
	    AssignRouteInfo(&rtinfo,&kmp->rtobj_enpt);
	    break;

	case KMP_ITPT:
	    kmp->itpt_mask[0] = mask1;
	    kmp->itpt_mask[1] = mask2;
	    AssignRouteInfo(&rtinfo,&kmp->rtobj_itpt);
	    break;

	default:
	    break;
    }

    return err;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextKTPT
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextKTPT()\n");

    kcl_fall_t kfall;
    ResetKclFall(&kfall);

    ScanParam_t ptab[] =
    {
	{ "KTPT-HEAD-VALUE",	SPM_U16,    kmp->value+KMP_KTPT },
	KCL_FALL_VARS(kmp,kfall)
	{0}
    };

    List_t *dlist = kmp->dlist + KMP_KTPT;
    dlist->used = 0;

    #if HAVE_PRINT0
	si->debug++;
    #endif

    for(;;)
    {
     #if HAVE_PRINT0
	{
	    ccp eol = FindNextLineFeedSI(si,false);
	    PRINT("##LINE %4u: %.*s\n",si->line,(int)(eol-si->ptr),si->ptr);
	}
     #endif

	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

     #if HAVE_PRINT0
	{
	    ccp eol = FindNextLineFeedSI(si,false);
	    PRINT("##LINE %4u: %.*s\n",si->line,(int)(eol-si->ptr),si->ptr);
	}
     #endif

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    CheckKclFall(&kfall);
	    continue;
	}

	u16 idx;
	if (ScanU16SI(si,&idx,1,0))
	    break;

	kmp_ktpt_entry_t * entry = AppendList(dlist);
	ScanFloatV3SI(si,entry->position,2);
	ScanU16SI(si,(u16*)&entry->player_index,1,0);
	ScanU16SI(si,(u16*)&entry->unknown,1,si->cur_file->szs_modifier);
	CalcKclFall(&kfall,entry->position);
	CheckEolSI(si);
    }
    CheckLevelSI(si);

 #if HAVE_PRINT0
    si->debug--;
 #endif

    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextCKPT
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextCKPT()\n");

    int auto_mode = 0, auto_next = 0, auto_respawn = 0;
    int auto_group_name = 0, rotate_group_links = 0;
    memset(kmp->ckpt_auto_respawn,0,sizeof(kmp->ckpt_auto_respawn));

    ScanParam_t ptab[] =
    {
	{ "AUTO-MODE",			SPM_INT,    &auto_mode },
	{ "AUTO-NEXT",			SPM_INT,    &auto_next },
	{ "AUTO-GROUP-NAME",		SPM_INT,    &auto_group_name },
	{ "AUTO-RESPAWN",		SPM_INT,    &auto_respawn },
	{ "ROTATE-GROUP-LINKS",		SPM_INT,    &rotate_group_links },
	{ "ROUTE-OBJECT-MODE",		SPM_INT,    &kmp->ckpt_object_mode },
	{ "ROUTE-OBJECT1",		SPM_INT,    &kmp->ckpt_object1 },
	{ "ROUTE-OBJECT2",		SPM_INT,    &kmp->ckpt_object2 },
	{ "ROUTE-OBJECT1-SCALE",	SPM_VAR,    &kmp->ckpt_object1_scale },
	{ "ROUTE-OBJECT2-SCALE",	SPM_VAR,    &kmp->ckpt_object2_scale },
	{ "ROUTE-OBJECT1-BASE",		SPM_FLOAT,  &kmp->ckpt_object1_base },
	{ "ROUTE-OBJECT2-BASE",		SPM_FLOAT,  &kmp->ckpt_object2_base },
	{ "CKPT-HEAD-VALUE",		SPM_U16,    kmp->value+KMP_CKPT },
	{ "CKPH-HEAD-VALUE",		SPM_U16,    kmp->value+KMP_CKPH },
	{0}
    };

    kmp_group_list_t gl;
    InitializeGL(&gl,0);
    List_t *ptlist = kmp->dlist + KMP_CKPT;
    ptlist->used = 0;
    uint last_mode = 0;
    float max_x = 0.0, max_y = 0.0;

    const uint max_n = 0x100;
    float ftemp[max_n+1][4]; // x,y,unused,width/2

    enumError err = ERR_OK;
    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    continue;
	}

	if ( ch == '$' )
	{
	    AddGroupGL(&gl,si,'C',ptlist->used,2);
	    continue;
	}

	u16 idx;
	if (ScanU16SI(si,&idx,1,0))
	{
	    err = ERR_WARNING;
	    break;
	}

	kmp_ckpt_entry_t temp;
	ScanFloatV2SI(si,temp.left,2);

	if ( auto_respawn > 0 )
	{
	    si->no_warn++;
	    ScanU8SI(si,&temp.respawn,1,0);
	    si->no_warn--;
	}
	else
	    ScanU8SI(si,&temp.respawn,1,0);

	if ( si->cur_file->szs_modifier > 0 )
	{
	    ScanU8SI(si,&temp.mode,1,1);
	    ScanU8SI(si,&temp.prev,2,0);
	}
	else
	    ScanU8SI(si,&temp.mode,3,0);
	CheckEolSI(si);

	kmp_ckpt_entry_t * prev = GetListElem(ptlist,-1,&temp);
	if (!ptlist->used) // no previous point available
	    gl.line.n = 1;
	else if ( gl.line.n > max_n )
	    gl.line.n = max_n;

	float *p1 = 0, *p2 = 0;
	if ( gl.line.n > 1 )
	{
	    if ( gl.line.line_mode == LM_BORDER )
	    {
		p1 = ftemp[0];
		memcpy(p1,prev->left,4*sizeof(float));
		p2 = ftemp[gl.line.n];
		memcpy(p2,temp.left,4*sizeof(float));
	    }
	    else
	    {
		p1 = ftemp[0];
		p1[0] = ( prev->left[0] + prev->right[0] ) / 2;
		p1[1] = ( prev->left[1] + prev->right[1] ) / 2;
		p1[2] = 0.0;
		double dx = prev->left[0] - prev->right[0];
		double dy = prev->left[1] - prev->right[1];
		p1[3] = sqrt ( dx*dx + dy*dy ) / 2;

		p2 = ftemp[gl.line.n];
		p2[0] = ( temp.left[0] + temp.right[0] ) / 2;
		p2[1] = ( temp.left[1] + temp.right[1] ) / 2;
		p2[2] = 0.0;
		dx = temp.left[0] - temp.right[0];
		dy = temp.left[1] - temp.right[1];
		p2[3] = sqrt ( dx*dx + dy*dy ) / 2;

		if ( gl.line.line_mode == LM_BEZIER && !gl.line.n_pt )
		{
		    double angle1 = CalcDirection2F(prev->left,prev->right);
		    double angle2 = CalcDirection2F(temp.left,temp.right);
		    PRINT("BINGO 2: %7.2f %7.2f -> %7.2f %7.2f\n",
				angle1, angle2,
				fmod(angle1+M_PI,M_PI), fmod(angle2+M_PI,M_PI) );

		    double dx = p1[0] - p2[0];
		    double dy = p1[1] - p2[1];
		    double range = sqrt( dx*dx + dy*dy ) * 0.707; // sin(pi/4)

		    gl.line.pt[0][0] = p1[0] + cos(angle1) * range;
		    gl.line.pt[0][1] = p1[1] - sin(angle1) * range;
		    gl.line.pt[1][0] = p2[0] - cos(angle2) * range;
		    gl.line.pt[1][1] = p2[1] + sin(angle2) * range;
		    gl.line.n_pt = 2;
		    PRINT("\n-> %11.3f %11.3f\n"
			    "   %11.3f %11.3f\n"
			    "   %11.3f %11.3f\n"
			    "   %11.3f %11.3f\n",
			  p1[0], p1[1],
			  gl.line.pt[0][0], gl.line.pt[0][1],
			  gl.line.pt[1][0], gl.line.pt[1][1],
			  p2[0], p2[1] );
		}
	    }
	}

	uint p;
	for ( p = 1; p <= gl.line.n; p++ )
	{
	    const uint grp_idx = AddPointGL(&gl,ptlist->used);
	    kmp_ckpt_entry_t * entry = AppendList(ptlist);
	    memcpy(entry,&temp,sizeof(*entry));
	    if ( p < gl.line.n )
		CalcLine(&gl.line,ftemp[p],p1,p2,4,p);

	    switch (auto_mode)
	    {
		//case KMP_AM_NORM:	// normalize modes
		//    --> delayed
		//    break;

		case KMP_AM_RENUMBER:	// renumber  modes
		    if (!entry->mode)
			last_mode = 0;
		    else if (!IS_M1(entry->mode))
			entry->mode = ++last_mode;
		    break;

		case KMP_AM_1_LAP:	// define modes as 1 lap race
		    entry->mode = ptlist->used < 7 ? ptlist->used % 2 - 1 : -1;
		    break;

		case KMP_AM_SHORT:	// define modes to get a very short race
		    entry->mode = ptlist->used % 2 - 1;
		    break;

		case KMP_AM_UNLIMIT:	// remove lap counter for unlimited race
		    {
			static const u8 tab[] = { 0, 3,0,1,2 };
			entry->mode = ptlist->used < sizeof(tab)/sizeof(*tab)
					? tab[ptlist->used] : -1;

			if ( max_x < entry->left[0] )
			     max_x = entry->left[0];
			if ( max_x < entry->right[0] )
			     max_x = entry->right[0];

			if ( max_y < entry->left[1] )
			     max_y = entry->left[1];
			if ( max_y < entry->right[1] )
			     max_y = entry->right[1];
		    }
		    break;

		default:
		    if ( !IS_M1(entry->mode) )
			last_mode = entry->mode;
		    break;
	    }

	    if (auto_next)
	    {
		entry->next = ~0;
		entry->prev = grp_idx ? ptlist->used - 2 : ~0;
		if ( grp_idx > 0 && IS_M1(entry[-1].next) )
		    entry[-1].next = ptlist->used - 1;
	    }

	    if ( auto_respawn > 0 )
	    {
		const uint idx = ptlist->used - 1;
		if ( idx < KMP_MAX_POINT )
		    kmp->ckpt_auto_respawn[idx] = 1;
	    }
	}

	if ( gl.line.line_mode == LM_BORDER )
	{
	    // now copy the calculated values
	    for ( p = 1; p < gl.line.n; p++ )
	    {
		kmp_ckpt_entry_t * entry = prev + p;
		memcpy(entry,ftemp[p],4*sizeof(float));
	    }
	}
	else
	{
	    // now calculate angle and real positions
	    for ( p = 1; p < gl.line.n; p++ )
	    {
		kmp_ckpt_entry_t * entry = prev + p;
		float (*f)[4] = ftemp + p-1;
		double angle1 = CalcDirection2F(f[0],f[1]);
		double angle2 = CalcDirection2F(f[1],f[2]);
		noPRINT("%2u: %6.0f %6.0f %6.0f | %6.0f %6.0f %6.0f"
			" | %6.0f %6.0f %6.0f | %7.2f %7.2f\n",
		    p,
		    f[0][0],f[0][1],f[0][3], f[1][0],f[1][1],f[1][3], f[2][0],f[2][1],f[2][3],
		    angle1 * 180.0 / M_PI, angle2  * 180.0 / M_PI );

		angle1 = ( angle1 + angle2 ) / 2;
		double d = cos(angle1) * f[1][3];
		entry->left [0] = f[1][0] + d;
		entry->right[0] = f[1][0] - d;
		d = sin(angle1) * f[1][3];
		entry->left [1] = f[1][1] - d;
		entry->right[1] = f[1][1] + d;
	    }
	}
	InitializeLine(&gl.line);
    }

    if ( auto_mode == KMP_AM_NORM )
    {
	uint min_search = 1, next_val = 1;
	kmp_ckpt_entry_t *pt_end = (kmp_ckpt_entry_t*)ptlist->list + ptlist->used;
	for(;;)
	{
	    uint min_found = 0xff;
	    kmp_ckpt_entry_t *pt;
	    for ( pt = (kmp_ckpt_entry_t*)ptlist->list; pt < pt_end; pt++ )
		if ( pt->mode >= min_search && pt->mode < min_found )
		    min_found = pt->mode;

	    if ( min_found == 0xff )
		break;

	    for ( pt = (kmp_ckpt_entry_t*)ptlist->list; pt < pt_end; pt++ )
		if ( pt->mode == min_found )
		    pt->mode = next_val;

	    next_val++;
	    min_search = min_found + 1;
	}
    }
    else if ( auto_mode == KMP_AM_UNLIMIT && ptlist->used > 3 )
    {
	kmp_ckpt_entry_t *pt = ((kmp_ckpt_entry_t*)ptlist->list) + 3;
	pt->left[0]  = max_x + 20000.0;
	pt->left[1]  = max_y + 20000.0;
	pt->right[0] = max_x + 22000.0;
	pt->right[1] = max_y + 22000.0;
    }
    else if ( auto_mode == KMP_AM_SHORT || auto_mode == KMP_AM_1_LAP )
	kmp->multi_ckpt_mode_0_ok = true;

    TermGL(&gl,si,kmp,KMP_CKPH,kmp->ckph_name,rotate_group_links);
    if (auto_group_name)
	RenameGroup(kmp,KMP_CKPH,kmp->ckph_name);

    return err;
}

///////////////////////////////////////////////////////////////////////////////

static kmp_gobj_entry_t * AppendGOBJ ( kmp_t * kmp, u16 obj_id )
{
    kmp_gobj_entry_t *go = AppendList(kmp->dlist + KMP_GOBJ);
    if (go)
    {
	go->obj_id = obj_id;
	go->route_id = M1(go->route_id);
	go->presence_flags = 0x3f;
	go->scale[0] = go->scale[1] = go->scale[2] = 1.0;
    }
    return go;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextGOBJ
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextGOBJ()\n");

    kcl_fall_t kfall;
    ResetKclFall(&kfall);

    int auto_object_name = 0, sort_objects = 0;
    uint auto_enemy_item = 0, auto_enemy_index = 32;
    ScanParam_t ptab[] =
    {
	{ "GOBJ-HEAD-VALUE",	SPM_U16,  kmp->value+KMP_GOBJ },
	{ "AUTO-ENEMY-ITEM",	SPM_UINT, &auto_enemy_item },
	{ "AUTO-OBJECT-NAME",	SPM_INT,  &auto_object_name },
	{ "SORT-OBJECTS",	SPM_INT,  &sort_objects },
	KCL_FALL_VARS(kmp,kfall)
	{0}
    };

    List_t *dlist = kmp->dlist + KMP_GOBJ;
    dlist->used = 0;

    bool ignore_itembox
	=  kmp->rtobj_enpt.obj == GOBJ_ITEMBOX
	|| kmp->rtobj_itpt.obj == GOBJ_ITEMBOX
	|| kmp->rtobj_jgpt.obj == GOBJ_ITEMBOX
	|| kmp->rtobj_cnpt.obj == GOBJ_ITEMBOX
	|| kmp->rtobj_mspt.obj == GOBJ_ITEMBOX;
    if (!ignore_itembox)
    {
	uint pi;
	for ( pi = 0; pi < kmp->n_rtobj_poti; pi++ )
	    if ( kmp->rtobj_poti[pi].rtobj.obj == GOBJ_ITEMBOX )
	    {
		ignore_itembox = true;
		break;
	    }
    }

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    CheckKclFall(&kfall);
	    continue;
	}

	enumError err;
	ScanIndexSI(si,&err,dlist->used,kmp->index+KMP_GOBJ,0);
	if (err)
	    break;

	kmp_gobj_entry_t * entry = AppendList(dlist);

	if ( si->cur_file->szs_modifier > 0 )
	{
	    ScanU16SI(si,&entry->obj_id,1,1);
	    ScanU16SI(si,&entry->unknown,1,0);
	    ScanFloatV3SI(si,entry->position,3);
	    ScanBE32SI_swap(si,(u32*)&entry->route_id,5,1);
	}
	else
	{
	    ScanU16SI(si,&entry->obj_id,1,0);
	     ScanFloatV3SI(si,entry->position,1);
	     ScanU16SI(si,entry->setting,4,0);
	     ScanU16SI(si,&entry->route_id,1,0);
	    ScanU16SI(si,&entry->unknown,1,0);
	     ScanFloatV3SI(si,entry->rotation,1);
	     ScanU16SI(si,entry->setting+4,4,0);
	     ScanU16SI(si,&entry->presence_flags,1,0);
	    ScanFloatV3SI(si,entry->scale,1);
	}
	CalcKclFall(&kfall,entry->position);
	CheckEolSI(si);

	if ( ignore_itembox
		&& entry->obj_id < N_KMP_GOBJ
		&& ObjectInfo[entry->obj_id].flags & OBF_ITEMBOX )
	{
	    dlist->used--;
	}
	else if ( auto_enemy_item
		&& entry->obj_id == GOBJ_ITEMBOX
		&& !entry->setting[2] )
	{
	    for(;;)
	    {
		if ( auto_enemy_index > 32 )
		     auto_enemy_index = 0;
		if ( auto_enemy_item & ( 1 << auto_enemy_index++ ) )
		    break;
	    }
	    entry->setting[2] = auto_enemy_index - 1;
	}
    }
    CheckLevelSI(si);

    if (SortGOBJ(kmp,sort_objects))
	auto_object_name = 1;

    if ( auto_object_name > 0 )
    {
	ResetIL(kmp->index+KMP_GOBJ);
	FillIL(kmp->index+KMP_GOBJ,dlist->used);
    }

    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextAREA
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextAREA()\n");

    ScanParam_t ptab[] =
    {
	{ "AREA-HEAD-VALUE", SPM_U16, kmp->value+KMP_AREA },
	{0}
    };

    List_t *dlist = kmp->dlist + KMP_AREA;
    dlist->used = 0;

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    continue;
	}

	enumError err;
	ScanIndexSI(si,&err,dlist->used,kmp->index+KMP_AREA,2);
	if (err)
	    break;

	kmp_area_entry_t * entry = AppendList(dlist);

	if ( si->cur_file->szs_modifier > 0 )
	{
	    ScanBE32SI(si,(u32*)&entry->mode,1,1);
	    ScanFloatV3SI(si,entry->position,3);
	    ScanBE32SI_swap(si,(u32*)entry->setting,2,1);
	}
	else if ( si->cur_file->revision < 2956 )
	{
	    // 2011-08-26, v0.17a, r2956 -> new table layout

	    ScanBE16SI(si,(u16*)&entry->mode,2,0);
	     ScanFloatV3SI(si,entry->position,1);
	     ScanU16SI(si,entry->setting,4,0);
	    ScanFloatV3SI(si,entry->rotation,1);
	    ScanFloatV3SI(si,entry->scale,1);
	}
	else
	{
	    ScanU8SI(si,&entry->mode,2,0);
	     ScanFloatV3SI(si,entry->position,1);
	     ScanU16SI(si,entry->setting,4,0);
	    ScanU8SI(si,&entry->dest_id,2,0);
	     ScanFloatV3SI(si,entry->rotation,1);
	    ScanFloatV3SI(si,entry->scale,1);
	}
	CheckEolSI(si);
    }
    CheckLevelSI(si);

    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextCAME
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextCAME()\n");

    u8 *value = (u8*)(kmp->value + KMP_CAME);
    ScanParam_t ptab[] =
    {
	{ "CAME-HEAD-VALUE", SPM_U16, (u16*)value },
     #ifdef LITTLE_ENDIAN
	{ "OPENING-INDEX",   SPM_U8, value+1 },
	{ "SELECTION-INDEX", SPM_U8, value+0 },
     #else
	{ "OPENING-INDEX",   SPM_U8, value+0 },
	{ "SELECTION-INDEX", SPM_U8, value+1 },
     #endif
	{0}
    };

    List_t *dlist = kmp->dlist + KMP_CAME;
    dlist->used = 0;

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    continue;
	}

	enumError err;
	ScanIndexSI(si,&err,dlist->used,kmp->index+KMP_CAME,2);
	if (err)
	    break;

	kmp_came_entry_t * entry = AppendList(dlist);

	uint zoom_speed = 0, viewpt_speed = 0;

	if ( si->cur_file->szs_modifier > 0 )
	{
	    ScanBE32SI(si,(u32*)&entry->type,1,1);
	    ScanBE32SI_swap(si,(u32*)&entry->came_speed,2,1);
	    ScanFloatV3SI(si,entry->position,2);
	    ScanFloatSI(si,&entry->zoom_begin,2);
	    ScanFloatV3SI(si,entry->viewpt_begin,2);
	    ScanFloatSI(si,&entry->time,1);
	    zoom_speed = entry->zoom_speed;
	    viewpt_speed = entry->viewpt_speed;
	}
	else if ( si->cur_file->revision < 2955 )
	{
	    // 2011-08-26, v0.17a, r2956 -> new table layout

	    ScanU8SI(si,&entry->type,4,0);
	     ScanU16SI(si,&entry->came_speed,1,0);
	     ScanFloatV3SI(si,entry->position,1);
	     ScanFloatSI(si,&entry->time,1);
	    ScanU32SI(si,&zoom_speed,1,0);
	     ScanFloatV3SI(si,entry->rotation,1);
	    ScanU32SI(si,&viewpt_speed,1,0);
	     ScanFloatSI(si,&entry->zoom_begin,2);
	    ScanU16SI(si,&entry->unknown_0x0a,1,0);
	     ScanFloatV3SI(si,entry->viewpt_begin,2);
	}
	else
	{
	    ScanU8SI(si,&entry->type,1,0);
	     ScanU16SI(si,&entry->came_speed,1,0);
	     ScanFloatV3SI(si,entry->position,1);
	     ScanFloatSI(si,&entry->zoom_begin,1);
	    ScanU8SI(si,&entry->next,1,0);
	     ScanU32SI(si,&zoom_speed,1,0);
	     ScanFloatV3SI(si,entry->rotation,1);
	     ScanFloatSI(si,&entry->zoom_end,1);
	    ScanU8SI(si,&entry->unknown_0x02,1,0);
	     ScanU32SI(si,&viewpt_speed,1,0);
	     ScanFloatV3SI(si,entry->viewpt_begin,1);
	    ScanU8SI(si,&entry->route,1,0);
	     ScanU16SI(si,&entry->unknown_0x0a,1,0);
	     ScanFloatV3SI(si,entry->viewpt_end,1);
	     ScanFloatSI(si,&entry->time,1);
	}
	CheckEolSI(si);

	const double zoom_len = fabs(entry->zoom_begin-entry->zoom_end);
	entry->zoom_speed = zoom_speed;
	if ( zoom_speed == ~(u32)0 )
	    entry->zoom_speed = entry->time >= 0.001
			? (uint)floor( 100.0 * zoom_len / entry->time )
			: 1;

	const double viewpt_len = LengthF(entry->viewpt_begin,entry->viewpt_end);
	entry->viewpt_speed = viewpt_speed;
	if ( viewpt_speed == ~(u32)0 )
	    entry->viewpt_speed = entry->time >= 0.001
			? (uint)floor( 100.0 * viewpt_len / entry->time )
			: 1;

	DefineIndexDouble(si,"CAME.TIME",entry->time);
	DefineIndexDouble(si,"ZOOM.LEN",zoom_len);
	DefineIndexDouble(si,"VIEWPT.LEN",viewpt_len);
    }
    CheckLevelSI(si);

    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextPT2
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si,		// valid data
    uint		sect		// KMP section index
)
{
    DASSERT(kmp);
    DASSERT(si);
    DASSERT(sect<KMP_N_SECT);
    TRACE("ScanTextPT2(%s)\n",kmp_section_name[sect].name1);

    Var_t *snap2_to_enpt, *snap_to_enpt, *hsnap_to_enpt;
    kmp_rtobj_t *rtobj = 0;

    Var_t var_dummy, *auto_adjust = &var_dummy;
    int   int_dummy, *auto_define = &int_dummy, *auto_remove = &int_dummy;
    float float_dummy, *auto_distance = &float_dummy;
    u8 *pt_flags = 0;

    switch (sect)
    {
	case KMP_JGPT:
	    snap2_to_enpt	= &kmp->jgpt_snap2;
	    snap_to_enpt	= &kmp->jgpt_snap;
	    hsnap_to_enpt	= &kmp->jgpt_hsnap;
	    rtobj		= &kmp->rtobj_jgpt;
	    auto_define		= &kmp->jgpt_auto_define;
	    auto_distance	= &kmp->jgpt_auto_distance;
	    auto_adjust		= &kmp->jgpt_auto_adjust;
	    auto_remove		= &kmp->jgpt_auto_remove;
	    pt_flags		=  kmp->jgpt_flags;
	    break;

	case KMP_CNPT:
	    snap2_to_enpt	= &kmp->cnpt_snap2;
	    snap_to_enpt	= &kmp->cnpt_snap;
	    hsnap_to_enpt	= &kmp->cnpt_hsnap;
	    rtobj		= &kmp->rtobj_cnpt;
	    break;

	case KMP_MSPT:
	    snap2_to_enpt	= &kmp->mspt_snap2;
	    snap_to_enpt	= &kmp->mspt_snap;
	    hsnap_to_enpt	= &kmp->mspt_hsnap;
	    rtobj		= &kmp->rtobj_mspt;
	    break;

	default:
	    var_dummy.mode	= VAR_UNSET;
	    snap2_to_enpt	= &var_dummy;
	    snap_to_enpt	= &var_dummy;
	    hsnap_to_enpt	= &var_dummy;
	    break;
    }

    rtobj_info_t rtinfo;
    ResetRouteInfo(&rtinfo,&kmp->rtobj);
    kcl_fall_t kfall;
    ResetKclFall(&kfall);
    int auto_id = 0;
    char hd_name[20];
    snprintf( hd_name, sizeof(hd_name),
			"%s-HEAD-VALUE", kmp_section_name[sect].name1 );
    ScanParam_t ptab[] =
    {
	{ hd_name,		SPM_U16,    kmp->value+sect },
	{ "AUTO-ID",		SPM_INT,    &auto_id },
	{ "SNAP2-TO-ENPT",	SPM_VAR,    snap2_to_enpt },
	{ "SNAP-TO-ENPT",	SPM_VAR,    snap_to_enpt },
	{ "HSNAP-TO-ENPT",	SPM_VAR,    hsnap_to_enpt },
	{ "AUTO-DEFINE",	SPM_INT,    auto_define },
	{ "AUTO-DISTANCE",	SPM_FLOAT,  auto_distance },
	{ "AUTO-ADJUST",	SPM_VAR,    auto_adjust },
	{ "AUTO-REMOVE",	SPM_INT,    auto_remove },
	ROUTE_OBJECT_VARS(rtinfo)
	KCL_FALL_VARS(kmp,kfall)
	{0}
    };

    List_t *dlist = kmp->dlist + sect;
    dlist->used = 0;

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    CheckRouteInfo(&rtinfo);
	    CheckKclFall(&kfall);
	    continue;
	}

	enumError err;
	ScanIndexSI(si,&err,dlist->used,kmp->index+sect,2);
	if (err)
	    break;

	kmp_jgpt_entry_t * entry = AppendList(dlist);
	ScanFloatV3SI(si,entry->position,2);
	ScanU16SI(si,(u16*)&entry->id,1,0);
	ScanU16SI(si,(u16*)&entry->effect,1,si->cur_file->szs_modifier);

	const u8 flags = pt_flags && NextCharSI(si,false)
				? ScanPtFlags(si) : KPFL_M_DEFAULT;
	if ( pt_flags && (uint)(dlist->used-1) < KMP_MAX_POINT )
		pt_flags[dlist->used-1] = flags;

	if ( flags & KPFL_FALL )
	    CalcKclFall(&kfall,entry->position);
	CheckEolSI(si);

	if (auto_id)
	    entry->id = dlist->used - 1;
    }
    CheckLevelSI(si);

    PRINT_IF( snap2_to_enpt && snap2_to_enpt->mode,
		"SNAP-TO-ENPT: %d -> jgpt=%d, cnpt=%d, mspt=%d\n",
		snap2_to_enpt->mode, kmp->jgpt_snap2.mode,
		kmp->cnpt_snap2.mode, kmp->mspt_snap2.mode );
    PRINT_IF( snap_to_enpt && snap_to_enpt->mode,
		"SNAP-TO-ENPT: %d -> jgpt=%d, cnpt=%d, mspt=%d\n",
		snap_to_enpt->mode, kmp->jgpt_snap.mode,
		kmp->cnpt_snap.mode, kmp->mspt_snap.mode );

    if (rtobj)
	AssignRouteInfo(&rtinfo,rtobj);

    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextSTGI
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextSTGI()\n");

    ScanParam_t ptab[] =
    {
	{ "STGI-HEAD-VALUE", SPM_U16, kmp->value+KMP_STGI },
	{0}
    };

    List_t *dlist = kmp->dlist + KMP_STGI;
    dlist->used = 0;

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    continue;
	}

	enumError err;
	ScanIndexSI(si,&err,dlist->used,kmp->index+KMP_STGI,2);
	if (err)
	    break;

	kmp_stgi_entry_t * entry = AppendList(dlist);
	if ( si->cur_file->revision < 3379 )
	{
	    // 2012-01-28, v0.26a, r3379
	    ScanBE16SI(si,(u16*)&entry->lap_count,6,0);
	    entry->unknown_04 = be32(&entry->unknown_04);
	}
	else
	{
	    ScanU8SI(si,&entry->lap_count,4,0);
	    ScanU32SI(si,&entry->unknown_04,1,0);
	    ScanU8SI(si,&entry->unknown_08,2,0);

	    Var_t var;
	    ScanExprSI(si,&var);
	    if ( var.mode == VAR_DOUBLE || var.mode == VAR_VECTOR )
		entry->unknown_0a = SpeedMod2u16(GetDoubleV(&var));
	    else
	    {
		u8 v2;
		ScanU8SI(si,&v2,1,0);
		entry->unknown_0a = GetIntV(&var) << 8 | v2;
	    }
	}
	CheckEolSI(si);
    }
    CheckLevelSI(si);
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextSETUP
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextSETUP()\n");

    rtobj_info_t rtinfo;
    ResetRouteInfo(&rtinfo,&kmp->rtobj);

    ScanParam_t ptab[] =
    {
	ROUTE_OBJECT_VARS(rtinfo)
	{0}
    };

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    CheckRouteInfo(&rtinfo);
	    continue;
	}

	char name[50];
	if ( !ScanNameSI(si,name,sizeof(name),true,true,0)
		|| NextCharSI(si,true) != '=' )
	{
	    CheckEolSI(si);
	    continue;
	}
	DASSERT(si->cur_file);
	si->cur_file->ptr++;
	noPRINT("### %s = %.4s...\n",name,si->cur_file->ptr);

	if (!strcmp(name,"REVISION"))
	{
	    ScanU32SI(si,(u32*)&kmp->revision,1,0);
	    si->init_revision = si->cur_file->revision = kmp->revision;
	    DefineIntVar(&si->gvar,"REVISION$SETUP",kmp->revision);
	    DefineIntVar(&si->gvar,"REVISION$ACTIVE",kmp->revision);
	}
	else if (!strcmp(name,"KCL-FALL-ADD-ROAD"))
	{
	    Var_t val;
	    if ( ScanExprSI(si,&val) <= ERR_WARNING && GetBoolV(&val) )
		kmp->enable_fall_kcl = true;
	}
	else if (!strcmp(name,"UNKNOWN-0X0C"))
	    ScanU32SI(si,&kmp->unknown_0x0c,1,0);
	else if (!strcmp(name,"INFO"))
	    ScanStringSI(si,&kmp->info);
	else
	    GotoEolSI(si);
	CheckEolSI(si);
    }
    CheckLevelSI(si);
    if (rtinfo.valid)
	AssignRouteInfo(&rtinfo,&kmp->rtobj);
    return ERR_OK;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////		   ScanTextKMP() route functions	///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct route_info_t
{
    int		index;			// index of route

    int		f_xss;
    int		f_yss;
    int		f_zss;
    int		f_scale;
    int		f_shift;
    int		f_hrotate;		// [[[hrot]] [[2do]]

    rtobj_info_t rtinfo;
    kcl_fall_t	kfall;

    double	xss[4];
    double	yss[4];
    double	zss[4];
    Var_t	scale[2];
    Var_t	shift;
    Var_t	hrotate[2];		// [[[hrot]] [[2do]]
    int		reverse_order;

    double	route_time;		// >0.0: recalc setting 1

    int		show_route;		// >0: enable itemboxes for this route

} route_info_t;

///////////////////////////////////////////////////////////////////////////////

static void CloseRoute
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si,		// valid data
    const kmp_poti_group_t
			* grp,		// group to close
    route_info_t	* ri		// route info data
)
{
    DASSERT(kmp);
    DASSERT(grp);
    DASSERT(ri);

    const uint np = grp->n_point;
    DASSERT( np <= kmp->poti_point.used );

    kmp_poti_point_t *pp0 = (kmp_poti_point_t*)kmp->poti_point.list
			  + kmp->poti_point.used - np;

    //--- xss/yss/zss

    if ( ri->f_xss && fabs(ri->xss[0]-ri->xss[2]) < 1e-6 )
	ri->f_xss = 0;

    if ( ri->f_yss && fabs(ri->yss[0]-ri->yss[2]) < 1e-6 )
	ri->f_yss = 0;

    if ( ri->f_zss && fabs(ri->zss[0]-ri->zss[2]) < 1e-6 )
	ri->f_zss = 0;

    if ( ri->f_xss || ri->f_yss || ri->f_zss )
    {
	ri->f_shift = 1;
	ri->f_scale = 1;
	AssignVectorV3(&ri->shift,0.0,0.0,0.0);
	AssignVectorV3(ri->scale,1.0,1.0,1.0);

	if (ri->f_xss)
	{
	    double *d = ri->xss;
	    PRINT("XSS[%d]:     %11.3f %11.3f | %11.3f %11.3f\n",
		ri->f_xss, d[0], d[1], d[2], d[3] );
	    const double scale = ( d[3] - d[1] ) / ( d[2] - d[0] );
	    ri->scale[0].x = scale;
	    ri->shift.x = d[1] - d[0] * scale;
	}

	if (ri->f_yss)
	{
	    double *d = ri->yss;
	    PRINT("YSS[%d]:     %11.3f %11.3f | %11.3f %11.3f\n",
		ri->f_yss, d[0], d[1], d[2], d[3] );
	    const double scale = ( d[3] - d[1] ) / ( d[2] - d[0] );
	    ri->scale[0].y = scale;
	    ri->shift.y = d[1] - d[0] * scale;
	}

	if (ri->f_zss)
	{
	    double *d = ri->zss;
	    PRINT("ZSS[%d]:     %11.3f %11.3f | %11.3f %11.3f\n",
		ri->f_zss, d[0], d[1], d[2], d[3] );
	    const double scale = ( d[3] - d[1] ) / ( d[2] - d[0] );
	    ri->scale[0].z = scale;
	    ri->shift.z = d[1] - d[0] * scale;
	}
    }


    //--- scale

    if ( ri->f_scale > 1 )
    {
	Var_t *d = ri->scale;
	ToVectorV(d+0);
	ToVectorV(d+1);
	PRINT("SCALE[%d]:   %11.3f %11.3f %11.3f | %11.3f %11.3f %11.3f\n",
		ri->f_scale, d[0].x, d[0].y, d[0].z,  d[1].x, d[1].y, d[1].z );

	kmp_poti_point_t *pp = pp0;
	uint i;
	for ( i = 0; i < np; i++, pp++ )
	{
	    pp->position[0] = ( pp->position[0] - d[1].x ) * d[0].x + d[1].x;
	    pp->position[1] = ( pp->position[1] - d[1].y ) * d[0].y + d[1].y;
	    pp->position[2] = ( pp->position[2] - d[1].z ) * d[0].z + d[1].z;
	}
    }
    else if ( ri->f_scale > 0 )
    {
	Var_t *d = ri->scale;
	ToVectorV(d+0);
	PRINT("SCALE[%d]:   %11.3f %11.3f %11.3f\n",
		ri->f_scale, d->x, d->y, d->z );

	kmp_poti_point_t *pp = pp0;
	uint i;
	for ( i = 0; i < np; i++, pp++ )
	{
	    pp->position[0] *= d->x;
	    pp->position[1] *= d->y;
	    pp->position[2] *= d->z;
	}
    }


    //--- shift

    if ( ri->f_shift )
    {
	Var_t *d = &ri->shift;
	ToVectorV(d);
	PRINT("SHIFT[%d]:   %11.3f %11.3f %11.3f\n",
		ri->f_shift, d->x, d->y, d->z );

	kmp_poti_point_t *pp = pp0;
	uint i;
	for ( i = 0; i < np; i++, pp++ )
	{
	    pp->position[0] += d->x;
	    pp->position[1] += d->y;
	    pp->position[2] += d->z;
	}
     }


    //--- hrotate	// [[[hrot]] [[2do]]

    if ( ri->f_hrotate )
    {
	Var_t *d = ri->hrotate;
	ToVectorV(d+1);
	double rot = GetDoubleV(d++) * (M_PI/180.0);
	PRINT("HROTATE[%d]: %11.3f %11.3f %11.3f -> %6.3f\n",
		ri->f_hrotate, d->x, d->y, d->z,  rot );

	kmp_poti_point_t *pp = pp0;
	uint i;
	for ( i = 0; i < np; i++, pp++ )
	{
	    const double dx = pp->position[0] - d->x;
	    const double dz = pp->position[2] - d->z;
	    const double width = sqrt ( dx*dx + dz*dz );
	    if ( width >= 1e-6 )
	    {
		const double new_rot = atan2(dx,dz) + rot;
		pp->position[0] = d->x + sin(new_rot) * width;
		pp->position[2] = d->y + cos(new_rot) * width;
	    }
	}
    }


    //--- reverse order

    if (ri->reverse_order)
    {
	kmp_poti_point_t *pp1 = pp0;
	kmp_poti_point_t *pp2 = pp0 + np - 1;
	while ( pp1 < pp2 )
	{
	    noPRINT("P: %p %p / %zd %zd\n",pp1,pp2,pp1-pp0,pp2-pp0);
	    kmp_poti_point_t temp;
	    memcpy(&temp,pp1,  sizeof(temp));
	    memcpy(pp1++,pp2,  sizeof(*pp1));
	    memcpy(pp2--,&temp,sizeof(*pp2));
	}
    }


    //--- define variables

    uint i;
    double total_len = 0.0, total_tim = 0.0;
    kmp_poti_point_t *pp = pp0;
    for ( i = 0; i < np-1; i++, pp++ )
    {
	const double len = LengthF(pp->position,pp[1].position);
	total_len += len;
	if ( pp->speed > 0 )
	    total_tim += len / pp->speed;
    }

    DefineIndexDouble(si,"LEN",total_len);
    DefineIndexDouble(si,"TIME",total_tim);
    DefineIndexDouble(si,"TIME.SET",total_tim);


    //--- set route time

    if ( ri->route_time > 0.0 && total_tim > 0.0 )
    {
	PRINT("SET ROUTE TIME %s, N=%d+%u, time=%4.2f, len=%5.3f, set=%4.2f\n",
		last_index_name, kmp->poti_point.used - np, np,
		total_tim, total_len, ri->route_time );


	const double factor = ri->route_time / total_tim;
	double sum_src = 0.0, sum_dest = 0.0;
	pp = pp0;
	for ( i = 0; i < np-1; i++, pp++ )
	{
	    if ( pp->speed > 0 )
	    {
		const double len = LengthF(pp->position,pp[1].position);
		const double tim = len / pp->speed;
		sum_src += tim;
		const double wanted = sum_src * factor - sum_dest;
		DASSERT(wanted>=0.0);
		pp->speed = wanted > 0.0 ? floor( len / wanted + 0.5 ) : 0.0;
		if (!pp->speed)
		    pp->speed = 1;
		sum_dest += len / pp->speed;
		noPRINT("%6.2f/%4.2f - %6.2f/%4.2f - %6.2f -> %5u\n",
			sum_src/60.0, total_tim,
			sum_dest/60.0, total_tim*factor,
			wanted, pp->speed );
	    }
	}
	DefineIndexDouble(si,"TIME.SET",sum_dest);
    }


    //--- enable ruote itemboxes

    kmp_rtobj_t rtobj;
    if (AssignRouteInfo(&ri->rtinfo,&rtobj))
    {
	const uint max = sizeof(kmp->rtobj_poti)/sizeof(*kmp->rtobj_poti);
	uint i;
	for ( i = 0; i <= kmp->n_rtobj_poti; i++ )
	{
	    if ( i >= max )
	    {
		ERROR0(ERR_WARNING,
		    "Route list full -> @ROUTE-OBJECT ignored for route #%u",ri->index);
		break;
	    }

	    kmp_rtobj_list_t *rptr = kmp->rtobj_poti + i;
	    if ( i == kmp->n_rtobj_poti )
	    {
		rptr->index = ri->index;
		kmp->n_rtobj_poti++;

	    }

	    if ( rptr->index == ri->index )
	    {
		AssignRouteInfo(&ri->rtinfo,&rptr->rtobj);
		PRINT("@ROUTE-OBJECT[%u] = 0x%x -> %u\n",
			i, ri->index, rptr->rtobj.obj );
		break;
	    }
	}
    }

    memset(ri,0,sizeof(*ri));
    ResetRouteInfo(&ri->rtinfo,&kmp->rtobj);
    ResetKclFall(&ri->kfall);
}

///////////////////////////////////////////////////////////////////////////////

static enumError ScanTextPOTI
(
    kmp_t		* kmp,		// KMP data structure
    ScanInfo_t		* si		// valid data
)
{
    DASSERT(kmp);
    DASSERT(si);
    TRACE("ScanTextPOTI()\n");

    List_t *glist = kmp->dlist + KMP_POTI;
    List_t *plist = &kmp->poti_point;
    glist->used = plist->used = 0;

    Line_t line;
    InitializeLine(&line);
    kmp_poti_group_t *grp = 0;
    u8 g_smooth = 0, g_back = 0;

    int auto_route_name = 0;
    route_info_t ri = {0};
    ResetRouteInfo(&ri.rtinfo,&kmp->rtobj);
    ResetKclFall(&ri.kfall);

    ScanParam_t ptab[] =
    {
	{ "AUTO-ROUTE-NAME",    SPM_INT,    &auto_route_name },
	{ "XSS",		SPM_DOUBLE, ri.xss,     4,4, &ri.f_xss },
	{ "YSS",		SPM_DOUBLE, ri.yss,     4,4, &ri.f_yss },
	{ "ZSS",		SPM_DOUBLE, ri.zss,     4,4, &ri.f_zss },
	{ "SCALE",		SPM_VAR,    ri.scale,   1,2, &ri.f_scale },
	{ "SHIFT",		SPM_VAR,    &ri.shift,  1,1, &ri.f_shift },
// [[[hrot]] [[2do]]
	{ "HROTATE",		SPM_VAR,    ri.hrotate, 1,2, &ri.f_hrotate },
	{ "REVERSE-ORDER",	SPM_INT,    &ri.reverse_order },
	{ "SET-ROUTE-TIME",	SPM_DOUBLE, &ri.route_time },
	ROUTE_OBJECT_VARS(ri.rtinfo)
	KCL_FALL_VARS(kmp,ri.kfall)
	{0}
    };

    for(;;)
    {
	char ch = NextCharSI(si,true);
	if ( !ch || ch == '[' )
	    break;

	if ( ch == '@' )
	{
	    ScanParamSI(si,ptab);
	    CheckRouteInfo(&ri.rtinfo);
	    CheckKclFall(&ri.kfall);
	    continue;
	}

	if ( ch == '$' )
	{
	    char name[KMP_MAX_NAME_LEN+1];
	    ScanNameSI(si,name,sizeof(name),true,true,0);

	    const int is_bezier = !strcmp(name,"$BEZIER");
	    if ( is_bezier || !strcmp(name,"$LINE") )
	    {
		enumError err = ScanLine(&line,si,3,is_bezier);
		if (err)
		    return err;
		continue;
	    }

	    if ( !strcmp(name,"$COPY") )
	    {
		uint r_idx;
		if (ScanU32SI(si,&r_idx,1,0))
		    continue;

		if ( r_idx >= glist->used )
		{
		    ScanFile_t *sf = si->cur_file;
		    DASSERT(sf);
		    si->total_err++;
		    sf->line_err++;
		    ERROR0(ERR_WARNING,
				"Can't copy unknown route #%u [%s @%u]\n",
				r_idx, sf->name, sf->line );
		    GotoEolSI(si);
		}

		uint r_beg = 0, r_end = UINT_MAX;
		if (NextCharSI(si,false))
		{
		    if (ScanU32SI(si,&r_beg,1,0))
			continue;

		    if (NextCharSI(si,false))
		    {
			if (ScanU32SI(si,&r_end,1,0))
			    continue;
			r_end++;
		    }
		}

		const kmp_poti_group_t *srcgrp = GetListElem(glist,r_idx,0);
		DASSERT(srcgrp);
		if ( r_end > srcgrp->n_point )
		     r_end = srcgrp->n_point;
		PRINT("COPY ROUTE: %u,%u..%u\n",r_idx,r_beg,r_end);
		if ( r_beg >= r_end )
		    continue;

		const kmp_poti_point_t * psrc = GetListElem(plist,0,0);
		psrc += r_beg;
		while ( r_idx > 0 )
		{
		    r_idx--;
		    const kmp_poti_group_t *g = GetListElem(glist,r_idx,0);
		    DASSERT(g);
		    psrc += g->n_point;
		}
		DASSERT( psrc < (kmp_poti_point_t*)GetListElem(plist,0,0) + plist->used );

		if (!grp)
		{
		    grp = AppendList(glist);
		    grp->smooth = g_smooth;
		    grp->back   = g_back;
		}
		DASSERT(grp);
		grp->n_point += r_end - r_beg;

		for ( ; r_beg < r_end; r_beg++, psrc++ )
		{
		    kmp_poti_point_t * pdest = AppendList(plist);
		    memcpy(pdest,psrc,sizeof(*pdest));
		}
		kmp->value[KMP_POTI] = plist->used;
		continue;
	    }

	    const bool cmd_close = !strcmp(name,"$CLOSE");
	    if ( !cmd_close && strcmp(name,"$ROUTE") && strcmp(name,"$GROUP") )
	    {
		ScanFile_t *sf = si->cur_file;
		DASSERT(sf);
		si->total_err++;
		sf->line_err++;
		CheckEolSI(si);
		ERROR0(ERR_WARNING,
			"Line ignored [%s @%u]: %s ...\n",
			sf->name, sf->line, name );
		continue;
	    }

	    //--- close group

	    if (grp)
		CloseRoute(kmp,si,grp,&ri);
	    grp = 0;

	    if (cmd_close)
		continue;

	    ScanIndexSI(si,0,glist->used,kmp->index+KMP_POTI,2);

	    for(;;)
	    {
		char ch = NextCharSI(si,false);
		noPRINT("--> %02x : %.6s\n",ch,si->cur_file->ptr);
		if (!ch)
		    break;
		si->cur_file->ptr++;
		if ( ch == ':' )
		    break;
	    }
	    u16 g_temp;
	    ScanU16SI(si,&g_temp,1,0);
	    if (NextCharSI(si,false))
	    {
		g_smooth = g_temp;
		ScanU8SI(si,&g_back,1,0);
	    }
	    else
	    {
		g_smooth = g_temp >> 8;
		g_back   = (u8)g_temp;
	    }
	    noPRINT("ROUTE: smooth=%u, back=%u\n",g_smooth,g_back);
	    CheckEolSI(si);
	    continue;
	}

	if ( !grp || !grp->n_point ) // no previous point available
	    line.n = 1;
	enumError err;
	ScanIndexSI(si,&err, grp ? grp->n_point + line.n -1 : 0, 0,2);
	if (err)
	    break;

	if (!grp)
	{
	    ri.index = glist->used;
	    grp = AppendList(glist);
	    grp->smooth = g_smooth;
	    grp->back   = g_back;
	}
	DASSERT(grp);
	grp->n_point += line.n;

	kmp_poti_point_t temp;
	ScanFloatV3SI(si,temp.position,1);
	if ( si->cur_file->szs_modifier > 0 )
	    ScanBE32SI_swap(si,(u32*)&temp.speed,1,1);
	else
	    ScanU16SI(si,&temp.speed,2,0);
	CheckEolSI(si);

	const kmp_poti_point_t * prev = GetListElem(plist,-1,&temp);

	uint p;
	for ( p = 1; p <= line.n; p++ )
	{
	    kmp_poti_point_t * ppos = AppendList(plist);
	    memcpy(ppos,&temp,sizeof(*ppos));
	    CalcLine(&line,ppos->position,prev->position,temp.position,3,p);
	    CalcKclFall(&ri.kfall,ppos->position);
	}
	kmp->value[KMP_POTI] = plist->used;
	InitializeLine(&line);
    }
    if (grp)
	CloseRoute(kmp,si,grp,&ri);
    CheckLevelSI(si);

    if ( auto_route_name > 0 )
    {
	ResetIL(kmp->index+KMP_POTI);
	FillIL(kmp->index+KMP_POTI,glist->used);
    }

    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

static kmp_gobj_entry_t * AddObject
(
    kmp_t		* kmp,		// KMP data structure
    const float		* pos,		// 3D vector with position
    const float		* rot,		// NULL or a rotation 3D vector
    kmp_rtobj_t		* rtobj		// object to insert
)
{
    DASSERT(kmp);
    DASSERT(pos);
    DASSERT(rtobj);
    DASSERT( rtobj->obj > 0 && rtobj->obj < N_KMP_GOBJ );

    List_t *glist  = kmp->dlist + KMP_GOBJ;
    kmp_gobj_entry_t *gobj = AppendList(glist);
    gobj->obj_id = rtobj->obj;
    gobj->position[0] = pos[0] + rtobj->shift.x;
    gobj->position[1] = pos[1] + rtobj->shift.y;
    gobj->position[2] = pos[2] + rtobj->shift.z;
    memcpy(gobj->scale,rtobj->scale.v,sizeof(gobj->scale));
    memcpy(gobj->rotation,rtobj->rotate.v,sizeof(gobj->rotation));
    gobj->route_id = (typeof(gobj->route_id))~0;
    gobj->presence_flags = 0x3f;
    if (rot)
    {
	gobj->rotation[0] = NormDegreeF( gobj->rotation[0] + rot[0] );
	gobj->rotation[1] = NormDegreeF( gobj->rotation[1] + rot[1] );
	gobj->rotation[2] = NormDegreeF( gobj->rotation[2] + rot[2] );
    }
    return gobj;
}

///////////////////////////////////////////////////////////////////////////////

static void AddEnemyRouteObjects
(
    kmp_t		* kmp,		// KMP data structure
    uint		sect,		// KMP section index
    kmp_rtobj_t		*rtobj		// object to insert
)
{
    DASSERT(kmp);
    DASSERT(rtobj);
    DASSERT( sect == KMP_ENPT || sect == KMP_ITPT );

    if (rtobj->obj)
    {
	PRINT("AddEnemyRouteObjects(%s)\n",kmp_section_name[sect].name1);
	const uint n = kmp->dlist[sect].used;
	const kmp_enpt_entry_t * pt = (kmp_enpt_entry_t*)kmp->dlist[sect].list;

	uint i;
	for ( i = 0; i < n; i++ )
	{
	    int last = -1, link;
	    for ( link = 0; link < KMP_MAX_PH_LINK; link++ )
	    {
		int next = GetNextPointKMP(kmp,sect,i,link,false,-1,0,0,0);
		if ( next >= 0 && next != last )
		{
		    last = next;
		    float3 rot = CalcDirection3F(pt[i].position,pt[next].position);
		    AddObject(kmp,pt[i].position,rot.v,rtobj);
		}
	    }
	}
    }
}

///////////////////////////////////////////////////////////////////////////////

static void AddJugemRouteObjects
(
    kmp_t		* kmp,		// KMP data structure
    uint		sect,		// KMP section index
    kmp_rtobj_t		*rtobj		// object to insert
)
{
    DASSERT(kmp);
    DASSERT(rtobj);
    DASSERT( sect == KMP_JGPT || sect == KMP_CNPT || sect == KMP_MSPT );

    if (rtobj->obj)
    {
	PRINT("AddJugemRouteObjects(%s)\n",kmp_section_name[sect].name1);
	const uint n = kmp->dlist[sect].used;
	const kmp_jgpt_entry_t * pt = (kmp_jgpt_entry_t*)kmp->dlist[sect].list;

	uint i;
	for ( i = 0; i < n; i++, pt++ )
	    AddObject(kmp,pt->position,pt->rotation,rtobj);
    }
}

///////////////////////////////////////////////////////////////////////////////

static void AddRouteObjects
(
    kmp_t		* kmp		// KMP data structure
)
{
    DASSERT(kmp);
    PRINT("AddRouteObjects()\n");

    const uint ng = kmp->dlist[KMP_POTI].used;
    uint ilist;
    kmp_rtobj_list_t *rptr = kmp->rtobj_poti;
    for ( ilist = 0; ilist < kmp->n_rtobj_poti; ilist++, rptr++ )
    {
	int route_idx = rptr->index;
	if ( route_idx < 0 || route_idx >= ng || !rptr->rtobj.obj )
	    continue;

	PRINT("ROUTE-ITEMBOX %u/%u -> %d/%u\n",
		ilist, kmp->n_rtobj_poti, route_idx, ng );
	const kmp_poti_group_t *grp = (kmp_poti_group_t*)kmp->dlist[KMP_POTI].list;
	kmp_poti_point_t *pt = (kmp_poti_point_t*)kmp->poti_point.list;
	while ( route_idx-- > 0 )
	{
	    pt += grp->n_point;
	    grp++;
	}

	uint i;
	for ( i = 0; i < grp->n_point; i++ )
	{
	    float3 rot = CalcDirection3F( pt[i].position,
					  pt[ (i+1) % grp->n_point ].position );
	    AddObject( kmp, pt[i].position, rot.v, &rptr->rtobj );
	}
    }
}

///////////////////////////////////////////////////////////////////////////////

static void setup_scale ( Var_t *dest, const Var_t *src )
{
    DASSERT(dest);
    DASSERT(src);

    if ( src->mode == VAR_VECTOR )
	AssignVar(dest,src);
    else
	AssignVectorV3(dest,0.0,GetYDoubleV(src),0.0);
    DASSERT( dest->mode == VAR_VECTOR );

    if ( dest->x <= 0.0 )
	dest->x = RTOBJ_SCALE_X;
    else if ( dest->x < RTOBJ_SCALE_MIN )
	dest->y = RTOBJ_SCALE_MIN;

    if ( dest->y <= 0.0 )
	dest->y = RTOBJ_SCALE_Y;
    else if ( dest->y < RTOBJ_SCALE_MIN )
	dest->y = RTOBJ_SCALE_MIN;

    if ( dest->z <= 0.0 )
	dest->z = RTOBJ_SCALE_Z;
    else if ( dest->z < RTOBJ_SCALE_MIN )
	dest->y = RTOBJ_SCALE_MIN;
}

//-----------------------------------------------------------------------------

static void AddCheckpointObject
(
    kmp_t		* kmp		// KMP data structure
)
{
    DASSERT(kmp);
    PRINT("AddCheckpointObject()\n");

    struct obj_t
    {
	int    object;
	Var_t  scale;
	float  base;

    } obj1, obj2;

    obj1.object = kmp->ckpt_object1 > 0 && kmp->ckpt_object1 < N_KMP_GOBJ
			? kmp->ckpt_object1 : 0;
    setup_scale(&obj1.scale,&kmp->ckpt_object1_scale);
    obj1.base = kmp->ckpt_object1_base;

    obj2.object = kmp->ckpt_object2 > 0 && kmp->ckpt_object2 < N_KMP_GOBJ
			? kmp->ckpt_object2 : 0;
    setup_scale(&obj2.scale,&kmp->ckpt_object2_scale);
    obj2.base = kmp->ckpt_object2_base;

    const struct obj_t *obj_lap_lt = 0, *obj_lap_rt = 0;
    const struct obj_t *obj_man_lt = 0, *obj_man_rt = 0;
    const struct obj_t *obj_std_lt = 0, *obj_std_rt = 0;

    switch (kmp->ckpt_object_mode)
    {
	case 1:
	    obj_lap_lt = &obj2;
	    obj_lap_rt = &obj2;
	    obj_man_lt = &obj1;
	    obj_man_rt = &obj1;
	    break;

	case 2:
	    obj_lap_lt = &obj2;
	    obj_lap_rt = &obj2;
	    obj_man_lt = &obj2;
	    obj_man_rt = &obj2;
	    obj_std_lt = &obj1;
	    obj_std_rt = &obj1;
	    break;

	case 3:
	    obj_lap_lt = &obj1;
	    obj_lap_rt = &obj2;
	    obj_man_lt = &obj1;
	    obj_man_rt = &obj2;
	    obj_std_lt = &obj1;
	    obj_std_rt = &obj2;
	    break;

	default:
	    return;
    }

    opt_auto_add = true;

    List_t *glist = kmp->dlist + KMP_GOBJ;
    const kmp_ckpt_entry_t *pt = (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list;
    uint i, n = kmp->dlist[KMP_CKPT].used;
    for ( i = 0; i < n; i++, pt++ )
    {
	double direction, factor = IS_M1(pt->mode) ? 1.0 : 3.0;
	WidthAndDirection(0,&direction,true,pt->left,pt->right);

	const struct obj_t *obj;
	obj = IS_M1(pt->mode) ? obj_std_lt : pt->mode ? obj_man_lt : obj_lap_lt;
	if (obj)
	{
	    kmp_gobj_entry_t *go = AppendList(glist);
	    go->obj_id		= obj->object;
	    go->position[0]	= pt->left[0];
	    go->position[1]	= obj->base;
	    go->position[2]	= pt->left[1];
	    go->rotation[1]	= direction;
	    go->scale[0]	= obj->scale.x * factor;
	    go->scale[1]	= obj->scale.y;
	    go->scale[2]	= obj->scale.z * factor;
	    go->route_id	= (typeof(go->route_id))~0;
	    go->presence_flags	= 0x3f;
	}

	obj = IS_M1(pt->mode) ? obj_std_rt : pt->mode ? obj_man_rt : obj_lap_rt;
	if (obj)
	{
	    kmp_gobj_entry_t *go = AppendList(glist);
	    go->obj_id		= obj->object;
	    go->position[0]	= pt->right[0];
	    go->position[1]	= obj->base;
	    go->position[2]	= pt->right[1];
	    go->rotation[1]	= direction;
	    go->scale[0]	= obj->scale.x * factor;
	    go->scale[1]	= obj->scale.y;
	    go->scale[2]	= obj->scale.z * factor;
	    go->route_id	= (typeof(go->route_id))~0;
	    go->presence_flags	= 0x3f;
	}
    }
}

///////////////////////////////////////////////////////////////////////////////

static void SnapToENPT
(
    kmp_t		* kmp,		// KMP data structure
    uint		sect		// KMP section index
)
{
    const Var_t * add2;  // not NULL: enter 2D-full mode && value to add
    const Var_t	* add1;	 // not NULL: enter 1D-full mode && value to add
    const Var_t	* hadd;  // not NULL: enter horiz-mode && value to add
    const kmp_pt_flag_t *pt_flags;

    switch(sect)
    {
	case KMP_JGPT:
	    add2	= &kmp->jgpt_snap2;
	    add1	= &kmp->jgpt_snap;
	    hadd	= &kmp->jgpt_hsnap;
	    pt_flags	=  kmp->jgpt_flags;
	    break;

	case KMP_CNPT:
	    add2	= &kmp->cnpt_snap2;
	    add1	= &kmp->cnpt_snap;
	    hadd	= &kmp->cnpt_hsnap;
	    pt_flags	= 0;
	    break;

	case KMP_MSPT:
	    add2	= &kmp->mspt_snap2;
	    add1	= &kmp->mspt_snap;
	    hadd	= &kmp->mspt_hsnap;
	    pt_flags	= 0;
	    break;
	    
	default:
	    return;
    }

    const Var_t *add;
    enum { MODE2, MODE1, HMODE } mode;

    if ( add2 && add2->mode != VAR_UNSET )
    {
	mode = MODE2;
	add = add2;
    }
    else if ( add1 && add1->mode != VAR_UNSET )
    {
	mode = MODE1;
	add = add1;
    }
    else if ( hadd && hadd->mode != VAR_UNSET )
    {
	mode = HMODE;
	add = hadd;
    }
    else
	return;

    PRINT("SnapToENPT(%s) mode=%d\n",kmp_section_name[sect].name1,mode);

    const uint n_enpt = kmp->dlist[KMP_ENPT].used;
    if (!n_enpt)
	return;
    const kmp_enpt_entry_t *enpt1 = (kmp_enpt_entry_t*)kmp->dlist[KMP_ENPT].list;

    kmp_jgpt_entry_t *pt = (kmp_jgpt_entry_t*)kmp->dlist[sect].list;
    kmp_jgpt_entry_t *pt_end = pt + kmp->dlist[sect].used;

    uint j;
    for ( j = 0; pt < pt_end; j++, pt++ )
    {
	if ( pt_flags && j < KMP_MAX_POINT && !( pt_flags[j] & KPFL_SNAP ) )
	    continue;

	int i, enpt_idx = -1;
	double min_dist = 0.0;
	const kmp_enpt_entry_t *enpt;
	for ( i = 0, enpt = enpt1; i < n_enpt; i++, enpt++ )
	{
	    if ( i < KMP_MAX_POINT && !( kmp->enpt_flags[i] & KPFL_SNAP ))
		continue;

	    const double dx = pt->position[0] - enpt->position[0];
	    const double dy = pt->position[1] - enpt->position[1];
	    const double dz = pt->position[2] - enpt->position[2];
	    const double dist = dx*dx + dy*dy + dz*dz;
	    if ( enpt_idx < 0 || min_dist > dist )
	    {
		min_dist = dist;
		enpt_idx = i;
	    }
	}
	if ( enpt_idx < 0 ) 
	    break; // no usable ENPT found -> abort

	uint next_idx = GetNextPointKMP(kmp,KMP_ENPT,enpt_idx,-1,false,0,0,0,0);
	DASSERT( next_idx < n_enpt );

	enpt = enpt1 + enpt_idx;

	Var_t var;
	var.mode = VAR_VECTOR;
	var.x = enpt->position[0];
	var.y = mode == HMODE ? pt->position[1] : enpt->position[1];
	var.z = enpt->position[2];
	DASSERT(add);
	CalcExpr(&var,add,OPI_ADD,0);
	pt->position[0] = var.x;
	pt->position[1] = var.y;
	pt->position[2] = var.z;

	const kmp_enpt_entry_t *next = enpt1 + next_idx;
	if ( mode == MODE2 )
	{
	    double3 p1, p2;
	    p1.x = enpt->position[0];
	    p1.y = enpt->position[1];
	    p1.z = enpt->position[2];
	    p2.x = next->position[0];
	    p2.y = next->position[1];
	    p2.z = next->position[2];
	    double3 dir = CalcDirection3D(&p1,&p2);
	    pt->rotation[0] = dir.x;
	    pt->rotation[1] = dir.y;
	    pt->rotation[2] = dir.z;
	}
	else
	{
	    pt->rotation[0] = pt->rotation[2] = 0.0;
	    pt->rotation[1] = atan2( next->position[0] - enpt->position[0],
				 next->position[2] - enpt->position[2] ) * (180/M_PI);
	}
    }
}

///////////////////////////////////////////////////////////////////////////////

static int FindNearestJGPT2
(
    kmp_t	*kmp,		// valid KMP
    double	x,		// x-coordinate
    double	z,		// z-coordinate
    double	*res_dist2	// not NULL: store distance^2 here
)
{
    DASSERT(kmp);

    int found = -1;
    double max_dist2 = 1e30;

    const uint n_jgpt = kmp->dlist[KMP_JGPT].used;
    kmp_jgpt_entry_t *jgpt = (kmp_jgpt_entry_t*)kmp->dlist[KMP_JGPT].list;

    uint i;
    for ( i = 0; i < n_jgpt; i++, jgpt++ )
    {
	const double dx = x - jgpt->position[0];
	const double dz = z - jgpt->position[2];
	const double dist2 = dx*dx + dz*dz;
	if ( max_dist2 > dist2 )
	{
	    max_dist2 = dist2;
	    found = i;
	}
    }

    if ( res_dist2 )
	*res_dist2 = max_dist2;
    return found;
}

///////////////////////////////////////////////////////////////////////////////

static int FindNearestJGPT3
(
    kmp_t		*kmp,		// valid KMP
    const float32	*pos,		// 3d coordinates
    double		*res_dist2	// not NULL: store distance^2 here
)
{
    DASSERT(kmp);
    DASSERT(pos);

    int found = -1;
    double max_dist2 = 1e30;

    const uint n_jgpt = kmp->dlist[KMP_JGPT].used;
    kmp_jgpt_entry_t *jgpt = (kmp_jgpt_entry_t*)kmp->dlist[KMP_JGPT].list;

    uint i;
    for ( i = 0; i < n_jgpt; i++, jgpt++ )
    {
	double3 d;
	d.x = pos[0] - jgpt->position[0];
	d.y = pos[1] - jgpt->position[1];
	d.z = pos[2] - jgpt->position[2];
	const double dist2 = LengthSqare3(d);
	if ( max_dist2 > dist2 )
	{
	    max_dist2 = dist2;
	    found = i;
	}
    }

    if ( res_dist2 )
	*res_dist2 = max_dist2;
    return found;
}

///////////////////////////////////////////////////////////////////////////////

static void AutoRespawn ( kmp_t * kmp )
{
    DASSERT(kmp);


    //--- auto add respawn points

    const uint min_rm = kmp->jgpt_auto_remove > 1 ? 0 : kmp->dlist[KMP_JGPT].used;
    const int sect = kmp->jgpt_auto_define - 1;
    if ( sect == KMP_ENPT || sect == KMP_ITPT )
    {
	const uint n = kmp->dlist[sect].used;
	const kmp_enpt_entry_t *pt0 = (kmp_enpt_entry_t*)kmp->dlist[sect].list;
	const kmp_enpt_entry_t *pt = pt0;
	const u8* pt_flags = sect == KMP_ENPT ? kmp->enpt_flags : kmp->itpt_flags;
	uint i;
	for ( i = 0; i < n && kmp->dlist[KMP_JGPT].used < KMP_MAX_POINT; i++, pt++ )
	{
	    if ( i < KMP_MAX_POINT && !( pt_flags[i] & KPFL_JGPT ) )
		continue;

	    double dist2;
	    FindNearestJGPT3(kmp,pt->position,&dist2);
	    if ( dist2 > kmp->jgpt_auto_distance )
	    {
		kmp_jgpt_entry_t * entry = AppendList(kmp->dlist+KMP_JGPT);
		entry->position[0] = pt->position[0] + kmp->jgpt_auto_adjust.x;
		entry->position[1] = pt->position[1] + kmp->jgpt_auto_adjust.y;
		entry->position[2] = pt->position[2] + kmp->jgpt_auto_adjust.z;
		entry->id = kmp->dlist[KMP_JGPT].used - 1;
		entry->effect = M1(entry->effect);

		uint next_idx = GetNextPointKMP(kmp,KMP_ENPT,i,-1,false,0,0,0,0);
		DASSERT( next_idx < n );
		const kmp_enpt_entry_t *next = pt0 + next_idx;

		double3 p1, p2;
		p1.x = pt->position[0];
		p1.y = pt->position[1];
		p1.z = pt->position[2];
		p2.x = next->position[0];
		p2.y = next->position[1];
		p2.z = next->position[2];
		double3 dir = CalcDirection3D(&p1,&p2);
		entry->rotation[0] = dir.x;
		entry->rotation[1] = dir.y;
		entry->rotation[2] = dir.z;
	    }
	}
    }


    //--- auto connect check points to respawn points

    u8 jgpt_used[KMP_MAX_POINT];
    memset(jgpt_used,0,sizeof(jgpt_used));

    const uint n_ckpt	= kmp->dlist[KMP_CKPT].used < KMP_MAX_POINT
			? kmp->dlist[KMP_CKPT].used : KMP_MAX_POINT;
    kmp_ckpt_entry_t *ckpt = (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list;
    uint i;
    for ( i = 0; i < n_ckpt; i++, ckpt++ )
    {
	if ( kmp->ckpt_auto_respawn[i] )
	{
	    const double x = ( ckpt->left[0] + ckpt->right[0] ) / 2;
	    const double z = ( ckpt->left[1] + ckpt->right[1] ) / 2;
	    ckpt->respawn = FindNearestJGPT2(kmp,x,z,0);
	}
	if ( (uint)ckpt->respawn < KMP_MAX_POINT )
	    jgpt_used[ckpt->respawn] = 1;
    }


    //--- remove unused respawn points

    const uint n_jgpt = kmp->dlist[KMP_JGPT].used;
    if ( kmp->jgpt_auto_remove > 0 && n_jgpt > min_rm )
    {
	PRINT("--> JGPT: auto remove: %u..%u\n",min_rm,n_jgpt-1);

	u8 map[KMP_MAX_POINT];
	memset(map,0,sizeof(map));

	kmp_jgpt_entry_t *base = (kmp_jgpt_entry_t*)kmp->dlist[KMP_JGPT].list;
	kmp_jgpt_entry_t *src = base, *dest = base;

	uint i, new_idx = 0;
	for ( i = 0; i < n_jgpt; i++, src++ )
	{
	    if (jgpt_used[i])
	    {
		new_idx = dest - base;
		memcpy(dest,src,sizeof(*dest));
		dest++;
	    }
	    map[i] = new_idx;
	}
	kmp->dlist[KMP_JGPT].used = dest - base;
	//HEXDUMP16(0,0,map,sizeof(map));

	kmp_ckpt_entry_t *ckpt = (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list;
	for ( i = 0; i < n_ckpt; i++, ckpt++ )
	    ckpt->respawn = map[ckpt->respawn];
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////		   ScanTextKMP(): reference KCL		///////////////
///////////////////////////////////////////////////////////////////////////////

static void DefineReferenceKCLbyKMP ( kmp_t *kmp )
{
    kcl_ref_prio = 0;

    if ( kmp && kmp->enable_fall_kcl )
    {
	if (kmp->fall_kcl)
	{
	    ResetKCL(kmp->fall_kcl);
	    FREE(kmp->fall_kcl);
	    kmp->fall_kcl = 0;
	}

	DrawKCL_t dk;
	memset(&dk,0,sizeof(dk));

	const uint pflags = ( KMP_DRAW & DMD_M_PFLAGS ) >> DMD_SHIFT_PFLAGS;

	uint n = kmp->dlist[KMP_GOBJ].used;
	PRINT("\e[36;1m### DefineReferenceKCLbyKMP(), N(GOBJ) = %u\e[0m\n",n);
	const kmp_gobj_entry_t *op = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	for ( ; n > 0; n--, op++ )
	{
	    if ( op->obj_id < N_KMP_GOBJ
		&& ObjectInfo[op->obj_id].flags & OBF_ROAD
		&& ( !pflags || pflags & op->presence_flags )
	       )
	    {
		if (!kmp->fall_kcl)
		{
		    kcl_t *ref = LoadReferenceKCL();
		    PRINT("\e[36;1m### CREATE KMP, kcl_ref=%d\e[0m\n",ref!=0);
		    if (!ref)
			break;

		    kmp->fall_kcl = MALLOC(sizeof(kcl_t));
		    CopyKCL(kmp->fall_kcl,true,ref);
		    kmp->fall_kcl->silent_octree = true;
		    kmp->fall_kcl->fast = true;
		    dk.kcl = dk.draw = kcl_ref_prio = kmp->fall_kcl;
		    dk.kmp = kmp;
		    dk.detailed = dk.as_road = true;
		}
		AddObject2KCL(&dk,op);
	    }
	}
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  ScanTextKMP()			///////////////
///////////////////////////////////////////////////////////////////////////////

enumError ScanTextKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    const void		* data,		// data to scan
    uint		data_size	// size of 'data'
)
{
    PRINT("ScanTextKMP(init=%d)\n",init_kmp);
    KMP_ACTION_LOG(false,"ScanTextKMP() %s\n",kmp->fname);

    DASSERT(kmp);
    if (init_kmp)
	InitializeKMP(kmp);
    else if (kmp->fname)
	DefineAutoLoadReferenceKCL(kmp->fname,".kmp.txt");

    ScanInfo_t si;
    InitializeSI(&si,data,data_size,kmp->fname,kmp->revision);
    si.predef = SetupVarsKMP();

    enumError max_err = ERR_OK;
    bool pass2 = false;

    for(;;)
    {
	PRINT("----- SCAN SECTIONS, PASS2=%u ...\n",pass2);

	max_err = ERR_OK;
	si.kmp = kmp;
	si.no_warn = !pass2;
	si.total_err = 0;
	DefineIntVar(&si.gvar, "$PASS", pass2 ? 2 : 1 );

	for(;;)
	{
	    char ch = NextCharSI(&si,true);
	    if (!ch)
		break;

	    if ( ch != '[' )
	    {
		NextLineSI(&si,true,false);
		continue;
	    }
	    ResetLocalVarsSI(&si,kmp->revision);

	    si.cur_file->ptr++;
	    char name[20];
	    ScanNameSI(&si,name,sizeof(name),true,true,0);
	    noPRINT("--> #%04u: %s\n",si.line,name);

	    int abbrev_count;
	    const KeywordTab_t *cmd = ScanKeyword(&abbrev_count,name,kmp_section_name);
	    if ( !cmd || abbrev_count )
		continue;
	    NextLineSI(&si,false,false);
	    noPRINT("--> %-6s #%-4u |%.3s|\n",cmd->name1,si.line,si.ptr);

	    if ( cmd->id < 0 || cmd->id > KMP_N_SECT )
		continue;

	    enumError err = ERR_OK;
	    switch (cmd->id)
	    {
		case KMP_KTPT:
		    err = ScanTextKTPT(kmp,&si);
		    break;

		case KMP_ENPT:
		    err = ScanText_PT_PH( kmp, &si, KMP_ENPT, KMP_ENPH,
					kmp->enph_name, kmp->enpt_flags );
		    break;

		case KMP_ITPT:
		    err = ScanText_PT_PH( kmp, &si, KMP_ITPT, KMP_ITPH,
					kmp->itph_name,  kmp->itpt_flags );
		    break;

		case KMP_CKPT:
		    err = ScanTextCKPT(kmp,&si);
		    break;

		case KMP_GOBJ:
		    err = ScanTextGOBJ(kmp,&si);
		    break;

		case KMP_POTI:
		    err = ScanTextPOTI(kmp,&si);
		    break;

		case KMP_AREA:
		    err = ScanTextAREA(kmp,&si);
		    break;

		case KMP_CAME:
		    err = ScanTextCAME(kmp,&si);
		    break;

		case KMP_JGPT:
		    err = ScanTextPT2(kmp,&si,KMP_JGPT);
		    break;

		case KMP_CNPT:
		    err = ScanTextPT2(kmp,&si,KMP_CNPT);
		    break;

		case KMP_MSPT:
		    err = ScanTextPT2(kmp,&si,KMP_MSPT);
		    break;

		case KMP_STGI:
		    err = ScanTextSTGI(kmp,&si);
		    break;

		case KMP_N_SECT: // alias for [SETUP]
		    err = ScanTextSETUP(kmp,&si);
		    break;

		default:
		    // ignore all other section without any warnings
		    break;
	    }

	    if ( max_err < err )
		 max_err = err;
	}

	if (pass2)
	    break;

	pass2 = true;

	DefineReferenceKCLbyKMP(kmp);
	RestartSI(&si);
	ClearSectionsKMP(kmp);
    }

 #if HAVE_PRINT0
    printf("VAR DUMP/GLOBAL:\n");
    DumpVarMap(stdout,3,&si.gvar,false);
 #endif

    CheckLevelSI(&si);
    if ( max_err < ERR_WARNING && si.total_err )
	max_err = ERR_WARNING;
    PRINT("ERR(ScanTextKMP) = %u (errcount=%u)\n", max_err, si.total_err );
    ResetSI(&si);

    SnapToENPT(kmp,KMP_JGPT);
    SnapToENPT(kmp,KMP_CNPT);
    SnapToENPT(kmp,KMP_MSPT);

    AutoRespawn(kmp);

    AddEnemyRouteObjects(kmp,KMP_ENPT,&kmp->rtobj_enpt);
    AddEnemyRouteObjects(kmp,KMP_ITPT,&kmp->rtobj_itpt);
    AddJugemRouteObjects(kmp,KMP_JGPT,&kmp->rtobj_jgpt);
    AddJugemRouteObjects(kmp,KMP_CNPT,&kmp->rtobj_cnpt);
    AddJugemRouteObjects(kmp,KMP_MSPT,&kmp->rtobj_mspt);
    AddRouteObjects(kmp);
    AddCheckpointObject(kmp);

    if (opt_generic)
    {
	uint sect;
	for ( sect = 0; sect < KMP_N_SECT; sect++ )
	{
	    ResetIL(kmp->index+sect);
	    FillIL(kmp->index+sect,kmp->dlist[sect].used);
	}
    }

    DefineReferenceKCLbyKMP(0);
    return max_err;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    ScanRawKMP()		///////////////
///////////////////////////////////////////////////////////////////////////////

enumError ScanRawKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    const void		* data,		// data to scan
    uint		data_size	// size of 'data'
)
{
    DASSERT(kmp);
    if (init_kmp)
	InitializeKMP(kmp);

    if ( IsValidKMP(data,data_size,data_size,kmp->fname) >= VALID_ERROR )
    {
	return ERROR0(ERR_INVALID_DATA,
		"Invalid KMP file: %s\n"
		"Add option --kmp=force or --force to ignore some validity checks.",
		kmp->fname ? kmp->fname : "?" );
    }

    KMP_ACTION_LOG(false,"ScanRawKMP() %s\n",kmp->fname);
    kmp->fform = FF_KMP;
    kmp->unknown_0x0c = be32( data + 0x0c );
    kmp->max_scanned  = 0;


    //--- test for unknown sections

    test_unknown_sections(data,data_size,kmp->fname);


    //--- scan KTPT entries

    uint n;
    const kmp_ktpt_entry_t * ktpt = get_kmp_pointer(&n,0,kmp,data,data_size,KMP_KTPT);
    for ( ; n > 0; n--, ktpt++ )
    {
	kmp_ktpt_entry_t * entry = AppendList(kmp->dlist+KMP_KTPT);
	DASSERT(entry);
	bef4n(entry->position,ktpt->position,6);
	be16n((u16*)&entry->player_index,(u16*)&ktpt->player_index,2);
    }


    //--- scan ENPT + ENPH entries

    scan_kmp_pt(kmp,data,data_size,KMP_ENPT);
    scan_kmp_ph(kmp,data,data_size,KMP_ENPH,kmp->enph_name);


    //--- scan ITPT + ITPH entries

    scan_kmp_pt(kmp,data,data_size,KMP_ITPT);
    scan_kmp_ph(kmp,data,data_size,KMP_ITPH,kmp->itph_name);


    //--- scan CKPT entries

    const kmp_ckpt_entry_t * ckpt = get_kmp_pointer(&n,0,kmp,data,data_size,KMP_CKPT);
    for ( ; n > 0; n--, ckpt++ )
    {
	kmp_ckpt_entry_t * entry = AppendList(kmp->dlist+KMP_CKPT);
	DASSERT(entry);
	memcpy(entry,ckpt,sizeof(*entry));
	bef4n(entry->left,ckpt->left,4);
    }


    //--- scan CKPH entries

    scan_kmp_ph(kmp,data,data_size,KMP_CKPH,kmp->ckph_name);


    //--- scan GOBJ entries

    const kmp_gobj_entry_t * gobj = get_kmp_pointer(&n,0,kmp,data,data_size,KMP_GOBJ);
    for ( ; n > 0; n--, gobj++ )
    {
	kmp_gobj_entry_t * entry = AppendList(kmp->dlist+KMP_GOBJ);
	DASSERT(entry);
	be16n((u16*)&entry->obj_id,(u16*)&gobj->obj_id,2);
	bef4n(entry->position,gobj->position,9);
	be16n(&entry->route_id,&gobj->route_id,10);
    }


    //--- scan POTI entries

    uint n_pt;
    const kmp_poti_group_t * pgroup = get_kmp_pointer(&n,&n_pt,kmp,data,data_size,KMP_POTI);
    while ( n_pt > 0 && n-- > 0 )
    {
	kmp_poti_group_t * group = AppendList(kmp->dlist+KMP_POTI);
	DASSERT(group);
	group->smooth	= pgroup->smooth;
	group->back	= pgroup->back;
	group->n_point	= be16((u16*)&pgroup->n_point);
	if ( group->n_point > n_pt )
	    group->n_point = n_pt;
	n_pt -= group->n_point;

	const kmp_poti_point_t * point = (kmp_poti_point_t*)(pgroup+1);
	uint count;
	for ( count = group->n_point; count > 0; count--, point++ )
	{
	    kmp_poti_point_t * entry = AppendList(&kmp->poti_point);
	    DASSERT(entry);
	    bef4n(entry->position,point->position,3);
	    be16n(&entry->speed,&point->speed,2);
	}
	pgroup = (kmp_poti_group_t*)point;
    }
    if ( kmp->max_scanned < (u8*)pgroup )
	 kmp->max_scanned = (u8*)pgroup;


    //--- scan AREA entries

    const kmp_area_entry_t * area = get_kmp_pointer(&n,0,kmp,data,data_size,KMP_AREA);
    for ( ; n > 0; n--, area++ )
    {
	kmp_area_entry_t * entry = AppendList(kmp->dlist+KMP_AREA);
	DASSERT(entry);
	memcpy(entry,area,sizeof(*entry));
	bef4n(entry->position,area->position,9);
	be16n(entry->setting,area->setting,4);
    }


    //--- scan CAME entries

    const kmp_came_entry_t * came = get_kmp_pointer(&n,0,kmp,data,data_size,KMP_CAME);
    for ( ; n > 0; n--, came++ )
    {
	kmp_came_entry_t * entry = AppendList(kmp->dlist+KMP_CAME);
	DASSERT(entry);
	memcpy(entry,came,sizeof(*entry));
	be16n(&entry->came_speed,&came->came_speed,4);
	bef4n(entry->position,came->position,15);
    }


    //--- scan JGPT + CNPT + MSPT entries

    scan_kmp_pt2(kmp,data,data_size,KMP_JGPT);
    scan_kmp_pt2(kmp,data,data_size,KMP_CNPT);
    scan_kmp_pt2(kmp,data,data_size,KMP_MSPT);


    //--- scan STGI entries

    const kmp_stgi_entry_t * stgi = get_kmp_pointer(&n,0,kmp,data,data_size,KMP_STGI);
    for ( ; n > 0; n--, stgi++ )
    {
	kmp_stgi_entry_t * entry = AppendList(kmp->dlist+KMP_STGI);
	DASSERT(entry);
	memcpy(entry,stgi,sizeof(*entry));
	entry->unknown_04 = be32(&stgi->unknown_04);
	entry->unknown_0a = be16(&stgi->unknown_0a);
    }


    //--- info string

    if (kmp->max_scanned)
    {
	PRINT("SCANNED: %zx/%x\n", kmp->max_scanned - (u8*)data, data_size );
	uint info_off = kmp->max_scanned - (u8*)data;
	if ( info_off < data_size )
	{
	    ccp src = (ccp)kmp->max_scanned;
	    const int info_len = strlen(src);
	    if ( info_off + info_len < data_size )
	    {
		FreeString(kmp->info);
		kmp->info = MALLOC(info_len+1);
		char *dest = kmp->info;
		for(;;)
		{
		    const uchar ch = *src++;
		    if (!ch)
			break;
		    *dest++ = ch == '\r' || ch == '\n'
				? ' '
				: ch == '"'
					? '\''
					: ch;
		}
		*dest = 0;
		DASSERT( dest == kmp->info + info_len );
		PRINT("INFO FOUND: |%s|\n",kmp->info);
	    }
	}
    }


    //--- term

    kmp->max_scanned = 0;
    return ERR_OK;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			KMP: scan and load		///////////////
///////////////////////////////////////////////////////////////////////////////

enumError ScanKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    const void		* data,		// data to scan
    uint		data_size,	// size of 'data'
    CheckMode_t		mode		// not NULL: call CheckKMP(mode)
)
{
    DASSERT(kmp);
    PRINT("ScanKMP(mode=%x) size=%u\n",mode,data_size);
    //HEXDUMP16(0,0,data,16);

    enumError err;
    switch (GetByMagicFF(data,data_size,data_size))
    {
	case FF_KMP:
	    kmp->fform = FF_KMP;
	    err = ScanRawKMP(kmp,init_kmp,data,data_size);
	    break;

	case FF_KMP_TXT:
	    kmp->fform = FF_KMP_TXT;
	    err =  ScanTextKMP(kmp,init_kmp,data,data_size);
	    break;

	default:
	    if (init_kmp)
		InitializeKMP(kmp);
	    return ERROR0(ERR_INVALID_DATA,
		"No KMP file: %s\n", kmp->fname ? kmp->fname : "?");
    }

    if ( disable_patch_on_load <= 0 )
	PatchKMP(kmp);

    if ( err <= ERR_WARNING && mode )
	CheckKMP(kmp,mode);

    return err;
}

///////////////////////////////////////////////////////////////////////////////

enumError ScanRawDataKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    struct raw_data_t	* raw,		// valid raw data
    CheckMode_t		mode		// not NULL: call CheckKMP(mode)
)
{
    DASSERT(kmp);
    DASSERT(raw);
    if (init_kmp)
	InitializeKMP(kmp);
    else
	ResetKMP(kmp);

    kmp->fname = raw->fname;
    raw->fname = 0;

    return raw->is_0 ? ERR_OK : ScanKMP(kmp,false,raw->data,raw->data_size,mode);
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

enumError LoadKMP
(
    kmp_t		* kmp,		// KMP data structure
    bool		init_kmp,	// true: initialize 'kmp' first
    ccp			fname,		// valid pointer to filenname
    bool		ignore_no_file,	// ignore if file does not exists
					// and return warning ERR_NOT_EXISTS
    CheckMode_t		mode		// not NULL: call CheckKMP(mode)
)
{
    DASSERT(kmp);
    DASSERT(fname);
    if (init_kmp)
	InitializeKMP(kmp);
    else
	ResetKMP(kmp);

    //--- load and scan data

    raw_data_t raw;
    enumError err = LoadRawData(&raw,true,fname,0,ignore_no_file,0);
    if (!err)
    {
	kmp->fname = raw.fname;
	raw.fname = 0;
	err = ScanKMP(kmp,false,raw.data,raw.data_size,mode);
    }

    ResetRawData(&raw);
    return err;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			CreateRawKMP() helpers		///////////////
///////////////////////////////////////////////////////////////////////////////

static void * setup_raw_kmp
(
    const kmp_t		* kmp,		// input: KMP
    uint		sect,		// input: section number
    kmp_file_t		* kfile,	// input: file header
    kmp_list_head_t	** hd,		// modify: header pointer
    const List_t	** dlist	// result: list pointer
)
{
    DASSERT(kmp);
    DASSERT( sect < KMP_N_SECT );
    DASSERT(kfile);
    DASSERT(dlist);
    DASSERT(hd);

    const List_t *dl = kmp->dlist + sect;
    *dlist = dl;
    const uint n = dl->used;

    kmp_list_head_t *h = *hd;
    *hd = (kmp_list_head_t*)( (u8*)(h+1) + kmp_entry_size[sect] * n );
    write_be32( kfile->sect_off + sect, (u8*)h - (u8*)(kfile+1) );

    memcpy(h->magic, kmp_section_name[sect].name1,sizeof(h->magic));
    write_be16(&h->n_entry,n);
    write_be16(&h->value,kmp->value[sect]);

    return h+1;
}

///////////////////////////////////////////////////////////////////////////////

static void create_kmp_pt
(
    const kmp_t		* kmp,		// input: KMP
    uint		sect,		// input: section number
    kmp_file_t		* kfile,	// input: file header
    kmp_list_head_t	** hd,		// modify: header pointer
    const u16		* mask		// AND mask for settings
)
{
    DASSERT(kmp);
    DASSERT(kfile);
    DASSERT(hd);
    DASSERT(mask);

    const List_t *dlist;
    kmp_enpt_entry_t * pt = setup_raw_kmp(kmp,sect,kfile,hd,&dlist);
    uint i;
    for ( i = 0; i < dlist->used; i++, pt++ )
    {
	DASSERT( (u8*)pt < (u8*)*hd );
	const kmp_enpt_entry_t * src = (kmp_enpt_entry_t*)dlist->list + i;
	write_bef4n(pt->position,src->position,4);
	//write_be16n(pt->prop,src->prop,2);
	write_be16(pt->prop+0,src->prop[0]&mask[0]);
	write_be16(pt->prop+1,src->prop[1]&mask[1]);
    }
}

///////////////////////////////////////////////////////////////////////////////

static void create_kmp_ph
(
    const kmp_t		* kmp,		// input: KMP
    uint		sect,		// input: section number
    kmp_file_t		* kfile,	// input: file header
    kmp_list_head_t	** hd		// modify: header pointer
)
{
    DASSERT(kmp);
    DASSERT(kfile);
    DASSERT(hd);

    const List_t *dlist;
    kmp_enph_entry_t * ph = setup_raw_kmp(kmp,sect,kfile,hd,&dlist);
    uint i;
    for ( i = 0; i < dlist->used; i++, ph++ )
    {
	DASSERT( (u8*)ph < (u8*)*hd );
	const kmp_enph_entry_t * src = (kmp_enph_entry_t*)dlist->list + i;
	memcpy(ph,src,sizeof(*ph));
    }
}

///////////////////////////////////////////////////////////////////////////////

static void create_kmp_pt2
(
    const kmp_t		* kmp,		// input: KMP
    uint		sect,		// input: section number
    kmp_file_t		* kfile,	// input: file header
    kmp_list_head_t	** hd		// modify: header pointer
)
{
    DASSERT(kmp);
    DASSERT(kfile);
    DASSERT(hd);

    const List_t *dlist;
    kmp_jgpt_entry_t * pt = setup_raw_kmp(kmp,sect,kfile,hd,&dlist);
    uint i;
    for ( i = 0; i < dlist->used; i++, pt++ )
    {
	DASSERT( (u8*)pt < (u8*)*hd );
	const kmp_jgpt_entry_t * src = (kmp_jgpt_entry_t*)dlist->list + i;
	write_bef4n(pt->position,src->position,6);
	write_be16n((u16*)&pt->id,(u16*)&src->id,2);
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			CreateRawKMP()			///////////////
///////////////////////////////////////////////////////////////////////////////

enumError CreateRawKMP
(
    kmp_t		* kmp		// pointer to valid KMP
)
{
    DASSERT(kmp);
    TRACE("CreateRawKMP()\n");


    //--- calculate data size

    const uint info_size = kmp->info ? ALIGN32(strlen(kmp->info)+1,4) : 0;
    uint data_size = sizeof(kmp_file_t)
		   + sizeof(kmp_list_head_t) * KMP_N_SECT
		   + sizeof(kmp_poti_point_t) * kmp->poti_point.used
		   + info_size;

    uint sect;
    for ( sect = 0; sect < KMP_N_SECT; sect++ )
	data_size += kmp_entry_size[sect] * kmp->dlist[sect].used;

    PRINT("CREATE KMP: size = %#x = %u\n",data_size,data_size);
    KMP_ACTION_LOG(false,"CreateRawKMP() size=%u\n",data_size);


    //--- alloc data

    FREE(kmp->raw_data);
    kmp_file_t * kfile = CALLOC(1,data_size);
    kmp->raw_data = (u8*)kfile;
    kmp->raw_data_size = data_size;


    //--- setup file header

    memcpy(kfile->magic,KMP_MAGIC,sizeof(kfile->magic));
    write_be32(&kfile->file_size,data_size);
    write_be16(&kfile->n_sect,KMP_N_SECT);
    write_be16(&kfile->sect_base_off,sizeof(*kfile));
    write_be32(&kfile->unknown_0x0c,kmp->unknown_0x0c);

    kmp_list_head_t * hd = (kmp_list_head_t*)(kfile+1);

    if (info_size)
	strcpy((char*)kfile + data_size - info_size, kmp->info );

    uint i;
    const List_t *dlist;


    //--- save KTPT entries

    kmp_ktpt_entry_t * ktpt = setup_raw_kmp(kmp,KMP_KTPT,kfile,&hd,&dlist);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );
    for ( i = 0; i < dlist->used; i++, ktpt++ )
    {
	DASSERT( (u8*)ktpt < (u8*)hd );
	const kmp_ktpt_entry_t * src = (kmp_ktpt_entry_t*)dlist->list + i;
	write_bef4n(ktpt->position,src->position,6);
	write_be16n((u16*)&ktpt->player_index,(u16*)&src->player_index,2);
    }


    //--- save ENPT + ENPH entries

    create_kmp_pt(kmp,KMP_ENPT,kfile,&hd,kmp->enpt_mask);
    create_kmp_ph(kmp,KMP_ENPH,kfile,&hd);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );


    //--- save ITPT + ITPH entries

    create_kmp_pt(kmp,KMP_ITPT,kfile,&hd,kmp->itpt_mask);
    create_kmp_ph(kmp,KMP_ITPH,kfile,&hd);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );


    //--- save CKPT entries

    kmp_ckpt_entry_t * ckpt = setup_raw_kmp(kmp,KMP_CKPT,kfile,&hd,&dlist);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );
    for ( i = 0; i < dlist->used; i++, ckpt++ )
    {
	DASSERT( (u8*)ckpt < (u8*)hd );
	const kmp_ckpt_entry_t * src = (kmp_ckpt_entry_t*)dlist->list + i;
	memcpy(ckpt,src,sizeof(*ckpt));
	write_bef4n(ckpt->left,src->left,4);
    }


    //--- save CKPH entries

    create_kmp_ph(kmp,KMP_CKPH,kfile,&hd);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );


    //--- save GOBJ entries

    kmp_gobj_entry_t * gobj = setup_raw_kmp(kmp,KMP_GOBJ,kfile,&hd,&dlist);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );
    for ( i = 0; i < dlist->used; i++, gobj++ )
    {
	DASSERT( (u8*)gobj < (u8*)hd );
	const kmp_gobj_entry_t * src = (kmp_gobj_entry_t*)dlist->list + i;
	write_be16n((u16*)&gobj->obj_id,(u16*)&src->obj_id,2);
	write_bef4n(gobj->position,src->position,9);
	write_be16n(&gobj->route_id,&src->route_id,10);
    }


    //--- save POTI entries

    kmp_list_head_t * hd2 = hd;
    u8 * poti = setup_raw_kmp(kmp,KMP_POTI,kfile,&hd,&dlist);
    uint ng = dlist->used;
    uint np = kmp->poti_point.used;
    write_be16(&hd2->n_entry,ng);
    write_be16(&hd2->value,np);

    hd = (kmp_list_head_t*)( (u8*)hd + sizeof(kmp_poti_point_t) * np );
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );

    const kmp_poti_group_t * pg = (kmp_poti_group_t*)dlist->list;
    const kmp_poti_point_t * pp = (kmp_poti_point_t*)kmp->poti_point.list;

    uint gi;
    for ( gi = 0; gi < ng && np > 0; gi++, pg++ )
    {
	DASSERT( (u8*)poti < (u8*)hd );
	write_be16((u16*)poti,pg->n_point);
	poti[2] = pg->smooth;
	poti[3] = pg->back;

	poti += sizeof(*pg);

	uint en = pg->n_point;
	DASSERT( en <= np );
	for ( ; en > 0; en--, i++, pp++ )
	{
	    noPRINT("%p/%p %p/%p\n",
		poti, hd, pp, dlist->list + dlist->used * dlist->size );

	    DASSERT( (u8*)poti < (u8*)hd );
	    write_bef4n((float32*)poti,pp->position,3);
	    poti += sizeof(pp->position);

	    DASSERT( (u8*)poti < (u8*)hd );
	    write_be16n((u16*)poti,&pp->speed,2);
	    poti += 2*sizeof(pp->speed);
	}
    }


    //--- save AREA entries

    kmp_area_entry_t * area = setup_raw_kmp(kmp,KMP_AREA,kfile,&hd,&dlist);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );
    for ( i = 0; i < dlist->used; i++, area++ )
    {
	DASSERT( (u8*)area < (u8*)hd );
	const kmp_area_entry_t * src = (kmp_area_entry_t*)dlist->list + i;
	memcpy(area,src,sizeof(*area));
	write_bef4n(area->position,src->position,9);
	write_be16n(area->setting,src->setting,4);
    }


    //--- save CAME entries

    hd2 = hd;
    kmp_came_entry_t * came = setup_raw_kmp(kmp,KMP_CAME,kfile,&hd,&dlist);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );

    for ( i = 0; i < dlist->used; i++, came++ )
    {
	DASSERT( (u8*)came < (u8*)hd );
	const kmp_came_entry_t * src = (kmp_came_entry_t*)dlist->list + i;
	memcpy(came,src,sizeof(*came));
	write_be16n(&came->came_speed,&src->came_speed,4);
	write_bef4n(came->position,src->position,15);
    }


    //--- scan JGPT + CNPT + MSPT entries

    create_kmp_pt2(kmp,KMP_JGPT,kfile,&hd);
    create_kmp_pt2(kmp,KMP_CNPT,kfile,&hd);
    create_kmp_pt2(kmp,KMP_MSPT,kfile,&hd);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );


    //--- save STGI entries

    kmp_stgi_entry_t * stgi = setup_raw_kmp(kmp,KMP_STGI,kfile,&hd,&dlist);
    DASSERT( (u8*)hd <= (u8*)kfile + data_size );
    for ( i = 0; i < dlist->used; i++, stgi++ )
    {
	DASSERT( (u8*)stgi < (u8*)hd );
	const kmp_stgi_entry_t * src = (kmp_stgi_entry_t*)dlist->list + i;
	memcpy(stgi,src,sizeof(*stgi));
	write_be32(&stgi->unknown_04,src->unknown_04);
	write_be16(&stgi->unknown_0a,src->unknown_0a);
    }

    DASSERT( (u8*)hd + info_size == (u8*)kfile + data_size );
    return ERR_OK;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    SaveRawKMP()		///////////////
///////////////////////////////////////////////////////////////////////////////

enumError SaveRawKMP
(
    kmp_t		* kmp,		// pointer to valid KMP
    ccp			fname,		// filename of destination
    bool		set_time	// true: set time stamps
)
{
    DASSERT(kmp);
    DASSERT(fname);
    PRINT("SaveRawKMP(%s,%d)\n",fname,set_time);
    KMP_ACTION_LOG(false,"SaveRawKMP() %s\n",fname);

    //--- create raw data

    enumError err = CreateRawKMP(kmp);
    if (err)
	return err;
    DASSERT(kmp->raw_data);
    DASSERT(kmp->raw_data_size);

    //--- write to file

    File_t F;
    err = CreateFileOpt(&F,true,fname,testmode,kmp->fname);
    if ( err > ERR_WARNING || !F.f )
	return err;
    SetFileAttrib(&F.fatt,&kmp->fatt,0);

    if ( fwrite(kmp->raw_data,1,kmp->raw_data_size,F.f) != kmp->raw_data_size )
	FILEERROR1(&F,ERR_WRITE_FAILED,"Write failed: %s\n",fname);
    return ResetFile(&F,set_time);
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			SaveTextKMP() helpers		///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct print_kmp_group_t
{
    int				index;		// current group index
    int				max;		// max group index
    const kmp_enph_entry_t	* ph;		// pointer to next ph element
    const kmp_group_name_t	* gname;	// list with names

} print_kmp_group_t;

//-----------------------------------------------------------------------------

static void init_print_kmp_group
(
    print_kmp_group_t		* grp,		// group element, will be initialized
    const kmp_t			* kmp,		// KMP data structure
    uint			sect_ph,	// related ph section index, may be invalid
    const kmp_group_name_t	* gname		// list with names
)
{
    DASSERT(grp);
    DASSERT(kmp);

    memset(grp,0,sizeof(*grp));

    if ( sect_ph < KMP_N_SECT )
    {
	grp->max   = kmp->dlist[sect_ph].used;
	grp->ph    = (kmp_enph_entry_t*)kmp->dlist[sect_ph].list;
	grp->gname = gname;
    }
}

//-----------------------------------------------------------------------------

static ccp print_kmp_group
(
    File_t		* F,		// output file
    const kmp_t		* kmp,		// KMP data structure
    print_kmp_group_t	* grp,		// group element, will be initialized
    uint		index,		// index of next element
    ccp			eol_text,	// text, printed to close previous line
    bool		check_eline	// print cross ref of GOBJ/ELINE_CONTROL
)
{
    DASSERT(F);
    DASSERT(grp);

    if ( grp->index < grp->max && index == grp->ph->pt_start )
    {
	if (eol_text)
	{
	    fputs(eol_text,F->f);
	    eol_text = 0;
	}

	const uint idx = grp->index++;
	fprintf(F->f,"\r\n $GROUP %s,  next:", grp->gname[idx] );

	const u8 *ptr, *end;
	for ( ptr = grp->ph->next, end = ptr + KMP_MAX_PH_LINK; ptr < end; )
	{
	    const u8 *start = ptr;
	    const u8 code = *ptr++;
	    while ( ptr < end && *ptr == code )
		ptr++;
	    if ( code != 0xff )
	    {
		const uint n = ptr - start;
		if ( n > 1 )
		    fprintf(F->f," %s*%u",grp->gname[code],n);
		else
		    fprintf(F->f," %s",grp->gname[code]);
	    }
	}

	fprintf(F->f,"\r\n  # zero based index: %u,  prev:", idx );
	for ( ptr = grp->ph->prev, end = ptr + KMP_MAX_PH_LINK; ptr < end; ptr++ )
	    if ( *ptr != 0xff )
		fprintf(F->f," %u",*ptr);

	fprintf(F->f,",  next:");
	for ( ptr = grp->ph->next, end = ptr + KMP_MAX_PH_LINK; ptr < end; ptr++ )
	    if ( *ptr != 0xff )
		fprintf(F->f," %u",*ptr);

	if (check_eline)
	{
	    uint gidx;
	    for ( gidx = 0; gidx < kmp->dlist[KMP_GOBJ].used; gidx++ )
	    {
		const kmp_gobj_entry_t *op = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list + gidx;
		if ( op->obj_id == GOBJ_ELINE_CONTROL && op->setting[7] == idx )
		    fprintf(F->f,
			"\r\n  # GOBJ/ELINE_CONTROL #%u links to this group:"
			" enter=%d, sec=%d, id=%d, next=%d",
			gidx, op->setting[0], op->setting[1], op->setting[2], op->setting[3] );
	    }
	}

	fputs("\r\n\r\n",F->f);
	grp->ph++;
    }

    return eol_text;
}

///////////////////////////////////////////////////////////////////////////////

static ccp print_kmp_flags ( kmp_pt_flag_t flags, kmp_pt_flag_t pmask )
{
    if (!opt_kmp_flags)
	return EmptyString;

    static char buf[4];

    char *d = buf;
    if ( pmask & KPFL_FALL )	*d++ =  flags & KPFL_FALL	? 'F' : '-';
    if ( pmask & KPFL_SNAP )	*d++ =  flags & KPFL_SNAP	? 'S' : '-';
    if ( pmask & KPFL_JGPT )	*d++ =  flags & KPFL_JGPT	? 'J' : '-';

    DASSERT( d < buf + sizeof(buf) );
    *d = 0;
    return buf;
}

///////////////////////////////////////////////////////////////////////////////

static void print_kmp_pt
(
    File_t			* F,		// output file
    const kmp_t			* kmp,		// KMP data structure
    uint			sect_pt,	// KMP section index
    uint			sect_ph		// related ph section index
)
{
    DASSERT(F);
    DASSERT(kmp);
    DASSERT( sect_pt < KMP_N_SECT );
    DASSERT( sect_ph < KMP_N_SECT );

    ccp pt_name = kmp_section_name[sect_pt].name1;
    ccp ph_name = kmp_section_name[sect_ph].name1;

    uint show_route = sect_pt == KMP_ENPT ? (1<<SHRT_ENPT) : (1<<SHRT_ITPT);
    kmp_pt_flag_t flag_mask;
    const u8 *pt_flags;
    const kmp_group_name_t *gname;
    const u16 *mask;

    if ( sect_pt == KMP_ENPT )
    {
	show_route	= 1 << SHRT_ENPT;
	flag_mask	= KPFL_M_ENPT;
	pt_flags	= kmp->enpt_flags;
	gname		= kmp->enph_name;
	mask		= kmp->enpt_mask;
    }
    else
    {
	DASSERT( sect_pt == KMP_ITPT );
	show_route	= 1 << SHRT_ITPT;
	flag_mask	= KPFL_M_ITPT;
	pt_flags	= kmp->itpt_flags;
	gname		= kmp->itph_name;
	mask		= kmp->itpt_mask;
    }

    fprintf(F->f,text_kmp_pt_info_cr,
		REVISION_NUM,
		pt_name, kmp->value[sect_pt],	// @%s-HEAD-VALUE = %#x
		ph_name, kmp->value[sect_ph],	// @%s-HEAD-VALUE = %#x
		show_route );

    if ( mask[0] != 0xffff || mask[1] != 0xffff )
	fprintf(F->f,text_kmp_pt_mask_cr, mask[0], mask[1] );

    fprintf(F->f, text_kmp_pt_head_cr, pt_name );

    const uint n = kmp->dlist[sect_pt].used;
    const kmp_enpt_entry_t * pt = (kmp_enpt_entry_t*)kmp->dlist[sect_pt].list;

    print_kmp_group_t grp;
    init_print_kmp_group(&grp,kmp,sect_ph,gname);
    ccp eol = "";

    uint i, gi = 0;
    for ( i = 0; i < n; i++, gi++, pt++ )
    {
	eol = print_kmp_group(F,kmp,&grp,i,eol,sect_pt==KMP_ENPT);
	if (!eol)
	{
	    gi = 0;
	    eol = "";
	}
	else if ( gi > 1 )
	{
	    fprintf(F->f," %7.2f",
		AngleVector(pt[-2].position,pt[-1].position,pt->position,true) );
	}

	fprintf(F->f,"%s%4u  %11.3f %11.3f %11.3f  %7.3f  %#6x %#6x",
		eol, i,
		pt->position[0], pt->position[1], pt->position[2],
		pt->scale,
		pt->prop[0], pt->prop[1] );
	eol = "\r\n";

	if (pt_flags)
	{
	    const kmp_pt_flag_t flags = i < KMP_MAX_POINT ? pt_flags[i] : KPFL_M_DEFAULT;
	    fprintf(F->f,"  %-3s",print_kmp_flags(flags,flag_mask));
	}

	if ( gi > 0 )
	    fprintf(F->f," # %9.3f", LengthF(pt[-1].position,pt->position) );
    }
    fputs(eol,F->f);
}

///////////////////////////////////////////////////////////////////////////////

static void print_kmp_ph
(
    File_t			* F,		// output file
    const kmp_t			* kmp,		// KMP data structure
    uint			sect_ph,	// KMP section index
    uint			sect_pt,	// related pt section index
    const kmp_group_name_t	* gname		// list with names
)
{
    DASSERT(F);
    DASSERT(kmp);
    DASSERT( sect_ph < KMP_N_SECT );
    DASSERT( sect_pt < KMP_N_SECT );

    ccp pt_name = kmp_section_name[sect_pt].name1;
    ccp ph_name = kmp_section_name[sect_ph].name1;

    fprintf(F->f,text_kmp_ph_cr,
		pt_name,
		REVISION_NUM,
		ph_name, kmp->value[sect_ph],
		ph_name );

    const uint n = kmp->dlist[sect_ph].used;
    const kmp_enph_entry_t * p = (kmp_enph_entry_t*)kmp->dlist[sect_ph].list;
    uint i;
    for ( i = 0; i < n; i++, p++ )
    {
	fprintf(F->f,"%4u  %3u %3u %3u ",
			i, p->pt_start, p->pt_start+p->pt_len-1, p->pt_len );
	DASSERT( p->prev + KMP_MAX_PH_LINK == p->next ); // assume 'prev' before 'next'
	const u8 * d = p->prev;
	uint j;
	for ( j = 0; j < 12; j++, d++ )
	    if ( *d == 0xff )
		fprintf(F->f,"%s   -", j==6 ? " " : "" );
	    else
		fprintf(F->f,"%s%4u", j==6 ? " " : "", *d );
     #ifdef TEST
	fprintf(F->f," %#6x  # %-6s %s\r\n",
		p->unknown, gname[i],
		i < kmp->index[sect_ph].used ? kmp->index[sect_ph].list[i] : "-" );
     #else
	fprintf(F->f," %#6x  # %s\r\n",p->unknown,gname[i]);
     #endif
    }
}

///////////////////////////////////////////////////////////////////////////////

static void print_kmp_pt2
(
    File_t		* F,		// output file
    const kmp_t		* kmp,		// KMP data structure
    uint		sect,		// KMP section index
    uint		* ref_count	// NULL or reference counter array (len=0x100)
)
{
    DASSERT(F);
    DASSERT(kmp);
    DASSERT( sect < KMP_N_SECT );

    uint i;
    const uint n = kmp->dlist[sect].used;
    const kmp_jgpt_entry_t * pt = (kmp_jgpt_entry_t*)kmp->dlist[sect].list;
    IndexList_t *il = (IndexList_t*)kmp->index + sect; // cast because remove const
    FillIL(il,n);

    bool auto_id = true;
    for ( i = 0; i < n; i++, pt++ )
	if ( pt->id != i )
	{
	    auto_id = false;
	    break;
	}

    ccp pt_name = kmp_section_name[sect].name1;
    uint show_route = sect == KMP_JGPT
				? (1<<SHRT_JGPT)
				: sect == KMP_CNPT
					? (1<<SHRT_CNPT)
					: (1<<SHRT_MSPT);
    const u8 *pt_flags = sect == KMP_JGPT ? kmp->jgpt_flags : 0;

    fprintf( F->f, text_kmp_pt2_cr,
		REVISION_NUM,
		pt_name, kmp->value[sect],
		auto_id,
		show_route );

    if ( sect == KMP_JGPT )
	fprintf( F->f, text_kmp_jgpt_auto_cr );
    else
	fprintf( F->f, text_kmp_pt2_head_cr, pt_name );

    pt = (kmp_jgpt_entry_t*)kmp->dlist[sect].list;
    for ( i = 0; i < n; i++, pt++ )
    {
	PrintNameIL(F->f,il,i,5,5,0);
	fprintf(F->f," %11.3f %11.3f %11.3f %7.2f %7.2f %7.2f %6d %#6x",
		pt->position[0], pt->position[1], pt->position[2],
		pt->rotation[0], pt->rotation[1], pt->rotation[2],
		pt->id, pt->effect );

	if (pt_flags)
	{
	    const kmp_pt_flag_t flags = i < KMP_MAX_POINT ? pt_flags[i] : KPFL_M_DEFAULT;
	    fprintf(F->f,"  %s",print_kmp_flags(flags,KPFL_M_JGPT));
	}

	fprintf(F->f,"%s\r\n", ref_count && !ref_count[i] ? "  # unused" : "" );
    }
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  SaveTextKMP()			///////////////
///////////////////////////////////////////////////////////////////////////////

enumError SaveTextKMP
(
    const kmp_t		* kmp,		// pointer to valid KMP
    ccp			fname,		// filename of destination
    bool		set_time	// true: set time stamps
)
{
    // use DOS/Windows line format -> unix can handle it ;)

    DASSERT(kmp);
    DASSERT(fname);
    PRINT("SaveTextKMP(%s,%d)\n",fname,set_time);
    KMP_ACTION_LOG(false,"SaveTextKMP() %s\n",fname);

    char temp_buf[100];


    //--- open file

    File_t F;
    enumError err = CreateFileOpt(&F,true,fname,testmode,kmp->fname);
    if ( err > ERR_WARNING || !F.f )
	return err;
    SetFileAttrib(&F.fatt,&kmp->fatt,0);


    //--- print header + syntax info

    if ( print_header && !brief_count && !export_count )
	fprintf(F.f,text_kmp_head_cr);
    else
	fprintf(F.f,
		"%s\r\n"
		"# KMP syntax & semantics: http://szs.wiimm.de/info/kmp-syntax.html\r\n"
		,KMP_TEXT_MAGIC);


    //--- var defs

    uint i, n, gi, ng, np;
    char buf1[0x100], buf2[0x20];


    //--- print info section & section counter alphabetically

    if ( brief_count || export_count )
    {
	fprintf(F.f,
		"%s[SETUP]\r\n\r\n"
		"TOOL     = %s\r\n"
		"SYSTEM   = %s\r\n"
		"VERSION  = %s\r\n"
		"REVISION = %u\r\n"
		"DATE     = %s\r\n"
		"INFO     = \"%s\"\r\n"
		"\r\n"
		"UNKNOWN-0x0c = %#x\r\n"
		"\r\n"
		"KCL-FALL-ADD-ROAD = 0\r\n"
		"\r\n",
		section_sep,
		tool_name, SYSTEM, VERSION, REVISION_NUM, DATE,
		kmp->info ? kmp->info : "", kmp->unknown_0x0c );
    }
    else
    {
	fprintf(F.f, text_kmp_setup_cr,
		tool_name, SYSTEM, VERSION, REVISION_NUM, DATE,
		kmp->info ? kmp->info : "", kmp->unknown_0x0c );

	ccp last_name = "";
	for(;;)
	{
	    ccp found_name = 0;
	    uint sect, found_sect = 0;
	    for ( sect = 0; sect < KMP_N_SECT; sect++ )
	    {
		ccp name = kmp_section_name[sect].name1;
		if ( strcmp(name,last_name) > 0
		    && ( !found_name || strcmp(name,found_name) < 0 ) )
		{
		    found_name = name;
		    found_sect = sect;
		}
	    }

	    if (!found_name)
		break;

	    if ( found_sect == KMP_POTI )
		fprintf(F.f, " # N-%s = %5u,%5u\r\n",
			    found_name,
			    kmp->dlist[found_sect].used, kmp->poti_point.used );
	    else
		fprintf(F.f, " # N-%s = %5u\r\n",
			    found_name,
			    kmp->dlist[found_sect].used );
	    last_name = found_name;
	}
    }


    //--- print default settings

    fprintf(F.f, text_kmp_defaults_cr,
	GetObjectName(kmp->rtobj.obj),
	kmp->rtobj.shift.x, kmp->rtobj.shift.y, kmp->rtobj.shift.z,
	kmp->rtobj.scale.x, kmp->rtobj.scale.y, kmp->rtobj.scale.z );


    //--- print parameter support

    if (print_param)
    {
	if ( !brief_count && !export_count )
	    fwrite(text_kmp_param1_cr,1,sizeof(text_kmp_param1_cr)-1,F.f);
	else
	    fprintf(F.f,"#%.78s\r\n# Parameter support\r\n",Minus300);

	fwrite(text_kmp_param2_cr,1,sizeof(text_kmp_param2_cr)-1,F.f);

	if ( !brief_count && !export_count )
	    fwrite(text_kmp_param3_cr,1,sizeof(text_kmp_param3_cr)-1,F.f);
    }


    //--- fill index lists

    IndexList_t *ill = (IndexList_t*)kmp->index; // cast because remove const
    for ( i = 0; i < KMP_N_SECT; i++ )
	FillIL(ill+i,kmp->dlist[i].used);


    //--- print macros

    fwrite(text_kmp_macros_cr,1,sizeof(text_kmp_macros_cr)-1,F.f);


    //--- print AREA entries

    fprintf(F.f,text_kmp_area_cr,REVISION_NUM,kmp->value[KMP_AREA]);

    n = kmp->dlist[KMP_AREA].used;
    const kmp_area_entry_t * area = (kmp_area_entry_t*)kmp->dlist[KMP_AREA].list;
    for ( i = 0; i < n; i++, area++ )
    {
	*buf1 = 0;
	PrintNameIL(F.f,ill+KMP_AREA,i,6,8,0);
	fprintf(F.f,
		"%#4x %#4x  %12.3f %11.3f %11.3f  %#6x %#6x %#6x %#6x\r\n >",
		area->mode, area->type,
		 area->position[0], area->position[1], area->position[2],
		 area->setting[0], area->setting[1], area->setting[2], area->setting[3] );

	if ( area->type == 0x00 )
	    PrintNameIL(F.f,ill+KMP_CAME,area->dest_id,10,10,2);
	else
	    fprintf(F.f,"%#10x",area->dest_id);

	fprintf(F.f,
		" %#4x  %12.3f %11.3f %11.3f \r\n"
		    " > %28.3f %11.3f %11.3f\r\n"
		"#%.84s\r\n",
		area->unknown,
		 area->rotation[0], area->rotation[1], area->rotation[2],
		area->scale[0], area->scale[1], area->scale[2],
		Minus300 );
    }


    //--- print CAME entries

    uint ocam = kmp->value[KMP_CAME] >> 8;
    uint scam = kmp->value[KMP_CAME] & 0xff;

    fprintf(F.f,text_kmp_came_cr,REVISION_NUM,
		GetNameIL(ill+KMP_CAME,ocam),
		GetNameIL(ill+KMP_CAME,scam) );

    // find opening cameras

    n = kmp->dlist[KMP_CAME].used;
    const kmp_came_entry_t * came = (kmp_came_entry_t*)kmp->dlist[KMP_CAME].list;

    uint ocount = 0, oidx = ocam;
    memset(buf1,0,sizeof(buf1));
    if (print_param)
	while ( oidx < n && ocount < 0xff && oidx < sizeof(buf1) && !buf1[oidx] )
	{
	    buf1[oidx] = ++ocount;
	    oidx = came[oidx].next;
	}

    for ( i = 0; i < n; i++, came++ )
    {
	//--- AREA reference

	const kmp_area_entry_t * area = (kmp_area_entry_t*)kmp->dlist[KMP_AREA].list;
	uint count, idx, nn = kmp->dlist[KMP_AREA].used;
	for ( count = idx = 0; idx < nn; idx++, area++ )
	    if ( area->type == 0x00 && area->dest_id == i )
		fprintf(F.f,"%s %s",
		    count++ ? "" : "\t# AREA reference:",
		    GetNameIL(ill+KMP_AREA,idx) );
	if (count)
	    fputs("\r\n",F.f);


	//--- CAME reference

	count = 0;
	if ( i == ocam )
	    fprintf(F.f,"%s @OPENING-INDEX",
		count++ ? "" : "\t# CAME reference:" );
	if ( i == scam )
	    fprintf(F.f,"%s @SELECTION-INDEX",
		count++ ? "" : "\t# CAME reference:" );

	const kmp_came_entry_t * cref = (kmp_came_entry_t*)kmp->dlist[KMP_CAME].list;
	nn = kmp->dlist[KMP_CAME].used;
	for ( idx = 0; idx < nn; idx++, cref++ )
	    if ( cref->next == i )
		fprintf(F.f,"%s %s",
			count++ ? "" : "\t# CAME reference:",
			GetNameIL(ill+KMP_CAME,idx) );
	if (count)
	    fputs("\r\n",F.f);


	//--- CAME object

	PrintNameIL(F.f,ill+KMP_CAME,i,6,8,0);
	fprintf(F.f," %3u %#7x  %11.3f %11.3f %11.3f %10.3f",
		came->type, came->came_speed,
		 came->position[0], came->position[1], came->position[2],
		 came->zoom_begin );
	const double came_time = came->time/60.0;
	if ( came_time >= 0.01 )
	    fprintf(F.f,"  # came time:%8.2fs",came_time);
	fputs("\r\n >",F.f);

	if (buf1[i])
	{
	    uint idx = (uchar)buf1[i];
	    ccp term = "\r\n >            ";
	    if ( idx == ocount )
		fprintf(F.f,
		    " ( mode$ocam == %u ? %s : mode$test ? %s : %s )%s",
		    idx, GetNameIL(ill+KMP_CAME,i),
		    GetNameIL(ill+KMP_CAME,kmp->value[KMP_CAME]>>8),
		    GetNameIL(ill+KMP_CAME,came->next), term );
	    else
		fprintf(F.f,
		    " ( mode$ocam == %u ? %s : %s )%s",
		    idx, GetNameIL(ill+KMP_CAME,i),
		    GetNameIL(ill+KMP_CAME,came->next), term );
	}
	else
	    PrintNameIL(F.f,ill+KMP_CAME,came->next,10,12,2);

	fprintf(F.f," %5u  %11.3f %11.3f %11.3f %10.3f",
		came->zoom_speed,
		 came->rotation[0], came->rotation[1], came->rotation[2],
		 came->zoom_end );
	const double zoom_length = fabs(came->zoom_begin-came->zoom_end);
	if ( zoom_length >= 0.01 && came->zoom_speed )
	    fprintf(F.f,"  # zoom time:%8.2fs",
			zoom_length / came->zoom_speed * (10.0/6.0) );

	fprintf(F.f,"\r\n > %#9x %7u  %11.3f %11.3f %11.3f",
		came->unknown_0x02, came->viewpt_speed,
		 came->viewpt_begin[0], came->viewpt_begin[1], came->viewpt_begin[2] );
	const double viewpt_len = LengthF(came->viewpt_begin,came->viewpt_end);
	if ( viewpt_len >= 0.01 && came->viewpt_speed )
	    fprintf(F.f,"             # v.pt time:%8.2fs",
			viewpt_len / came->viewpt_speed * (10.0/6.0) );
	fputs("\r\n >",F.f);

	PrintNameIL(F.f,ill+KMP_POTI,came->route,10,11,2);
	fprintf(F.f,
		" %#6x  %11.3f %11.3f %11.3f %10.3f",
		came->unknown_0x0a,
		 came->viewpt_end[0], came->viewpt_end[1], came->viewpt_end[2],
		 came->time );
	if ( viewpt_len >= 0.01 )
	    fprintf(F.f,"  # v.pt len:%9.2f",viewpt_len);
	fprintf(F.f,"\r\n#%.91s\r\n",Minus300);
    }


    //--- print CKPH entries

    if ( brief_count < 2 && !export_count )
    {
	fwrite(text_kmp_ckph_cr,1,sizeof(text_kmp_ckph_cr)-1,F.f);
	print_kmp_ph(&F,kmp,KMP_CKPH,KMP_CKPT,kmp->ckph_name);
    }


    //--- print CKPT entries

    bool auto_next = true;
    n  = kmp->dlist[KMP_CKPT].used;
    ng = kmp->dlist[KMP_CKPH].used;
    const kmp_ckph_entry_t * ckph = (kmp_ckph_entry_t*)kmp->dlist[KMP_CKPH].list;
    for ( gi = 0; gi < ng && auto_next; gi++, ckph++ )
    {
	uint start = ckph->pt_start;
	uint end = start + ckph->pt_len - 1;
	if ( !ckph->pt_len || end >= n )
	{
	    auto_next = false;
	    break;
	}

	const kmp_ckpt_entry_t * ckpt
		= (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list + start;
	uint j;
	for ( j = start; j <= end; j++, ckpt++ )
	{
	    if (   ckpt->prev != ( j == start ? -1 : j-1 )
		|| ckpt->next != ( j == end   ? -1 : j+1 ) )
	    {
		auto_next = false;
		break;
	    }
	}
    }

    fprintf(F.f,text_kmp_ckpt_cr,
		REVISION_NUM,
		kmp->value[KMP_CKPT], kmp->value[KMP_CKPH],
		auto_next,
		ObjectInfo[RTOBJ_1].name, ObjectInfo[RTOBJ_2].name,
		RTOBJ_SCALE_X, RTOBJ_SCALE_Y, RTOBJ_SCALE_Z,
		RTOBJ_SCALE_X, RTOBJ_SCALE_Y, RTOBJ_SCALE_Z,
		RTOBJ_BASE, RTOBJ_BASE );

    const kmp_ckpt_entry_t * ckpt = (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list;

    print_kmp_group_t grp;
    init_print_kmp_group(&grp,kmp,KMP_CKPH,kmp->ckph_name);
    uint jgpt_count[0x100];
    memset(jgpt_count,0,sizeof(jgpt_count));

    u8 prev_mode = M1(prev_mode);
    for ( i = 0; i < n; i++, ckpt++ )
    {
	if ( print_kmp_group(&F,kmp,&grp,i,"",false) && IS_M1(ckpt->mode) && !IS_M1(prev_mode) )
	    fputs("\r\n",F.f);
	prev_mode = ckpt->mode;
	jgpt_count[(uint)ckpt->respawn]++;

	double width, direction;
	WidthAndDirection(&width,&direction,true,ckpt->left,ckpt->right);

	const uint indent
	    = fprintf(F.f,"%4u %11.3f %11.3f %11.3f %11.3f ",
		i,
		ckpt->left[0], ckpt->left[1],
		ckpt->right[0], ckpt->right[1] );
	PrintNameIL(F.f,ill+KMP_JGPT,ckpt->respawn,5,5,indent);
	fprintf(F.f," %3d %3d %3d # %9.3f %7.2f\r\n",
		IS_M1(ckpt->mode) ? -1 : (uint)ckpt->mode,
		IS_M1(ckpt->prev) ? -1 : (uint)ckpt->prev,
		IS_M1(ckpt->next) ? -1 : (uint)ckpt->next,
		width, direction );
    }


    //--- print CNPT entries

    fwrite(text_kmp_cnpt_cr,1,sizeof(text_kmp_cnpt_cr)-1,F.f);
    print_kmp_pt2(&F,kmp,KMP_CNPT,0);


    //--- print ENPH entries

    if ( brief_count < 2 && !export_count )
    {
	fwrite(text_kmp_enph_cr,1,sizeof(text_kmp_enph_cr)-1,F.f);
	print_kmp_ph(&F,kmp,KMP_ENPH,KMP_ENPT,kmp->enph_name);
    }


    //--- print ENPT entries

    fwrite(text_kmp_enpt_cr,1,sizeof(text_kmp_enpt_cr)-1,F.f);
    print_kmp_pt(&F,kmp,KMP_ENPT,KMP_ENPH);


    //--- print GOBJ entries

    n = kmp->dlist[KMP_GOBJ].used;
    fprintf(F.f,text_kmp_gobj_cr,REVISION_NUM,kmp->value[KMP_GOBJ]);

    //- print object statistics

    uint min_id = 0;
    for(;;)
    {
	uint current_id = UINT_MAX, count = 0;

	const kmp_gobj_entry_t * gobj = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	for ( i = 0; i < n; i++, gobj++ )
	{
	    if ( gobj->obj_id >= min_id )
	    {
		if ( gobj->obj_id == current_id )
		    count++;
		else if ( gobj->obj_id < current_id )
		{
		    count = 1;
		    current_id = gobj->obj_id;
		}
	    }
	}

	if (!count)
	    break;

	if ( current_id < N_KMP_GOBJ && ObjectInfo[current_id].name )
	{
	    const ObjectInfo_t *oi = ObjectInfo + current_id;
	    fprintf(F.f,
		    "##%4u %#6x  %-*s %s\r\n",
		    count, current_id, OBJ_FW_NAME, oi->name,
		    oi->charact );
	}
	else
	    fprintf(F.f,"##%4u %#6x\r\n", count, current_id );

	min_id = current_id + 1;
    }
    fprintf(F.f,"##%.83s\r\n\r\n",Minus300);

    fwrite(text_kmp_gobj_tabhead_cr,1,sizeof(text_kmp_gobj_tabhead_cr)-1,F.f);

    u8 done[N_KMP_GOBJ];
    memset(done,0,sizeof(done));

    const kmp_gobj_entry_t * gobj = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
    for ( i = 0; i < n; i++, gobj++ )
    {
	ccp form1 = settings_format_uuuu;
	ccp form2 = settings_format_uuuu;

	if ( gobj->obj_id < N_KMP_GOBJ )
	{
	    const ObjectInfo_t *oi = ObjectInfo + gobj->obj_id;
	    if (oi->name)
	    {
		ccp char_sep = *oi->charact ? ", " : "";

		if (!done[gobj->obj_id] )
		{
		    done[gobj->obj_id] = 1;

		    fprintf(F.f,
			"## o$%s = %#x = %u%s%s\r\n",
			oi->name, gobj->obj_id, gobj->obj_id,
			char_sep, oi->charact );

		    if (oi->info)
		    {
			fputs("## Description: ",F.f);
			PutLines(F.f,0,80,13,"## ",oi->info,"\r\n");
		    }

		    if (oi->fname)
			fprintf(F.f,"## File names:  %s\r\n",oi->fname);

		    uint p;
		    for ( p = 0; p < 8; p++ )
			if ( oi->param[p] )
			{
			    fprintf(F.f,"## Setting #%u:  ",p+1);
			    PutLines(F.f,13,80,13,"## ",oi->param[p],"\r\n");
			}
		    fprintf(F.f,"#%.84s\r\n",Minus300);
		}

		fprintf(F.f,"# o$%s, S",oi->name);
		uint p;
		for ( p = 0; p < 8; p++ )
		    fputc( oi->param[p] ? '1'+p : '-', F.f );
		fprintf(F.f,"%s%s\r\n",char_sep, oi->charact);
	    }

	    if (oi->format1)
		form1 = oi->format1;
	    if (oi->format2)
		form2 = oi->format2;
	}

	snprintf(buf1,sizeof(buf1),form1,
			gobj->setting[0], gobj->setting[1],
			gobj->setting[2], gobj->setting[3] );
	switch (gobj->obj_id)
	{
	 case GOBJ_ELINE_CONTROL:
	    {
		const int idx = IS_M1(gobj->setting[7]) ? -1 : gobj->setting[7];
		if ( (uint)idx < KMP_MAX_GROUP && kmp->enph_name[idx] )
		    snprintf(temp_buf,sizeof(temp_buf),"E.%s",kmp->enph_name[idx]);
		else
		    snprintf(temp_buf,sizeof(temp_buf),"%6d",idx);
		snprintf(buf2,sizeof(buf2),"%6u %6u %6u %6s",
			gobj->setting[4], gobj->setting[5],
			gobj->setting[6], temp_buf );
	    }
	    break;

	 default:
	    snprintf(buf2,sizeof(buf2),form2,
			gobj->setting[4], gobj->setting[5],
			gobj->setting[6], gobj->setting[7] );
	}

	PrintNameIL(F.f,ill+KMP_GOBJ,i,5,5,0);
	fprintf(F.f,
		       " %#6x %11.3f %11.3f %11.3f %s  %6s\r\n"
		  "    > %#6x %11.3f %11.3f %11.3f %s  %#6x\r\n"
		"    >        %11.3f %11.3f %11.3f\r\n"
		"#%.84s\r\n",
		gobj->obj_id,
		 gobj->position[0], gobj->position[1], gobj->position[2],
		 buf1, GetNameIL(ill+KMP_POTI,gobj->route_id),
		gobj->unknown,
		 gobj->rotation[0], gobj->rotation[1], gobj->rotation[2],
		 buf2, gobj->presence_flags,
		gobj->scale[0], gobj->scale[1], gobj->scale[2],
		Minus300 );
    }


    //--- print ITPH entries

    if ( brief_count < 2 && !export_count )
    {
	fwrite(text_kmp_itph_cr,1,sizeof(text_kmp_itph_cr)-1,F.f);
	print_kmp_ph(&F,kmp,KMP_ITPH,KMP_ITPT,kmp->itph_name);
    }


    //--- print ITPT entries

    fwrite(text_kmp_itpt_cr,1,sizeof(text_kmp_itpt_cr)-1,F.f);
    print_kmp_pt(&F,kmp,KMP_ITPT,KMP_ITPH);


    //--- print JGPT entries

    fwrite(text_kmp_jgpt_cr,1,sizeof(text_kmp_jgpt_cr)-1,F.f);
    print_kmp_pt2(&F,kmp,KMP_JGPT,jgpt_count);


    //--- print KTPT entries

    fprintf(F.f,text_kmp_ktpt_cr,REVISION_NUM,kmp->value[KMP_KTPT]);

    n = kmp->dlist[KMP_KTPT].used;
    const kmp_ktpt_entry_t * ktpt = (kmp_ktpt_entry_t*)kmp->dlist[KMP_KTPT].list;
    for ( i = 0; i < n; i++, ktpt++ )
    {
	fprintf(F.f,"%4u  %11.3f %11.3f %11.3f  %7.2f %7.2f %7.2f %5d %#6x\r\n",
		i,
		ktpt->position[0], ktpt->position[1], ktpt->position[2],
		ktpt->rotation[0], ktpt->rotation[1], ktpt->rotation[2],
		ktpt->player_index, ktpt->unknown );
    }


    //--- print MSPT entries

    fwrite(text_kmp_mspt_cr,1,sizeof(text_kmp_mspt_cr)-1,F.f);
    print_kmp_pt2(&F,kmp,KMP_MSPT,0);


    //--- print POTI entries

    fprintf(F.f,text_kmp_poti_cr,REVISION_NUM,1<<SHRT_POTI);

    ng = kmp->dlist[KMP_POTI].used;
    const kmp_poti_group_t * pg = (kmp_poti_group_t*)kmp->dlist[KMP_POTI].list;
    np = kmp->poti_point.used;
    const kmp_poti_point_t * pp = (kmp_poti_point_t*)kmp->poti_point.list;

    for ( gi = i = 0; gi < ng && i < np; gi++, pg++ )
    {
	uint en = pg->n_point;
	DASSERT( en <= np-i );
	fprintf(F.f,"\r\n $ROUTE %s, settings: %u %u # section %u\r\n",
			GetNameIL(ill+KMP_POTI,gi),
			pg->smooth, pg->back, gi+1 );
	const bool is_cyclic = !pg->back;

	//--- GOBJ reference

	const kmp_gobj_entry_t * gobj = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	uint count, idx, nn = kmp->dlist[KMP_GOBJ].used;
	for ( count = idx = 0; idx < nn; idx++, gobj++ )
	    if ( gobj->route_id == (uint)gi )
	    {
		fprintf(F.f,"\t# GOBJ reference: %-8s type %#5x",
			GetNameIL(ill+KMP_GOBJ,idx), gobj->obj_id );
		if ( gobj->obj_id < N_KMP_GOBJ )
		{
		    const ObjectInfo_t *oi = ObjectInfo + gobj->obj_id;
		    if (oi->name)
			fprintf(F.f," = %s",oi->name);
		}
		fputs("\r\n",F.f);
	    }

	//--- CAME reference

	ccp came_ref = 0;
	const kmp_came_entry_t * came = (kmp_came_entry_t*)kmp->dlist[KMP_CAME].list;
	nn = kmp->dlist[KMP_CAME].used;
	for ( count = idx = 0; idx < nn; idx++, came++ )
	    if ( came->route == gi )
	    {
		ccp came_name = GetNameIL(ill+KMP_CAME,idx);
		if (!came_ref)
		    came_ref = came_name;
		fprintf(F.f,"%s %s",
			count++ ? "" : "\t# CAME reference:",
			came_name );
	    }

	if (came_ref)
	    fprintf(F.f,"\r\n\r\n"
		"   # @set-route-time = %s.came.time  # in 1/60 sec\r\n",came_ref);


	//--- print points

	fputs("\r\n",F.f);
	uint ei;
	double total_len = 0.0, total_tim = 0.0;
	for ( ei = 0; ei < en; ei++, i++, pp++ )
	{
	    fprintf(F.f, "%4u  %11.3f %11.3f %11.3f %5u %5u  #",
		ei,
		pp->position[0], pp->position[1], pp->position[2],
		pp->speed, pp->unknown );

	    if ( ei < en-1 || is_cyclic )
	    {
		const kmp_poti_point_t * pnext = pp + 1;
		if ( ei >= en-1 )
		    pnext -= en;
		double len = LengthF(pp->position,pnext->position);
		total_len += len;
		double tim = len > 0.0 && pp->speed > 0
				? len / (pp->speed*60) : 0.0;
		if ( tim >= 0.01 && tim <= 999.99 )
		{
		    total_tim += tim;
		    fprintf(F.f," %6.2f %9.3f", tim, len );
		}
		else
		    fprintf(F.f,"      - %9.3f", len );

		if ( ei || is_cyclic )
		{
		    const kmp_poti_point_t * pprev = pp - 1;
		    if (!ei)
			pprev += en;
		    fprintf(F.f," %7.2f",
			AngleVector(pprev->position,pp->position,pnext->position,true) );
		}
	    }
	    fputs("\r\n",F.f);
	}

	if ( en > 1 )
	    fprintf(F.f,"%56s %6.2f %9.3f\r\n","#TOTAL#",total_tim,total_len);
    }


    //--- print STGI entries

    fprintf(F.f,text_kmp_stgi_cr,REVISION_NUM,kmp->value[KMP_STGI]);

    n = kmp->dlist[KMP_STGI].used;
    const kmp_stgi_entry_t * stgi = (kmp_stgi_entry_t*)kmp->dlist[KMP_STGI].list;

    for ( i = 0; i < n; i++, stgi++ )
    {
	PrintNameIL(F.f,ill+KMP_STGI,i,4,5,0);
	fprintf(F.f," %#4x %4u %5u %#6x %#10x %#5x %#4x %#4x %#4x # ",
		stgi->lap_count, stgi->pole_pos,
		stgi->narrow_start, stgi->unknown_03,
		stgi->unknown_04,
		stgi->unknown_08, stgi->unknown_09,
		stgi->unknown_0a >> 8, stgi->unknown_0a & 0xff );
	if (stgi->unknown_0a)
	    fprintf(F.f,"%7.3f\r\n",SpeedMod2float(stgi->unknown_0a));
	else
	    fputs("   -\r\n",F.f);
    }


    //--- terminate

    fputs(section_end,F.f);
    ResetFile(&F,set_time);
    return ERR_OK;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			Sort GOBJ			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct sort_gobj_t
{
    int			rank;	// sort 1: rank of object
    int			angle;	// sort 2: h-angle around (0,0) in degree*1000000
    int			pos;	// sort 3: old pos of object
    kmp_gobj_entry_t	data;	// raw object data

} sort_gobj_t;

//-----------------------------------------------------------------------------

static int sort_gobj ( const sort_gobj_t * a, const sort_gobj_t * b )
{
    int stat = a->rank - b->rank;
    if (!stat)
    {
	stat = a->angle - b->angle;
	if (!stat)
	    stat = a->pos - b->pos;
    }
    return stat;
}

//-----------------------------------------------------------------------------

bool SortGOBJ
(
    kmp_t		* kmp,		// pointer to valid KMP
    int			sort_mode	// <=0: dont sort
					//   1: group only
					//   2: secondary sort by angle
)
{
    DASSERT(kmp);
    PRINT("SortGOBJ(%d)\n",sort_mode);

    List_t *dlist = kmp->dlist + KMP_GOBJ;
    if ( sort_mode < 1 || dlist->used < 2 )
	return false;


    //--- build a group rank table

    int rank[N_KMP_GOBJ];
    memset(rank,0,sizeof(rank));

    kmp_gobj_entry_t *gobj = (kmp_gobj_entry_t*)dlist->list;
    int i, rank_count = 0;
    for ( i = dlist->used - 1; i >= 0; i-- )
    {
	const typeof(gobj->obj_id) obj_id = gobj[i].obj_id;
	if ( obj_id < N_KMP_GOBJ )
	    rank[obj_id] = --rank_count;
    }
    rank[0] = 0;


    //--- create sort array and copy data

    sort_gobj_t *sg_beg = CALLOC(dlist->used,sizeof(*sg_beg));
    sort_gobj_t *sg_end = sg_beg + dlist->used;
    sort_gobj_t *sg;
    for ( sg = sg_beg; sg < sg_end; sg++, gobj++ )
    {
	const typeof(gobj->obj_id) obj_id = gobj->obj_id;
	sg->pos  = sg - sg_beg;
	sg->rank = obj_id < N_KMP_GOBJ ? rank[obj_id] : obj_id;
	if ( sort_mode > 1 )
	{
	    sg->angle = atan2(gobj->position[0],gobj->position[2]) * (180000000/M_PI);
	    if ( sg->angle < 0)
		sg->angle += 360000000;
	}
	memcpy(&sg->data,gobj,sizeof(sg->data));
    }

    //--- sort data

    qsort( sg_beg, dlist->used, sizeof(*sg_beg), (qsort_func)sort_gobj );


    //--- store sorted data

    gobj = (kmp_gobj_entry_t*)dlist->list;
    for ( sg = sg_beg; sg < sg_end; sg++, gobj++ )
    {
	PRINT("SORT: %5d %10d %5d\n",sg->rank,sg->angle,sg->pos);
	memcpy(gobj,&sg->data,sizeof(*gobj));
    }

    //--- terminate

    FREE(sg_beg);
    return true;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////		   Startpoint Calculations		///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct kmp_start_pos_t
{
    double x;		// relative x position, range -800.0 .. +800.0
    double z;		// relative z position, <0.0 if unknown
    double z_narrow;	// narrow z position, <0.0 if unknown
}
kmp_start_pos_t;

static const kmp_start_pos_t StartPosTab[] =
{
	{    0.0,    0.00,    0.00 },	//  1/1

	{ -400.0,    0.00,    0.00 },	//  1/2
	{  400.0,  461.88,  373.06 },	//  2/2

	{ -800.0,    0.00,    0.00 },	//  1/3
	{    0.0,  461.88,  373.06 },	//  2/3
	{  800.0,  923.75,  746.12 },	//  3/3

	{ -800.0,    0.00,    0.00 },	//  1/4
	{  400.0,  692.81,  559.56 },	//  2/4
	{ -400.0, 1530.94, 1386.56 },	//  3/4
	{  800.0, 2223.75, 1946.12 },	//  4/4

	{ -800.0,    0.00,    0.00 },	//  1/5
	{    0.0,  461.88,  373.06 },	//  2/5
	{  800.0,  923.75,  746.12 },	//  3/5
	{ -400.0, 1530.94, 1386.56 },	//  4/5	unsicher
	{  400.0, 1992.81, 1759.62 },	//  5/5	unsicher

	{ -800.0,    0.00,    0.00 },	//  1/6
	{ -160.0,  369.50,  298.44 },	//  2/6
	{  480.0,  739.00,  596.88 },	//  3/6
	{ -480.0, 1484.75, 1349.19 },	//  4/6
	{  160.0, 1854.25, 1647.62 },	//  5/6
	{  800.0, 2223.75, 1946.06 },	//  6/6

	{ -400.0,    0.00,    0.00 },	//  1/7		??? narrow geraten
	{  400.0,  461.88,  373.06 },	//  2/7
	{ -800.0, 1069.06, 1013.50 },	//  3/7	unsicher
	{    0.0, 1530.94, 1386.56 },	//  4/7	unsicher
	{  800.0, 1992.81, 1759.62 },	//  5/7	unsicher
	{ -400.0, 2900.00, 2650.00 },	//  6/7
	{  400.0, 3361.88, 3023.06 },	//  7/7

	{ -800.0,    0.00,    0.00 },	//  1/8
	{    0.0,  461.88,  373.06 },	//  2/8
	{  800.0,  923.75,  746.12 },	//  3/8
	{ -400.0, 1530.94, 1386.56 },	//  4/8	unsicher
	{  400.0, 1992.81, 1759.62 },	//  5/8	unsicher
	{ -800.0, 2900.00, 2650.00 },	//  6/8	unsicher
	{    0.0, 3361.88, 3023.06 },	//  7/8
	{  800.0, 3823.75, 3396.12 },	//  8/8

	{ -800.0,    0.00,    0.00 },	//  1/9
	{ -160.0,  369.50,  298.44 },	//  2/9
	{  480.0,  739.00,  596.88 },	//  3/9
	{ -480.0, 1484.75, 1349.19 },	//  4/9
	{  160.0, 1854.25, 1647.62 },	//  5/9
	{  800.0, 2223.75, 1946.07 },	//  6/9
	{ -800.0, 2900.00, 2650.00 },	//  7/9
	{ -160.0, 3269.50, 2948.44 },	//  8/9
	{  480.0, 3639.00, 3246.88 },	//  9/9

	{ -800.0,    0.00,    0.00 },	//  1/10
	{    0.0,  461.88,  373.06 },	//  2/10
	{  800.0,  923.75,  746.13 },	//  3/10
	{ -400.0, 1530.94, 1386.56 },	//  4/10
	{  400.0, 1992.82, 1759.56 },	//  5/10
	{ -800.0, 2900.00, 2650.00 },	//  6/10
	{    0.0, 3361.88, 3023.06 },	//  7/10
	{  800.0, 3823.75, 3396.12 },	//  8/10
	{ -400.0, 4430.94, 4036.56 },	//  9/10
	{  400.0, 4892.82, 4409.56 },	// 10/10

	{ -800.0,    0.00,    0.00},	//  1/11
	{ -160.0,  369.50,  298.44 },	//  2/11
	{  480.0,  739.00,  596.88 },	//  3/11
	{ -480.0, 1484.75, 1349.19 },	//  4/11
	{  160.0, 1854.25, 1647.62 },	//  5/11
	{  800.0, 2223.75, 1946.07 },	//  6/11
	{ -800.0, 2900.00, 2650.00 },	//  7/11
	{ -160.0, 3269.50, 2948.44 },	//  8/11
	{  480.0, 3639.00, 3246.88 },	//  9/11
	{ -480.0, 4384.75, 3999.21 },	// 10/11
	{  160.0, 4754.25, 4297.62 },	// 11/11

	{ -800.0,    0.00,    0.00},	//  1/12
	{ -160.0,  369.50,  298.44 },	//  2/12
	{  480.0,  739.00,  596.88 },	//  3/12
	{ -480.0, 1484.75, 1349.19 },	//  4/12
	{  160.0, 1854.25, 1647.62 },	//  5/12
	{  800.0, 2223.75, 1946.07 },	//  6/12
	{ -800.0, 2900.00, 2650.00 },	//  7/12
	{ -160.0, 3269.50, 2948.44 },	//  8/12
	{  480.0, 3639.00, 3246.88 },	//  9/12
	{ -480.0, 4384.75, 3999.21 },	// 10/12
	{  160.0, 4754.25, 4297.62 },	// 11/12
	{  800.0, 5123.75, 4596.06 },	// 12/12
};

//-----------------------------------------------------------------------------

uint kmp_start_pos_mirror = false;
uint kmp_start_pos_narrow = false;
bool kmp_start_pos_auto   = false;

const double3 * GetStartPosKMP
(
    // Return NULL on failure.
    // Otherwise return a pointer to 'tab' or to an internal static buffer
    // with N elements. This buffer is overwritten on next call.

    double3	*tab,		// table with 'n' elements to store the result.
				//   if NULL: use internal static buffer
    uint	n,		// number of players: 1..12
    bool	mirror,		// true: mirror the data (pole is right)
    bool	narrow,		// true: use 'narrow' mode

    kmp_t	*kmp,		// not NULL: Use KTPT/#0 and STGI/#0
    uint	automode	// bit fieled: Use KMP to override parameters
				//  1: return absolute positions, if KMP available
				//  2: use pole position of KMP if available
				//  4: use narrow mode of KMP if available
)
{
    //--- valid 'n'?

    if ( n < 1 || n > 12 )
	return 0;


    //--- auto mode

    kmp_start_pos_auto   = 0;
    kmp_start_pos_mirror = mirror;
    kmp_start_pos_narrow = narrow;

    if ( kmp && kmp->dlist[KMP_STGI].used )
    {
	const kmp_stgi_entry_t *stgi = (kmp_stgi_entry_t*)kmp->dlist[KMP_STGI].list;
	if ( automode & 2 )
	{
	    kmp_start_pos_auto |= 2;
	    kmp_start_pos_mirror = stgi->pole_pos;
	    mirror = kmp_start_pos_mirror == 1;
	}
	if ( automode & 4 )
	{
	    kmp_start_pos_auto |= 4;
	    kmp_start_pos_narrow = stgi->narrow_start;
	    narrow = kmp_start_pos_narrow == 1;
	}
    }


    //--- assign table values

    static double3 itab[12];
    if (!tab)
	tab = itab;
    double3 *d;
    const kmp_start_pos_t *sp = StartPosTab + n * ( n-1 ) / 2;

    uint i;
    for ( i = 0, d = tab; i < n; i++, d++, sp++ )
    {
	d->x = mirror ? -sp->x : sp->x;
	d->y = 0.0;
	d->z = narrow ? sp->z_narrow : sp->z;
	if (IsNormalD(d->z))
	    d->z = ldexp(trunc(ldexp(d->z,4)+0.5),-4);
    }


    //--- valid KMP/KTPT?

    if ( !(automode&1) || !kmp || !kmp->dlist[KMP_KTPT].used )
	return tab;
    kmp_start_pos_auto |= 1;


    //--- transformation step 1: y-rotate by 180 degree to get normalized vectors

    MatrixD_t matrix = {0};
    InitializeMatrixD(&matrix);
    SetRotateMatrixD(&matrix,1,180.0,0.0,0);
    TransformD3NMatrixD(&matrix,tab,n,sizeof(*tab));


    //--- transformation step 2: rotate and translate by start point settings

    const kmp_ktpt_entry_t * ktpt = (kmp_ktpt_entry_t*)kmp->dlist[KMP_KTPT].list;
    SetRotateMatrixD(&matrix,0,ktpt->rotation[0],0.0,0);
    SetRotateMatrixD(&matrix,1,ktpt->rotation[1],0.0,0);
    SetRotateMatrixD(&matrix,2,ktpt->rotation[2],0.0,0);
    double3 temp;
    temp.x = ktpt->position[0];
    temp.y = ktpt->position[1];
    temp.z = ktpt->position[2];
    SetTranslateMatrixD(&matrix,&temp);
    TransformD3NMatrixD(&matrix,tab,n,sizeof(*tab));

    return tab;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			Transform KMP			///////////////
///////////////////////////////////////////////////////////////////////////////

bool TransformKMP
(
    kmp_t		* kmp		// pointer to valid KMP
)
{
    if ( have_patch_count <= 0 || !transform_active || disable_transformation > 0 )
	return false;

    KMP_ACTION_LOG(true,"Transform KMP\n");
    PRINT("** TransformKMP() **\n");
    int n;

    //--- AREA

    n = kmp->dlist[KMP_AREA].used;
    if (n)
    {
	kmp_area_entry_t * area = (kmp_area_entry_t*)kmp->dlist[KMP_AREA].list;
	if ( KMP_TFORM & KMPTF_AREA_POS )
	    TransformPosFloat3D(area->position,n,sizeof(*area));
	if ( KMP_TFORM & KMPTF_AREA_ROTATE )
	    TransformRotFloat3D(area->rotation,n,sizeof(*area));
	if ( KMP_TFORM & KMPTF_AREA_SCALE )
	    TransformScaleFloat3D(area->scale,n,sizeof(*area));
    }

    //--- CAME

    n = kmp->dlist[KMP_CAME].used;
    if ( n && KMP_TFORM & KMPTF_CAME_POS )
    {
	kmp_came_entry_t * came = (kmp_came_entry_t*)kmp->dlist[KMP_CAME].list;
	TransformPosFloat3D(came->position,n,sizeof(*came));
	TransformPosFloat3D(came->viewpt_begin,n,sizeof(*came));
	TransformPosFloat3D(came->viewpt_end,n,sizeof(*came));
    }

    //--- CKPT

    n = kmp->dlist[KMP_CKPT].used;
    if ( n && KMP_TFORM & KMPTF_CKPT_POS )
    {
	kmp_ckpt_entry_t * ckpt = (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list;
	TransformPosFloat2D(ckpt->left,n,sizeof(*ckpt));
	TransformPosFloat2D(ckpt->right,n,sizeof(*ckpt));
    }

    //--- CNPT

    n = kmp->dlist[KMP_CNPT].used;
    if (n)
    {
	kmp_cnpt_entry_t * cnpt = (kmp_cnpt_entry_t*)kmp->dlist[KMP_CNPT].list;
	if ( KMP_TFORM & KMPTF_CNPT_POS )
	    TransformPosFloat3D(cnpt->position,n,sizeof(*cnpt));
	if ( KMP_TFORM & KMPTF_CNPT_ROTATE )
	    TransformRotFloat3D(cnpt->rotation,n,sizeof(*cnpt));
    }

    //--- ENPT

    n = kmp->dlist[KMP_ENPT].used;
    if (n)
    {
	kmp_enpt_entry_t * enpt = (kmp_enpt_entry_t*)kmp->dlist[KMP_ENPT].list;
	if ( KMP_TFORM & KMPTF_ENPT_POS )
	    TransformPosFloat3D(enpt->position,n,sizeof(*enpt));
	if ( KMP_TFORM & KMPTF_ENPT_SCALE )
	    TransformScaleFloat1D(&enpt->scale,n,sizeof(*enpt));
    }

    //--- GOBJ

    n = kmp->dlist[KMP_GOBJ].used;
    if (n)
    {
	kmp_gobj_entry_t * gobj = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	if ( KMP_TFORM & KMPTF_GOBJ_POS )
	    TransformPosFloat3D(gobj->position,n,sizeof(*gobj));
	if ( KMP_TFORM & KMPTF_GOBJ_ROTATE )
	    TransformRotFloat3D(gobj->rotation,n,sizeof(*gobj));
	if ( KMP_TFORM & KMPTF_GOBJ_SCALE )
	    TransformScaleFloat3D(gobj->scale,n,sizeof(*gobj));
    }

    //--- ITPT

    n = kmp->dlist[KMP_ITPT].used;
    if (n)
    {
	kmp_itpt_entry_t * itpt = (kmp_itpt_entry_t*)kmp->dlist[KMP_ITPT].list;
	if ( KMP_TFORM & KMPTF_ITPT_POS )
	    TransformPosFloat3D(itpt->position,n,sizeof(*itpt));
	if ( KMP_TFORM & KMPTF_ITPT_SCALE )
	    TransformScaleFloat1D(&itpt->scale,n,sizeof(*itpt));
    }

    //--- JGPT

    n = kmp->dlist[KMP_JGPT].used;
    if (n)
    {
	kmp_jgpt_entry_t * jgpt = (kmp_jgpt_entry_t*)kmp->dlist[KMP_JGPT].list;
	if ( KMP_TFORM & KMPTF_JGPT_POS )
	    TransformPosFloat3D(jgpt->position,n,sizeof(*jgpt));
	if ( KMP_TFORM & KMPTF_JGPT_ROTATE )
	    TransformRotFloat3D(jgpt->rotation,n,sizeof(*jgpt));
    }

    //--- KTPT

    n = kmp->dlist[KMP_KTPT].used;
    if (n)
    {
	kmp_ktpt_entry_t * ktpt = (kmp_ktpt_entry_t*)kmp->dlist[KMP_KTPT].list;
	if ( KMP_TFORM & KMPTF_KTPT_POS )
	    TransformPosFloat3D(ktpt->position,n,sizeof(*ktpt));
	if ( KMP_TFORM & KMPTF_KTPT_ROTATE )
	    TransformRotFloat3D(ktpt->rotation,n,sizeof(*ktpt));
    }

    //--- MSPT

    n = kmp->dlist[KMP_MSPT].used;
    if (n)
    {
	kmp_mspt_entry_t * mspt = (kmp_mspt_entry_t*)kmp->dlist[KMP_MSPT].list;
	if ( KMP_TFORM & KMPTF_MSPT_POS )
	    TransformPosFloat3D(mspt->position,n,sizeof(*mspt));
	if ( KMP_TFORM & KMPTF_MSPT_ROTATE )
	    TransformRotFloat3D(mspt->rotation,n,sizeof(*mspt));
    }

    //--- POTI

    n = kmp->poti_point.used;
    if ( n && KMP_TFORM & KMPTF_POTI_POS )
    {
	kmp_poti_point_t * pp = (kmp_poti_point_t*)kmp->poti_point.list;
	TransformPosFloat3D(pp->position,n,sizeof(*pp));
    }

    //--- STGI
    // nothing to do

    return true;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  Patch KMP helpers		///////////////
///////////////////////////////////////////////////////////////////////////////

static u8 rm_kmp_gobj[N_KMP_GOBJ] = {0};
static bool have_rm_kmp_gobj = false;

///////////////////////////////////////////////////////////////////////////////

int ScanOptRmGobj ( ccp arg )
{
    if (have_rm_kmp_gobj)
	have_patch_count--, have_kmp_patch_count--;
    have_rm_kmp_gobj = false;
    memset(rm_kmp_gobj,0,sizeof(rm_kmp_gobj));
    if (!arg)
	return 0;

    ScanInfo_t si;
    InitializeSI(&si,arg,strlen(arg),"Option --rm-gobj",0);
    si.predef = SetupVarsKMP();
    ScanFile_t *sf = si.cur_file;
    DASSERT(sf);
    sf->disable_comma++;
    for(;;)
    {
	while ( *sf->ptr > 0 && *sf->ptr <= ' ' || *sf->ptr == ',' )
	    sf->ptr++;
	if (!*sf->ptr)
	    break;

	Var_t val;
	if (ScanExprSI(&si,&val))
	    goto abort;
	int idx1 = GetIntV(&val);
	int idx2 = idx1;
	if ( NextCharSI(&si,false) == ':' )
	{
	    sf->ptr++;
	    if (ScanExprSI(&si,&val))
		goto abort;
	    idx2 = GetIntV(&val);
	}

	if ( idx1 < 0 )
	     idx1 = 0;
	if ( idx2 > N_KMP_GOBJ - 1 )
	     idx2 = N_KMP_GOBJ - 1;

	while ( idx1 <= idx2 )
	{
	    rm_kmp_gobj[idx1++] = 1;
	    have_rm_kmp_gobj = true;
	}
    }

    if (have_rm_kmp_gobj)
	have_patch_count++, have_kmp_patch_count++;

    #if defined(DEBUG) && defined(TEST)
    {
	PRINT("## have_rm_kmp_gobj = %d [%d,%d]\n",
		have_rm_kmp_gobj, have_patch_count, have_kmp_patch_count );
	uint i;
	for ( i = 0; i < N_KMP_GOBJ; i++ )
	    if (rm_kmp_gobj[i])
		PRINT("%6u,0x%02x := 0x%02x\n", i, i, rm_kmp_gobj[i] );
    }
    #endif

    return 0;

 abort:
    memset(rm_kmp_gobj,0,sizeof(rm_kmp_gobj));
    return 1;
};

///////////////////////////////////////////////////////////////////////////////

static bool FixPH
(
    kmp_t		* kmp,		// KMP data structure
    uint		sect_ph,	// KMP PH section index
    uint		sect_pt,	// related PT section index
    bool		allow_dup	// true: allow multiple links into same section
)
{
    DASSERT(kmp);
    DASSERT( sect_pt < KMP_N_SECT );
    DASSERT( sect_ph < KMP_N_SECT );

    //--- setup

    PRINT("FixPH(%s,%s,%d)\n",
		kmp_section_name[sect_ph].name1,
		kmp_section_name[sect_pt].name1, allow_dup );

    List_t *ph_list = kmp->dlist + sect_ph;
    List_t *pt_list = kmp->dlist + sect_pt;
    const uint max_pt = pt_list->used;


    //--- save in-data

    const uint saved_n = ph_list->used;
    const uint saved_size = saved_n * ph_list->elem_size;
    u8 *saved_data = MALLOC(saved_size);
    memcpy(saved_data,ph_list->list,saved_size);


    //--- first loop: find valid groups

    u8 linkref[0x100];
    memset(linkref,0xff,sizeof(linkref));

    uint pi, idx = 0;
    kmp_enph_entry_t *ph, *dest, *ph0 = (kmp_enph_entry_t*)ph_list->list;
    for ( pi = 0, ph = dest = ph0; pi < ph_list->used; pi++, ph++ )
    {
	uint end = ph->pt_start + ph->pt_len;
	if ( end > max_pt || pi == ph_list->used-1 )
	    end = max_pt;
	if ( end > idx )
	{
	    dest->pt_start = idx;
	    dest->pt_len   = end - idx;
	    idx = end;
	    memset(dest->prev,0xff,sizeof(dest->prev));
	    memcpy(dest->next,ph->next,sizeof(dest->next));
	    linkref[pi] = dest - ph0;
	    dest++;
	}
    }
    ph_list->used = dest - ph0;


    //--- second loop: analyse next links

    for ( pi = 0, ph = ph0; pi < ph_list->used; pi++, ph++ )
    {
	u8 new_next[KMP_MAX_PH_LINK];
	memset(new_next,0xff,sizeof(new_next));

	uint il;
	for ( il = 0; il < KMP_MAX_PH_LINK; il++ )
	{
	    const u8 link = linkref[ph->next[il]&0xff];
	    if ( link < ph_list->used )
	    {
		InsertLink(new_next,link,0,0,allow_dup);
		InsertLink(ph0[link].prev,pi,0,0,false);
	    }
	}

	if ( new_next[0] == 0xff )
	{
	    u8 link = pi + 1;
	    if ( link == ph_list->used )
		link = 0;
	    InsertLink(new_next,link,0,0,allow_dup);
	    InsertLink(ph0[link].prev,pi,0,0,false);
	}

	memcpy(ph->next,new_next,sizeof(ph->next));
    }


    //--- third loop: find empty 'prev' lists

    for ( pi = 0, ph = ph0; pi < ph_list->used; pi++, ph++ )
    {
	if ( ph->prev[0] == 0xff )
	{
	    const u8 link = pi ? pi - 1 : ph_list->used - 1;
	    InsertLink(ph->prev,link,0,0,false);
	}
    }


    //--- calc dirty

    const bool dirty = saved_n != ph_list->used
			|| memcmp(saved_data,ph_list->list,saved_size);
    FREE(saved_data);

    if (dirty)
	KMP_ACTION_LOG(true,"Section %s fixed.\n",kmp_section_name[sect_ph].name1);
    return dirty;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  Patch KMP			///////////////
///////////////////////////////////////////////////////////////////////////////

static bool ReversePT
(
    const kmp_t		* kmp,		// KMP data structure
    uint		sect		// KMP PT section index
)
{
    DASSERT(kmp);
    DASSERT( sect == KMP_ENPT || sect == KMP_ITPT );

    const List_t *ptlist, *phlist;
    if ( sect == KMP_ENPT )
    {
	ptlist = kmp->dlist + KMP_ENPT;
	phlist = kmp->dlist + KMP_ENPH;
    }
    else
    {
	DASSERT( sect == KMP_ITPT );
	ptlist = kmp->dlist + KMP_ITPT;
	phlist = kmp->dlist + KMP_ITPH;
    }

    //--- reverse points

    const uint np = ptlist->used;
    if ( np < 2 )
	return false;

    kmp_enpt_entry_t *p1 = (kmp_enpt_entry_t*)ptlist->list;
    kmp_enpt_entry_t *p2 = p1 + np - 1;
    while ( p1 < p2 )
    {
	kmp_enpt_entry_t temp = *p1;
	*p1++ = *p2;
	*p2-- = temp;
    }

    //--- reverse groups

    const uint ng = phlist->used;
    const uint ng1 = ng - 1;
    kmp_enph_entry_t *g1 = (kmp_enph_entry_t*)phlist->list;
    kmp_enph_entry_t *g2 = g1 + ng1;
    while ( g1 < g2 )
    {
	kmp_enph_entry_t temp = *g1;
	*g1++ = *g2;
	*g2-- = temp;
    }

    //--- fix groups

    g1 = (kmp_enph_entry_t*)phlist->list;
    g2 = g1 + ng;
    for ( ; g1 < g2; g1++ )
    {
	kmp_enph_entry_t temp = *g1;

	uint i;
	for ( i = 0; i < KMP_MAX_PH_LINK; i++ )
	{
	    g1->prev[i] = temp.next[i] == 0xff ? 0xff : ng1 - temp.next[i];
	    g1->next[i] = temp.prev[i] == 0xff ? 0xff : ng1 - temp.prev[i];
	}

	g1->pt_start = np - g1->pt_start - g1->pt_len;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////

static bool ReversePOTI
(
    const kmp_t		* kmp		// KMP data structure
)
{
    DASSERT(kmp);

    const uint ng = kmp->dlist[KMP_POTI].used;
    kmp_poti_group_t *pg = (kmp_poti_group_t*)kmp->dlist[KMP_POTI].list;
    const uint np = kmp->poti_point.used;
    kmp_poti_point_t *pp = (kmp_poti_point_t*)kmp->poti_point.list;
    
    uint pi, gi;
    for ( gi = pi = 0; gi < ng; gi++ )
    {
	uint en = pg[gi].n_point;
	if ( en > np - pi )
	     en = np - pi;

	kmp_poti_point_t *p1 = pp + pi;
	kmp_poti_point_t *p2 = p1 + en - 1;
	PRINT("==> POTI %02u: %d..%d /%u\n",gi,(int)(p1-pp),(int)(p2-pp),np);
	pi += en;

	while ( p1 < p2 )
	{
	    kmp_poti_point_t temp = *p1;
	    *p1++ = *p2;
	    *p2-- = temp;
	}
    }

    return false;
}

///////////////////////////////////////////////////////////////////////////////

static bool ReverseCKPT
(
    const kmp_t		* kmp		// KMP data structure
)
{
    DASSERT(kmp);
    // [[2do]] ???
    return false;
}

///////////////////////////////////////////////////////////////////////////////

static bool ReverseKTPT
(
    const kmp_t		* kmp		// KMP data structure
)
{
    DASSERT(kmp);
    // [[2do]] ???
    return false;
}

///////////////////////////////////////////////////////////////////////////////

static bool ReverseJGPT
(
    const kmp_t		* kmp		// KMP data structure
)
{
    DASSERT(kmp);
    // [[2do]] ???
    return false;
}

///////////////////////////////////////////////////////////////////////////////

static bool ReverseKMP
(
    kmp_t		* kmp		// pointer to valid KMP
)
{
    DASSERT(kmp);
    return ReversePT(kmp,KMP_ENPT)
	 | ReversePT(kmp,KMP_ITPT)
	 | ReversePOTI(kmp)
	 | ReverseCKPT(kmp)	// [[2do]]
	 | ReverseKTPT(kmp)	// [[2do]]
	 | ReverseJGPT(kmp);	// [[2do]]
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  Patch KMP			///////////////
///////////////////////////////////////////////////////////////////////////////

bool PatchKMP
(
    kmp_t		* kmp		// pointer to valid KMP
)
{
    DASSERT(kmp);
    bool dirty = ( KMP_MODE & KMPMD_NEW ) != 0;
    if ( have_patch_count <= 0 || have_kmp_patch_count <= 0 && !transform_active )
	return dirty;

    KMP_ACTION_LOG(false,"PatchKMP()\n");
    PRINT("** PatchKMP() **\n");

    dirty |= TransformKMP(kmp);

    
    //--- remove global objects (--rm-gobj, --slot)

    if ( opt_slot & SLOTMD_RM_ICE )
    {
	rm_kmp_gobj[GOBJ_ICE] = 1;
	have_rm_kmp_gobj = true;
    }
    if ( opt_slot & SLOTMD_RM_SUNDS )
    {
	rm_kmp_gobj[GOBJ_SUN_DS] = 1;
	rm_kmp_gobj[GOBJ_PYLON01] = 1;
	have_rm_kmp_gobj = true;
    }
    if ( opt_slot & SLOTMD_RM_SHYGUY )
    {
	rm_kmp_gobj[GOBJ_HEYHO_SHIP] = 1;
	rm_kmp_gobj[GOBJ_HEYHO_BALL] = 1;
	have_rm_kmp_gobj = true;
    }

    if (have_rm_kmp_gobj)
    {
	kmp_gobj_entry_t *glist = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	kmp_gobj_entry_t *src  = glist;
	kmp_gobj_entry_t *dest = glist;
	kmp_gobj_entry_t *end  = glist + kmp->dlist[KMP_GOBJ].used;
	for ( ; src < end; src++ )
	{
	    if (!rm_kmp_gobj[src->obj_id])
	    {
		memcpy(dest,src,sizeof(*dest));
		dest++;
	    }
	    else
		RemoveIL(kmp->index+KMP_GOBJ,dest-glist);
	}

	if ( dest < end )
	{
	    dirty = true;
	    kmp->dlist[KMP_GOBJ].used = dest - glist;
	}
    }


    //--- SLOTMD_ADD_SHYGUY

    if ( opt_slot & (SLOTMD_ADD_GICE|SLOTMD_ADD_SHYGUY) )
    {
	const kmp_gobj_entry_t *glist = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	const kmp_gobj_entry_t *go, *gend  = glist + kmp->dlist[KMP_GOBJ].used;
	bool ice_count = 0, shuyguy_count = 0;
	for ( go = glist; go < gend; go++ )
	{
	    if ( go->obj_id == GOBJ_ICE )
		ice_count++;
	    if ( go->obj_id == GOBJ_HEYHO_SHIP )
		shuyguy_count++;
	}

	if ( !ice_count && opt_slot & SLOTMD_ADD_GICE )
	{
	    PATCH_ACTION_LOG("Add","KMP","object ice\n");
	    AppendGOBJ(kmp,GOBJ_ICE);
	    dirty = true;
	}

	if ( !shuyguy_count && opt_slot & SLOTMD_ADD_SHYGUY )
	{
	    PATCH_ACTION_LOG("Add","KMP","object HeyhoShipGBA + route\n");

	    const uint faraway  = 150000;
	    const uint distance =   2000;
	    kmp_gobj_entry_t *go = AppendGOBJ(kmp,GOBJ_HEYHO_SHIP);
	    go->position[0] = go->position[2] = faraway;
	    go->position[1] = 1;
	    go->setting[0] = 20;
	    go->setting[1] = 100;
	    go->route_id = kmp->dlist[KMP_POTI].used;

	    kmp_poti_group_t * pg = AppendList(kmp->dlist+KMP_POTI);
	    pg->n_point = 3;

	    uint i;
	    for ( i = 0; i < 3; i++ )
	    {
		kmp_poti_point_t * pg = AppendList(&kmp->poti_point);
		pg->position[0] = faraway + ( i>0 ? distance : 0 );
		pg->position[1] = 1;
		pg->position[2] = faraway + ( i>1 ? distance : 0 );
	    }
	    dirty = true;
	}
    }


    //--- reverse track order: KMPMD_REVERSE

    if ( KMP_MODE & KMPMD_REVERSE )
	dirty |= ReverseKMP(kmp);


    //--- KMPMD_POLE_LEFT, KMPMD_POLE_RIGHT, KMPMD_SP_WIDE, KMPMD_SP_NARROW, KMPMD_*_LAPS

    if ( KMP_MODE & KMPMD_M_STGI || speed_mod_active )
    {
	int n = kmp->dlist[KMP_STGI].used;
	kmp_stgi_entry_t * stgi = (kmp_stgi_entry_t*)kmp->dlist[KMP_STGI].list;

	const uint new_lap_count = ( KMP_MODE & KMPMD_M_LAPS ) >> KMPMD_SHIFT_LAPS;
	bool laps_modified   = false;
	ccp  pole_modified   = 0;
	ccp  narrow_modified = 0;
	uint speed_modified  = false;
	u16  old_speed_mod = 0;

	uint i;
	for ( i = 0; i < n; i++, stgi++ )
	{
	    if ( KMP_MODE & KMPMD_M_LAPS && stgi->lap_count != new_lap_count )
	    {
		stgi->lap_count = new_lap_count;
		laps_modified = true;
	    }

	    if ( KMP_MODE & KMPMD_POLE_LEFT && stgi->pole_pos != 0 )
	    {
		stgi->pole_pos = 0;
		pole_modified = "left";
	    }
	    if ( KMP_MODE & KMPMD_POLE_RIGHT && stgi->pole_pos != 1 )
	    {
		stgi->pole_pos = 1;
		pole_modified = "right";
	    }

	    if ( KMP_MODE & KMPMD_SP_WIDE && stgi->narrow_start != 0 )
	    {
		stgi->narrow_start = 0;
		narrow_modified = "wide";
	    }
	    if ( KMP_MODE & KMPMD_SP_NARROW && stgi->narrow_start != 1 )
	    {
		stgi->narrow_start = 1;
		narrow_modified = "narrow";
	    }

	    if ( speed_mod_active && stgi->unknown_0a != speed_mod_val )
	    {
		old_speed_mod = stgi->unknown_0a;
		stgi->unknown_0a = speed_mod_val;
		speed_modified = true;
	    }
	}

	dirty |= laps_modified || pole_modified || narrow_modified || speed_modified;

	char lap_count_info[50] = {0};
	if (laps_modified)
	    snprintf(lap_count_info,sizeof(lap_count_info),
			" Lap counter set to '%u'.",new_lap_count);

	if ( pole_modified && narrow_modified )
	    KMP_ACTION_LOG(true,
		"STGI:%s Pole position set to '%s' and narrow mode to '%s'.\n",
		lap_count_info, pole_modified, narrow_modified );
	else if ( pole_modified )
	    KMP_ACTION_LOG(true,"STGI:%s Pole position set to '%s'.\n",
		lap_count_info, pole_modified );
	else if ( narrow_modified )
	    KMP_ACTION_LOG(true,"STGI:%s Narrow mode set to '%s'.\n",
		lap_count_info, narrow_modified);
	else if (laps_modified)
	    KMP_ACTION_LOG(true,"STGI:%s\n",lap_count_info);

	if (speed_modified)
	{
	    if ( n == 1 )
	    {
		if (speed_mod_val)
		    KMP_ACTION_LOG(true,
			"STGI: Speed modifier set from %5.3f to %5.3f (%04x->%04x/hex).\n",
			    SpeedMod2float(old_speed_mod), speed_mod_factor,
			    old_speed_mod, speed_mod_val );
		else
		    KMP_ACTION_LOG(true,
			"STGI: Speed modifier cleared (old: %5.3f = %04x/hex).\n",
			SpeedMod2float(old_speed_mod), old_speed_mod );
	    }
	    else
	    {
		if (speed_mod_val)
		    KMP_ACTION_LOG(true,"STGI: Speed modifier set to %5.3f (%04x/hex).\n",
			    speed_mod_factor, speed_mod_val );
		else
		    KMP_ACTION_LOG(true,"STGI: Speed modifier cleard.\n");
	    }
	}
    }


    //--- KMPMD_RM_SPCITEM

    if ( KMP_MODE & KMPMD_RM_SPCITEM )
    {
	const uint n = kmp->dlist[KMP_GOBJ].used;
	kmp_gobj_entry_t * gobj = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
	uint i, count = 0;
	for ( i = 0; i < n; i++, gobj++ )
	{
	    switch (gobj->obj_id)
	    {
		case 0x65: // itembox
		case 0x76: // s_itembox
		case 0xc9: // f_itembox
		case 0xd4: // w_itembox
		case 0xd5: // w_itemboxline
		case 0xee: // sin_itembox
		    if (gobj->setting[1])
		    {
			gobj->setting[1] = 0;
			count++;
		    }
		    break;
	    }
	}
	if (count)
	{
	    dirty = true;
	    KMP_ACTION_LOG(true,"%u special item%s for players removed.\n",
			count, count == 1 ? "" : "s" );
	}
    }


    //--- KMPMD_FIX_CKPH, KMPMD_FIX_ENPH, KMPMD_FIX_ITPH

    if ( KMP_MODE & KMPMD_FIX_CKPH )
	dirty |= FixPH(kmp,KMP_CKPH,KMP_CKPT,false);
    if ( KMP_MODE & KMPMD_FIX_ENPH )
	dirty |= FixPH(kmp,KMP_ENPH,KMP_ENPT,true);
    if ( KMP_MODE & KMPMD_FIX_ITPH )
	dirty |= FixPH(kmp,KMP_ITPH,KMP_ITPT,false);


    //--- KMPMD_FIX_CKNEXT & KMPMD_FIX_CKJGPT

    if ( KMP_MODE & (KMPMD_FIX_CKNEXT|KMPMD_FIX_CKJGPT) )
    {
	List_t *pt_list = kmp->dlist + KMP_CKPT;
	kmp_ckpt_entry_t *pt0 = (kmp_ckpt_entry_t*)pt_list->list;
	const uint max_pt = pt_list->used;

	if ( max_pt > 0 )
	{
	    if ( KMP_MODE & KMPMD_FIX_CKNEXT )
	    {
		// save data
		const uint saved_size = max_pt * pt_list->elem_size;
		u8 *saved_data = MALLOC(saved_size);
		memcpy(saved_data,pt_list->list,saved_size);

		uint ip;
		kmp_ckpt_entry_t *pt;
		for ( ip = 0, pt = pt0; ip < max_pt; ip++, pt++ )
		{
		    pt->prev = ip - 1;
		    pt->next = ip + 1;
		}
		pt0[max_pt-1].next = M1(pt0->next);

		uint ih;
		List_t *ph_list = kmp->dlist + KMP_CKPH;
		kmp_enph_entry_t *ph = (kmp_enph_entry_t*)ph_list->list;
		for ( ih = 0; ih < ph_list->used; ih++, ph++ )
		{
		    uint idx = ph->pt_start;
		    if ( idx < max_pt )
			pt0[idx].prev = M1(pt0->prev);

		    idx += ph->pt_len - 1;
		    if ( idx < max_pt )
			pt0[idx].next = M1(pt0->next);
		}

		const bool pt_dirty = memcmp(saved_data,pt_list->list,saved_size);
		FREE(saved_data);
		dirty |= pt_dirty;
		if (pt_dirty)
		    KMP_ACTION_LOG(true,"'prev' and 'next' of section CKPT fixed.\n");
	    }

	    if ( KMP_MODE & KMPMD_FIX_CKJGPT )
	    {
		const uint n_jgpt = kmp->dlist[KMP_JGPT].used;

		u8 last_valid = 0;
		uint ip, count = 0;
		kmp_ckpt_entry_t *pt;
		for ( ip = 0, pt = pt0; ip < max_pt; ip++, pt++ )
		{
		    if ( pt->respawn && pt->respawn >= n_jgpt )
		    {
			pt->respawn = last_valid;
			count++;
		    }
		    else
			last_valid = pt->respawn;
		}

		if (count)
		{
		    dirty = true;
		    KMP_ACTION_LOG(true,
			"Section CKPT: %u invalid respawn link%s fixed.\n",
			count, count==1 ? "" : "s" );
		}
	    }
	}
    }


    //--- term

    return dirty;
}

///////////////////////////////////////////////////////////////////////////////

bool PatchRawDataKMP
(
    void		* data,		// data to scan
    uint		data_size	// size of 'data'
)
{
    if ( have_patch_count <= 0 || have_kmp_patch_count <= 0 && !transform_active )
	return false;

    KMP_ACTION_LOG(false,"PatchRawDataKMP()\n");
    PRINT("** PatchRawDataKMP() **\n");

    kmp_t kmp;
    bool stat = false;
    enumError err = ScanRawKMP(&kmp,true,data,data_size);
    if ( !err && PatchKMP(&kmp) )
    {
	stat = !CreateRawKMP(&kmp) && kmp.raw_data_size <= data_size;
	if (stat)
	    memcpy(data,kmp.raw_data,kmp.raw_data_size);
    }
    ResetKMP(&kmp);
    return stat;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    CkeckKMP()			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct check_kmp_t
{
    //--- param

    const kmp_t		* kmp;		// pointer to valid KMP
    CheckMode_t		mode;		// print mode

    //--- status

    int			warn_count;	// number of warnings
    int			hint_count;	// number of hints
    int			info_count;	// number of hints
    bool		head_printed;	// true: header line printed
    bool		info_printed;	// true: info line printed
    int			sect;		// last section
    int			index;		// last object index

    //--- misc data

    ColorSet_t		col;		// color setup

} check_kmp_t;

///////////////////////////////////////////////////////////////////////////////

ccp print_name ( const kmp_t * kmp, uint sect, uint index )
{
    DASSERT(kmp);
    DASSERT( sect < KMP_N_SECT );

    // cast because remove const
    IndexList_t *ill = (IndexList_t*)kmp->index + sect;
    ccp name = index < ill->used ? GetNameIL(ill,index) : 0;
    if ( !name || !name )
	return "";

    const uint bufsize = 50;
    char *buf = GetCircBuf(bufsize);
    snprintf(buf,bufsize," '%.46s'",name);
    return buf;
}

///////////////////////////////////////////////////////////////////////////////

// predefine function to avoid a warning because of __attribute__ usage
static void print_check_error
(
    check_kmp_t		* chk,		// valid data structure
    CheckMode_t		print_mode,	// NULL | CMOD_HINT | CMOD_WARNING
    int			sect,		// section
    int			index,		// object index
    ccp			obj_type,	// NULL or object type
    ccp			format,		// format of message
    ...					// arguments

) __attribute__ ((__format__(__printf__,6,7)));

static void print_check_error
(
    check_kmp_t		* chk,		// valid data structure
    CheckMode_t		print_mode,	// NULL | CMOD_HINT | CMOD_WARNING
    int			sect,		// section
    int			index,		// object index
    ccp			obj_type,	// NULL or object type
    ccp			format,		// format of message
    ...					// arguments

)
{
    DASSERT(chk);
    DASSERT(chk->kmp);

    ccp col; // message color
    ccp bol; // begin of line
    switch (print_mode)
    {
	case CMOD_WARNING:
	    if (!( chk->mode & CMOD_WARNING ))
		return;
	    col = chk->col.warn;
	    bol = check_bowl;
	    chk->warn_count++;
	    break;

	case CMOD_HINT:
	    if (!( chk->mode & CMOD_HINT ))
		return;
	    col = chk->col.hint;
	    bol = check_bohl;
	    chk->hint_count++;
	    break;

	case CMOD_INFO:
	    if (!( chk->mode & CMOD_INFO ))
		return;
	    col = chk->col.info;
	    bol = check_boil;
	    chk->info_count++;
	    break;

	default:
	    col = 0;
	    bol = 0; // print head only
	    break;
    }

    if (!chk->head_printed)
    {
	chk->head_printed = true;
	if ( chk->mode & (CMOD_HEADER|CMOD_FORCE_HEADER) )
	    fprintf(stdlog,"%s* CHECK %s:%s%s\n",
			chk->col.heading, GetNameFF(chk->kmp->fform,0),
			chk->kmp->fname, chk->col.reset );
    }

    if (!bol)
	return;

    if (!chk->info_printed)
    {
	chk->info_printed = true;
	fprintf(stdlog,"  %s> All indices are zero based!%s\n",
			chk->col.info, chk->col.reset );
    }

    if ( chk->sect != sect || chk->index != index )
    {
	chk->sect  = sect;
	chk->index = index;

	if ( sect < 0 )
	    fprintf(stdlog,"  %s> Statistics%s\n",chk->col.info,chk->col.reset);
	else if ( index < 0 )
	    fprintf(stdlog,"  %s> Section %s%s%s\n",
			chk->col.info, kmp_section_name[sect].name1,
			index == -2 ? ", header" : "", chk->col.reset );
	else
	{
	    fprintf(stdlog,"  %s> Section %s, %s #%u of %u%s",
			chk->col.info,
			kmp_section_name[sect].name1,
			kmp_section_name[sect].name2,
			index, chk->kmp->dlist[sect].used,
			print_name(chk->kmp,sect,index) );

	    if (obj_type)
		fprintf(stdlog,", type '%s'%s\n",obj_type,chk->col.reset);
	    else
		fprintf(stdlog,"%s\n",chk->col.reset);
	}
    }

    if (format)
    {
	fputs(col,stdlog);
	fputs(bol,stdlog);
	fputs(chk->col.reset,stdlog);

	va_list arg;
	va_start(arg,format);
	vfprintf(stdlog,format,arg);
	va_end(arg);
    }
}

///////////////////////////////////////////////////////////////////////////////

static void print_item_hint
(
    check_kmp_t		* chk,		// valid data structure
    uint		n_total,	// total number of special item boxes
    uint		n_unsure,	// unsure number of special item boxes
    ccp			dest		// "players" or "enemies"
)
{
    DASSERT(chk);
    DASSERT(dest);

    if (n_total)
    {
	char buf[25];
	if (n_unsure)
	    snprintf(buf,sizeof(buf)," (%u unsure)",n_unsure);
	else
	    *buf = 0;

	if ( n_total > 1 )
	    print_check_error(chk,CMOD_HINT,KMP_GOBJ,-1,0,
		"%u item boxes give special items to %s%s.\n",
		n_total, dest, buf );
	else
	    print_check_error(chk,CMOD_HINT,KMP_GOBJ,-1,0,
		"One item box gives a special item to %s%s.\n",
		dest, buf );
    }
}

///////////////////////////////////////////////////////////////////////////////

static void check_pt
(
    check_kmp_t		*chk,		// valid data structure
    int			sect_pt,	// PT section
    kmp_path_t		*ck_path	// path data, already initialized
					//	data is calculated in first call
)
{
    DASSERT(chk);
    DASSERT(chk->kmp);
    DASSERT( sect_pt == KMP_ENPT || sect_pt == KMP_ITPT );
    DASSERT(ck_path);

    uint n = chk->kmp->dlist[sect_pt].used;
    if ( n > KMP_MAX_POINT )
	 n = KMP_MAX_POINT;
    const kmp_enpt_entry_t *pt = (kmp_enpt_entry_t*)chk->kmp->dlist[sect_pt].list;

    u8 fail[KMP_MAX_POINT];
    memset(fail,0,sizeof(fail));

    uint i, fail_count = 0;
    const kmp_link_t *last = 0;
    for ( i = 0; i < n; i++, pt++ )
    {
	const kmp_link_t *link
		= FindCKPT(chk->kmp,ck_path,pt->position[0],pt->position[2],last);
	if (link)
	    last = link;
	else
	    fail[i]++, fail_count++;
    }

    if (fail_count)
    {
	uint i;
	for ( i = 0; i < n; i++ )
	    if (fail[i])
	    {
		const int start = i;
		while ( ++i < n && fail[i] )
		    ;
		if ( --i == start )
		    print_check_error(chk,CMOD_WARNING,sect_pt,-1,0,
			"Point #%u is outside of all convex check quadrilaterals"
			" (pos bug?).\n",
			i );
		else
		{
		    print_check_error(chk,CMOD_WARNING,sect_pt,-1,0,
			"Points #%u..#%u are outside of all convex check quadrilaterals"
			" (pos bug?).\n",
			start, i );
		}
	    }
    }
}

///////////////////////////////////////////////////////////////////////////////

static void check_ph
(
    check_kmp_t		*chk,		// valid data structure
    int			sect_ph,	// PH section
    int			sect_pt,	// PT section
    kmp_path_t		*ck_path	// ckeck point path data, already initialized
					//	data is calculated in first call
)
{
    DASSERT(chk);
    const kmp_t *kmp = chk->kmp;
    DASSERT(kmp);

    const List_t *ph_list = kmp->dlist + sect_ph;
    const kmp_enph_entry_t *ph = (kmp_enph_entry_t*)ph_list->list;
    const uint n_ph = ph_list->used;

    const List_t *pt_list = kmp->dlist + sect_pt;
    const uint n_pt = pt_list->used;

    if (!n_ph)
    {
	print_check_error(chk,CMOD_WARNING,sect_ph,-1,0,
		"Empty section => track will freeze.\n" );
    }

    uint i, next_idx = 0;
    for ( i = 0; i < n_ph; i++, ph++ )
    {
	if ( ph->pt_start >= n_pt )
	    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
		"Route group starts at point %u, but max possible index is %u.\n",
		ph->pt_start, n_pt );
	else if ( ph->pt_start + ph->pt_len > n_pt )
	    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
		"Route group ends at point %u, but max possible index is %u.\n",
		ph->pt_start + ph->pt_len, n_pt );
	else if ( ph->pt_start != next_idx )
	    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
		"Route group starts at point %u,"
		" but %u is the expected is start point.\n",
		ph->pt_start, next_idx );
	next_idx = ph->pt_start + ph->pt_len;


	//--- test 'prev' links

	uint link, count;
	for ( link = count = 0; link < KMP_MAX_PH_LINK; link++ )
	{
	    if ( !IS_M1(ph->prev[link]) )
	    {
		if ( ph->prev[link] >= n_ph )
		    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
			"Invalid 'prev' link at index #%u:"
			" Link points to group %u, but max allowed is %u.\n",
			link, ph->prev[link], n_ph-1 );
		else
		    count++;
	    }
	}

	if (!count)
	    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
		"No valid 'prev' link found. Define at least one link.\n" );


	//--- test 'next' links

	for ( link = count = 0; link < KMP_MAX_PH_LINK; link++ )
	{
	    if ( !IS_M1(ph->next[link]) )
	    {
		if ( ph->next[link] >= n_ph )
		    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
			"Invalid 'next' link at index #%u:"
			" Link points to group %u, but max allowed is %u.\n",
			link, ph->next[link], n_ph-1 );
		else
		    count++;
	    }
	    else if ( !link && sect_ph == KMP_ITPH )
	    {
		print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
			"Missing 'next' link at index #0."
			" It's needed for the standard itemroute.\n" );
	    }
	}

	if (!count)
	    print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
		"No valid 'next' link found => track will freeze."
		" Define at least one link.\n" );
    }

    if ( sect_ph != KMP_CKPH )
    {
	// enemy and item routes
	check_pt(chk,sect_pt,ck_path);
	return;
    }


    //--- only CKPH/CKPT

    ph = (kmp_enph_entry_t*)ph_list->list;
    for ( i = 0; i < n_ph; i++, ph++ )
    {
	uint idx = ph->pt_start;
	uint end = idx + ph->pt_len;
	if ( end > n_pt )
	    end = n_pt;

	u8 done[0xff];
	memset(done,0,sizeof(done));

	uint j;
	const kmp_ckpt_entry_t *pt = ( (kmp_ckpt_entry_t*)pt_list->list ) + idx;
	for ( j = idx; j < end; j++, pt++ )
	{
	    const int good_prev = j == idx ? -1 : j - 1;
	    const int good_next = j == end-1 ? -1 : j + 1;
	    const int have_prev = IS_M1(pt->prev) ? -1 : pt->prev;
	    const int have_next = IS_M1(pt->next) ? -1 : pt->next;

	    if ( good_prev != have_prev )
		print_check_error(chk,CMOD_WARNING,sect_pt,j,0,
			"Invalid 'prev' link [have %d, should be %d]\n",
			have_prev, good_prev );

	    if ( good_next != have_next )
		print_check_error(chk,CMOD_WARNING,sect_pt,j,0,
			"Invalid 'next' link [have %d, should be %d]\n",
			have_next, good_next );

	    typeof(pt->mode) mode = pt->mode;
	    if ( !IS_M1(mode) && mode < sizeof(done)/sizeof(*done) )
		done[mode]++;
	}

	for ( j = 0; j < sizeof(done)/sizeof(*done); j++ )
	    if ( done[j] > 1 )
		print_check_error(chk,CMOD_WARNING,sect_ph,i,0,
		    "Check point mode %u used %u times in single group.\n",
		    j, done[j] );
    }
}

///////////////////////////////////////////////////////////////////////////////

void print_stat
(
    check_kmp_t		* chk,		// valid data structure
    int			sect_ph,	// -1 or PH section
    int			sect_pt,	// (PT) section
    bool		print_null,	// true: print if null pounts
    ccp			name_ph,	// name of ph
    ccp			name_pt		// name of pt
)
{
    DASSERT(chk);
    DASSERT(chk->kmp);
    DASSERT( sect_ph >= -1 && sect_ph < KMP_N_SECT );
    DASSERT( sect_pt >=  0 && sect_pt < KMP_N_SECT );
    DASSERT( sect_ph  <  0 || name_ph );
    DASSERT(name_pt);

    if ( sect_ph >= 0 )
    {
	const uint nh = chk->kmp->dlist[sect_ph].used;
	const uint nt = sect_pt == KMP_POTI
			? chk->kmp->poti_point.used
			: chk->kmp->dlist[sect_pt].used;
	print_check_error(chk,CMOD_INFO,-1,-1,0,
		"%3u %s%s for %u %s%s defined.\n",
		nt, name_pt, nt==1 ? "" : "s",
		nh, name_ph, nh==1 ? "" : "s" );
    }
    else
    {
	const uint np = chk->kmp->dlist[sect_pt].used;
	if ( np || print_null )
	    print_check_error(chk,CMOD_INFO,-1,-1,0,
			"%3u %s%s defined.\n", np, name_pt, np==1 ? "" : "s");
    }
}

///////////////////////////////////////////////////////////////////////////////

int CheckKMP
(
    // returns number of errors

    const kmp_t		* kmp,		// pointer to valid KMP
    CheckMode_t		mode		// print mode
)
{
    DASSERT(kmp);
    PRINT("CheckKMP(%d)\n",mode);

    //--- setup

    check_kmp_t chk;
    memset(&chk,0,sizeof(chk));
    SetupColorSet(&chk.col,stdlog);
    chk.kmp  = kmp;
    chk.mode = mode;
    chk.sect = -1;

    if ( mode & CMOD_FORCE_HEADER )
	print_check_error(&chk,0,0,0,0,0);


    //--- var defs

    uint i, n;
    char buf[50];

    u8 done[0x100];

    const uint route_usage_size = 0xff;
    u8 route_usage[route_usage_size];
    memset(route_usage,0,sizeof(route_usage));
    enum { RU_CAME = 1, RU_GOBJ = 2 };

    u16 jgpt_usage[0x100];
    memset(jgpt_usage,0,sizeof(jgpt_usage));

    const int ncame = kmp->dlist[KMP_CAME].used;
    const int npoti = kmp->dlist[KMP_POTI].used;
    const int njgpt = kmp->dlist[KMP_JGPT].used;


    //--- statistics

    if ( mode & CMOD_INFO )
    {
	print_stat(&chk, -1,       KMP_AREA, true,  0, "area");
	print_stat(&chk, -1,       KMP_CAME, true,  0, "camera");
	print_stat(&chk, KMP_ENPH, KMP_ENPT, true,  "group", "enemy point");
	print_stat(&chk, KMP_ITPH, KMP_ITPT, true,  "group", "item point");
	print_stat(&chk, KMP_CKPH, KMP_CKPT, true,  "group", "check point");
	print_stat(&chk, KMP_POTI, KMP_POTI, true,  "route", "route point");
	print_stat(&chk, -1,       KMP_GOBJ, true,  0, "global object");
	print_stat(&chk, -1,       KMP_JGPT, true,  0, "respawn position");
	print_stat(&chk, -1,       KMP_KTPT, true,  0, "starting position");
	print_stat(&chk, -1,       KMP_CNPT, false, 0, "cannon target");
	print_stat(&chk, -1,       KMP_MSPT, false, 0, "end battle position");
	print_stat(&chk, -1,       KMP_STGI, true,  0, "stage record");
    }


    //--- check AREA entries

    n = kmp->dlist[KMP_AREA].used;
    const kmp_area_entry_t * area = (kmp_area_entry_t*)kmp->dlist[KMP_AREA].list;
    for ( i = 0; i < n; i++, area++ )
    {
	if ( area->type != 0x00 ) // only CAME checked
	    continue;

	if ( area->dest_id == 0xff )
	    print_check_error(&chk,CMOD_WARNING,KMP_AREA,i,"CAME",
			"Missing camera link.\n");
	else if ( area->dest_id >= ncame )
	    print_check_error(&chk,CMOD_WARNING,KMP_AREA,i,"CAME",
			"Invalid link to not existing camera #%u (max=#%d).\n",
			area->dest_id, ncame-1 );
    }


    //--- check CAME entries

    i = kmp->value[KMP_CAME] >> 8;
    if ( i >= ncame )
	print_check_error(&chk,CMOD_WARNING,KMP_CAME,-2,0,
			"Invalid link to not existing opening camera #%u (max=#%d).\n",
			i, ncame-1 );
    memset(done,0,sizeof(done));
    while ( i < ncame && i < sizeof(done)/sizeof(*done) && !done[i]++ )
    {
	const kmp_came_entry_t * came = (kmp_came_entry_t*)kmp->dlist[KMP_CAME].list + i;
	if ( came->type != 4 && came->type != 5 )
	    print_check_error(&chk,CMOD_WARNING,KMP_CAME,i,"opening camera",
			"Opening camera type should be 4 or 5, but not %u.\n",
			came->type );
	i = came->next;
    }

    i = kmp->value[KMP_CAME] & 0xff;
    if ( i >= ncame )
	print_check_error(&chk,CMOD_WARNING,KMP_CAME,-2,0,
			"Invalid link to not existing selection camera #%u (max=#%d).\n",
			i, ncame-1 );

    const kmp_came_entry_t * came = (kmp_came_entry_t*)kmp->dlist[KMP_CAME].list;
    uint came0count = 0;
    for ( i = 0; i < ncame; i++, came++ )
    {
	snprintf(buf,sizeof(buf),"%u",came->type);

	const typeof(came->route) route = came->route;
	if (!IS_M1(route))
	{
	    if ( route >= npoti )
		print_check_error(&chk,CMOD_WARNING,KMP_CAME,i,buf,
			"Invalid link to not existing route #%u (max=#%d).\n",
			route, npoti-1 );

	    if ( route < route_usage_size )
		route_usage[route] |= RU_CAME;
	}

	switch (came->type)
	{
	    case 0:
		came0count++;
		// fall through
	#ifdef TEST // [[2do]] is this true?
	    case 1:
	    case 3:
	#endif
		if (!IS_M1(route))
		    print_check_error(&chk,CMOD_WARNING,KMP_CAME,i,buf,
			"Link to route #%u not needed.\n",route);
		break;

	#ifdef TEST // [[2do]] is this true?
	    case 2:
	#endif
	    case 5:
	    case 6:
		if (IS_M1(route))
		    print_check_error(&chk,CMOD_WARNING,KMP_CAME,i,buf,
			"Missing route link.\n");
		break;

	#ifndef TEST // [[2do]]
	    case 1:
	    case 2:
	    case 3:
	#endif
	    case 4:
		// may have a route
		break;

	    default:
		print_check_error(&chk,CMOD_WARNING,KMP_CAME,i,buf,
			"Unknown camera type %u.\n",came->type);
	}

	if ( came->next != 0xff && came->next >= ncame )
		print_check_error(&chk,CMOD_WARNING,KMP_CAME,i,buf,
			"Invalid next link to not existing index #%u (max=#%d).\n",
			came->next, ncame-1 );
    }

    if (!came0count)
	print_check_error(&chk,CMOD_WARNING,KMP_CAME,-1,"Camera type 0",
		"Missing a camera with type 0.\n" );
    else if ( came0count > 1 )
	print_check_error(&chk,CMOD_WARNING,KMP_CAME,-1,"Camera type 0",
		"Only one camera with type 0 expected, but %u exist.\n",
			came0count );


    //--- check *PH entries

    kmp_path_t ck_path;
    InitializeKmpPath(&ck_path);

    check_ph(&chk,KMP_ENPH,KMP_ENPT,&ck_path);
    check_ph(&chk,KMP_ITPH,KMP_ITPT,&ck_path);
    check_ph(&chk,KMP_CKPH,KMP_CKPT,&ck_path);

    ResetKmpPath(&ck_path);


    //--- check CKPT entries

    uint mode0count = 0, max_mode = 0;
    memset(done,0,sizeof(done));

    n  = kmp->dlist[KMP_CKPT].used;
    const kmp_ckpt_entry_t * ckpt = (kmp_ckpt_entry_t*)kmp->dlist[KMP_CKPT].list;
    for ( i = 0; i < n; i++, ckpt++ )
    {
	if ( ckpt->respawn < sizeof(jgpt_usage)/sizeof(*jgpt_usage) )
	    jgpt_usage[ckpt->respawn]++;

	typeof(ckpt->mode) mode = ckpt->mode;
	if ( !IS_M1(mode) && mode < sizeof(done)/sizeof(*done) )
	{
	    done[mode] = 1;
	    if ( max_mode < mode )
		 max_mode = mode;
	    if (!mode)
		mode0count++;

	}

	uint link;
	for ( link = 0; link < KMP_MAX_PH_LINK; link++ )
	{
	    int next = GetNextPointKMP(kmp,KMP_CKPT,i,link,true,-1,0,0,0);
	    if ( next >= 0 )
	    {
		noPRINT("CKPT: %3d.%d -> %3d\n",i,link,next);
		double min_dist;
		int stat = CheckConvexCKPT(kmp,i,next,&min_dist);

		if ( min_dist < KMP_DIST_WARN )
		{
		    print_check_error(&chk,CMOD_WARNING,KMP_CKPT,-1,0,
			    "Check points #%u and #%u are to near together (%5.3f)"
			    " (pos bug?).\n",
			    i, next, min_dist );
		    continue;
		}

		if ( min_dist < KMP_DIST_HINT )
		{
		    print_check_error(&chk,CMOD_HINT,KMP_CKPT,-1,0,
			    "Check points #%u and #%u are near together (%3.1f).\n",
			    i, next, min_dist );
		}

		if ( stat > 0 )
		{
		    print_check_error(&chk,CMOD_WARNING,KMP_CKPT,-1,0,
			    "Check points #%u and #%u do not form a convex quadrilateral"
			    " (pos bug?).\n",
			    i, next );
		}
	    }
	}
    }

    if (!mode0count)
	print_check_error(&chk,CMOD_WARNING,KMP_CKPT,-1,0,
		"No check point with mode 0 (lap counter).\n");
    else if ( mode0count > 1 && !kmp->multi_ckpt_mode_0_ok )
	print_check_error(&chk,CMOD_HINT,KMP_CKPT,-1,0,
		"%u check points with mode 0 (lap counter).\n",
		mode0count );

    for ( i = 1; i <= max_mode; i++ )
	if (!done[i])
	{
	    const int start = i;
	    while ( ++i <= max_mode && !done[i] )
		;
	    if ( --i == start )
		print_check_error(&chk,CMOD_HINT,KMP_CKPT,-1,0,
			"Missing mandatory check point with mode %u.\n",
		    i );
	    else
		print_check_error(&chk,CMOD_HINT,KMP_CKPT,-1,0,
			"Missing mandatory check points with modes %u..%u.\n",
		    start, i );
	}

    for ( i = njgpt; i < sizeof(jgpt_usage)/sizeof(*jgpt_usage); i++ )
	if (jgpt_usage[i])
	    print_check_error(&chk,CMOD_WARNING,KMP_CKPT,-1,0,
		"%u invalid link%s to not existing respawn point #%u (max=#%d).\n",
		jgpt_usage[i], jgpt_usage[i]>1 ? "s" : "", i, njgpt-1 );


    //--- check JGPT entries

    n = njgpt;
    if ( n > sizeof(jgpt_usage)/sizeof(*jgpt_usage) )
	 n = sizeof(jgpt_usage)/sizeof(*jgpt_usage);
    const kmp_jgpt_entry_t * jgpt = (kmp_jgpt_entry_t*)kmp->dlist[KMP_CKPT].list;
    for ( i = 0; i < n; i++, jgpt++ )
	if (!jgpt_usage[i])
	{
	    const int start = i;
	    while ( ++i < n && !jgpt_usage[i] )
		;
	    if ( --i == start )
		print_check_error(&chk,CMOD_HINT,KMP_JGPT,-1,0,
			"Respawn point #%u of %u%s not used/linked.\n",
			i, n, print_name(kmp,KMP_JGPT,i) );
	    else
	    {
		ccp t1 = print_name(kmp,KMP_JGPT,start);
		ccp t2 = print_name(kmp,KMP_JGPT,i);
		if ( *t1 && *t2 )
		{
		    t1++;
		    t2++;
		    print_check_error(&chk,CMOD_HINT,KMP_JGPT,-1,0,
			"Respawn points #%u..#%u of %u (%s..%s) not used/linked.\n",
			start, i, n, t1, t2 );
		}
		else
		    print_check_error(&chk,CMOD_HINT,KMP_JGPT,-1,0,
			"Respawn points #%u..#%u of %u not used/linked.\n",
			start, i, n );
	    }
	}

    //--- check GOBJ entries

    uint n_obj_0 = 0;
    uint n_pitem = 0, n_pitem_unsure = 0;
    uint n_eitem = 0, n_eitem_unsure = 0;

    n = kmp->dlist[KMP_GOBJ].used;
    if ( n >= 200 )
	print_check_error(&chk,CMOD_HINT,KMP_GOBJ,-1,0,
		"%u global object%s defined.\n",
		n, n==1 ? "" : "s" );

    const kmp_gobj_entry_t * gobj = (kmp_gobj_entry_t*)kmp->dlist[KMP_GOBJ].list;
    for ( i = 0; i < n; i++, gobj++ )
    {
	const uint id = gobj->obj_id;
	if ( id >= N_KMP_GOBJ || !ObjectInfo[id].name )
	{
	    if (!id)
		n_obj_0++;
	    else
		print_check_error(&chk,CMOD_WARNING,KMP_GOBJ,i,0,
			"Unknown object with ID 0x%x.\n",id);
	    continue;
	}

	const ObjectInfo_t *oi = ObjectInfo + id;
	const typeof(gobj->route_id) route = gobj->route_id;
	if (IS_M1(route))
	{
	    if ( oi->flags & OBF_ALWAYS_ROUTE )
		print_check_error(&chk,CMOD_WARNING,KMP_GOBJ,i,oi->name,
			"Missing route link.\n");
	}
	else
	{
	    if ( route >= npoti )
		print_check_error(&chk,CMOD_WARNING,KMP_GOBJ,i,oi->name,
			"Invalid link to not existing route #%u (max=#%d).\n",
			route, npoti-1 );

	    if ( oi->flags & OBF_NO_ROUTE )
		print_check_error(&chk,CMOD_WARNING,KMP_GOBJ,i,oi->name,
			"Link to route #%u not needed.\n",route);

	    if ( route < route_usage_size )
		route_usage[route] |= RU_GOBJ;
	}

	Itembox_t ibox;
	if (AnalyseItembox(&ibox,gobj))
	{
	    if ( ibox.item_player )
	    {
		n_pitem++;
		if ( ibox.unsure_player )
		    n_pitem_unsure++;
	    }

	    if ( ibox.item_enemy )
	    {
		n_eitem++;
		if ( ibox.unsure_enemy )
		    n_eitem_unsure++;
	    }
	}
    }

    if ( n_obj_0 > 0 )
	print_check_error(&chk,CMOD_HINT,KMP_GOBJ,-1,0,
		"%u object%s with ID 0 found.\n",
		n_obj_0, n_obj_0 > 1 ? "s" : "" );

    print_item_hint(&chk,n_pitem,n_pitem_unsure,"players");
    print_item_hint(&chk,n_eitem,n_eitem_unsure,"enemies");


    //--- check POTI entries

    uint ng = kmp->dlist[KMP_POTI].used;
    const kmp_poti_group_t * pg = (kmp_poti_group_t*)kmp->dlist[KMP_POTI].list;
    uint np = kmp->poti_point.used;
    const kmp_poti_point_t * pp = (kmp_poti_point_t*)kmp->poti_point.list;

    n = ng < route_usage_size ? ng : route_usage_size;
    uint gi;
    for ( gi = 0; gi < n; gi++ )
    {
	if (!route_usage[gi])
	{
	    const int start = gi;
	    while ( ++gi < n && !route_usage[gi] )
		;
	    if ( --gi == start )
		print_check_error(&chk,CMOD_HINT,KMP_POTI,-1,0,
		    "Route #%u of %u%s not used/linked.\n",
			gi, ng, print_name(kmp,KMP_POTI,gi) );
	    else
	    {
		ccp t1 = print_name(kmp,KMP_POTI,start);
		ccp t2 = print_name(kmp,KMP_POTI,gi);
		if ( *t1 && *t2 )
		{
		    t1++;
		    t2++;
		    print_check_error(&chk,CMOD_HINT,KMP_POTI,-1,0,
			"Routes #%u..#%u of %u (%s..%s) not used/linked.\n",
			    start, gi, ng, t1, t2 );
		}
		else
		    print_check_error(&chk,CMOD_HINT,KMP_POTI,-1,0,
			"Routes #%u..#%u of %u not used/linked.\n",
			    start, gi, ng );
	    }
	}
	else if ( route_usage[gi] == (RU_CAME|RU_GOBJ) )
	    print_check_error(&chk,CMOD_WARNING,KMP_POTI,-1,0,
		    "Route #%u of %u is linked by CAME and GOBJ (may force a freeze).\n",
		    gi,ng);
    }

    for ( gi = i = 0; gi < ng && i < np; gi++, pg++ )
    {
	uint en = pg->n_point;
	DASSERT( en <= np-i );

	const bool used_by_came = gi < route_usage_size && route_usage[gi] & RU_CAME;
	uint count = 0;
	uint ei;
	for ( ei = 0; ei < en; ei++, i++, pp++ )
	    if ( ei < en-1 && !pp->speed && used_by_came )
		count++;
	if (count)
	    print_check_error(&chk,CMOD_HINT,KMP_POTI,gi,0,
		"CAME route: %u route point%s of %u with zero speed.\n",
			count, count == 1 ? "" : "s", en );
    }


    //--- check STGI entries

    n = kmp->dlist[KMP_STGI].used;
    if (!n)
	print_check_error(&chk,CMOD_WARNING,KMP_STGI,-1,0,
		"No element defined.\n");
    else if (n>1)
	print_check_error(&chk,CMOD_WARNING,KMP_STGI,-1,0,
		"%u elements defined.\n", n );
    else
    {
	const kmp_stgi_entry_t *stgi = (kmp_stgi_entry_t*)kmp->dlist[KMP_STGI].list;

	if ( stgi->lap_count != 3 )
	    print_check_error(&chk,CMOD_HINT,KMP_STGI,0,"STGI",
		"Lap counter changed to %u.\n",stgi->lap_count);
	else if ( mode & CMOD_INFO )
	    print_check_error(&chk,CMOD_INFO,KMP_STGI,0,"STGI",
		"Lap counter is to %u.\n",stgi->lap_count);

	if ( stgi->pole_pos > 1 )
	    print_check_error(&chk,CMOD_WARNING,KMP_STGI,0,"STGI",
		"Undefined value for pole positions: %u\n",stgi->pole_pos);
	else if ( mode & CMOD_INFO )
	    print_check_error(&chk,CMOD_INFO,KMP_STGI,0,"STGI",
		"Pole position is '%s' (%u).\n",
		stgi->pole_pos ? "right" : "left", stgi->pole_pos );

	if ( stgi->narrow_start > 1 )
	    print_check_error(&chk,CMOD_WARNING,KMP_STGI,0,"STGI",
		"Undefined value for narrow mode: %u\n",stgi->narrow_start);
	else if ( mode & CMOD_INFO )
	    print_check_error(&chk,CMOD_INFO,KMP_STGI,0,"STGI",
		"Start mode is '%s' (%u).\n",
		stgi->narrow_start ? "narrow" : "wide", stgi->narrow_start );

	if ( stgi->unknown_0a )
	    print_check_error(&chk,CMOD_HINT,KMP_STGI,0,"STGI",
		"Speed modifier is set to %5.3f (%04x/hex).\n",
		SpeedMod2float(stgi->unknown_0a), stgi->unknown_0a);
	else if ( mode & CMOD_INFO )
	    print_check_error(&chk,CMOD_INFO,KMP_STGI,0,"STGI",
		"No speed modifier set (0000/hex).\n" );
    }


    //--- terminate

    if ( mode & CMOD_GLOBAL_STAT )
    {
	global_warn_count += chk.warn_count;
	global_hint_count += chk.hint_count;
	global_info_count += chk.info_count;
    }

    if ( mode & (CMOD_FOOTER|CMOD_FORCE_FOOTER) )
    {
	static ccp hint
	    = " => see http://szs.wiimm.de/cmd/wkmpt/check#desc for more info.";

	ccp bind;
	char hbuf[60], ibuf[60];
	if (chk.info_count)
	{
	    snprintf(ibuf,sizeof(ibuf)," and %s%u info%s%s",
		chk.col.info,
		chk.info_count, chk.info_count>1 ? "s" : "",
		chk.col.heading );
	    bind = ",";
	}
	else
	{
	    *ibuf = 0;
	    bind = " and";
	}

	if (chk.hint_count)
	    snprintf(hbuf,sizeof(hbuf),"%s %s%u hint%s%s",
		bind, chk.col.hint,
		chk.hint_count, chk.hint_count>1 ? "s" : "",
		chk.col.heading );
	else
	    *hbuf = 0;

	if (chk.warn_count)
	    fprintf(stdlog,
			" %s=> %s%u warning%s%s%s%s for %s:%s\n%s%s%s\n\n",
			chk.col.heading, chk.col.warn,
			chk.warn_count, chk.warn_count == 1 ? "" : "s",
			chk.col.heading,
			hbuf, ibuf, GetNameFF(kmp->fform,0), kmp->fname,
			chk.col.heading, hint, chk.col.reset );
	else if ( mode & CMOD_FORCE_FOOTER || chk.hint_count )
	    fprintf(stdlog," %s=> No warnings%s%s for %s:%s\n%s%s%s\n\n",
			chk.col.heading, hbuf, ibuf, GetNameFF(kmp->fform,0), kmp->fname,
			chk.col.heading, hint, chk.col.reset );
    }

    return chk.warn_count;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			command stgi			///////////////
///////////////////////////////////////////////////////////////////////////////

enumError cmd_stgi()
{
    stdlog = stderr;
    raw_data_t raw;
    InitializeRawData(&raw);

    if (print_header)
	printf("\n  start pos  speed laps LC : file\n%.*s\n",
		79, Minus300 );

    enumError max_err = ERR_OK;
    ParamList_t * param;
    for ( param = first_param; param; param = param->next )
    {
	NORMALIZE_FILENAME_PARAM(param);
	enumError err = LoadRawData(&raw,false,param->arg,"/course.kmp",opt_ignore>0,0);
	if ( err == ERR_NOT_EXISTS || err > ERR_WARNING && opt_ignore )
	    continue;
	if ( err > ERR_WARNING )
	{
	    if ( max_err < err )
		max_err = err;
	    continue;
	}

	if ( opt_ignore && raw.fform != FF_KMP && raw.fform != FF_KMP_TXT )
	    continue;

	kmp_t kmp;
	err = ScanRawDataKMP(&kmp,true,&raw,CMOD_SILENT);
	if ( err > ERR_WARNING )
	{
	    ResetKMP(&kmp);
	    if ( max_err < err )
		max_err = err;
	    continue;
	}

	uint lap_count = 0;
	const uint n_ckpt =  kmp.dlist[KMP_CKPT].used;
	if (n_ckpt)
	{
	    const kmp_ckpt_entry_t *ckpt = (kmp_ckpt_entry_t*)kmp.dlist[KMP_CKPT].list;
	    uint i;
	    for ( i = 0; i < n_ckpt; i++, ckpt++ )
		if (!ckpt->mode)
		    lap_count++;
	}

	const kmp_stgi_entry_t *stgi = (kmp_stgi_entry_t*)kmp.dlist[KMP_STGI].list;
	const uint n_stgi = kmp.dlist[KMP_STGI].used;
	if (!n_stgi)
	    printf("%s   no STGI found     -%s", colout->warn, colout->reset );
	else if ( opt_ignore > 1
		&& lap_count == 1
		&& stgi->pole_pos < 2
		&& stgi->narrow_start < 2
		&& !stgi->unknown_0a
		&& stgi->lap_count == 3
		)
	{
	    ResetKMP(&kmp);
	    continue;
	}
	else
	{
	    if ( stgi->pole_pos == 0 )
		fputs(" left ",stdout);
	    else if ( stgi->pole_pos == 1 )
		fputs(" right",stdout);
	    else
		printf("%s%5u%s ", colout->warn, stgi->pole_pos, colout->reset);

	    if ( stgi->narrow_start == 0 )
		fputs("  wide ",stdout);
	    else if ( stgi->narrow_start == 1 )
		fputs(" narrow",stdout);
	    else
		printf("%s%5u%s  ", colout->warn, stgi->narrow_start, colout->reset );

	    if (stgi->unknown_0a)
	    {
		float speed = SpeedMod2float(stgi->unknown_0a);
		printf("%s %4.2f%s",
		    speed == 1.0 ? "" : speed <= 1.0 ? colout->info : colout->hint,
		    speed, colout->reset );
	    }
	    else
		fputs("  -- ",stdout);

	    printf(" %s%3u%s",
		    stgi->lap_count == 3 ? ""
			    : stgi->lap_count < 3 ? colout->info : colout->hint,
		    stgi->lap_count, colout->reset );
	}
	printf(" %s%3u%s : %s\n",
		    lap_count == 1 ? "" : colout->hint,
		    lap_count, colout->reset,
		    param->arg );
	ResetKMP(&kmp);
    }
    if (print_header)
	printf( "%.*s\n\n", 79, Minus300 );

    ResetRawData(&raw);
    return max_err;
}

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    END				///////////////
///////////////////////////////////////////////////////////////////////////////

